package WO::Client::Common;

use strict;
use warnings;

use base qw( WO::Class );

use Data::Dumper;
use LWP::UserAgent qw( post );
use MIME::Base64 qw( decode_base64 encode_base64 );
use Compress::Zlib qw( compress uncompress );
use Digest::MD5 qw( md5_hex );
use HTTP::Request::Common qw( POST );
use Text::CSV_XS;
use Archive::Zip qw( :ERROR_CODES :CONSTANTS );
use FindBin qw( $Bin );
use File::Find qw( find );
use File::Remove qw( remove );
use File::Spec;
use File::Path qw( mkpath );
use File::Basename;
use POSIX qw( strftime );

sub sync_fo_information { die "sync_fo_information() method must be overriden..." }
sub sync_bk_information { die "sync_bk_information() method must be overriden..." }

sub init {
  my $s = shift;
  my %a = ( dir => undef, manual => undef, total_active => 0, info => {}, @_ );

  $s->_misc_validations( total_active => $a{total_active} );

  my $xml = $s->info();

  die "You can't have 'update_from_server' and 'populate_from_fo' at the same time in a <frontoffice>..."
    if $xml->{update_from_server} && $xml->{populate_from_fo};

  $s->handle_custom_files();
  
  if (exists $xml->{execute_command_before} && $xml->{execute_command_before}) {
    my $command = $s->_decode_command($xml->{execute_command_before});
    $s->debug("Will execute command: $command");
    my $res = system($command);
    $s->debug("Command '$command' returned an error: $res") unless $res == 0;
  }

  if (defined $xml->{weckoffice}->{version} && ($xml->{weckoffice}->{version} == 2)) { 
    my $cconfig = $s->company_configuration();

    $cconfig->{client_method} = $cconfig->{client_method} ? $cconfig->{client_method} : 0;

    if ($cconfig->{client_method} == 0) {
      #BO => FO
      $xml->{update_from_server} = 1;
      $xml->{populate_from_fo} = 0;
    } else {
      $xml->{update_from_server} = 0;
      $xml->{populate_from_fo} = 1;
=pod
      $xml->{sync_users} = 0;
      $xml->{sync_articles} = 0;
      $xml->{sync_families} = 0;
      $xml->{sync_clients} = 0;
      $xml->{sync_parameters} = 0;
      $xml->{sync_locals} = 0;
      $xml->{sync_menus} = 0;
=cut
    }
  } else {
    if (! $xml->{update_from_server} && ! $xml->{populate_from_fo}) {
      $s->debug("WARNING!!!!!! both 'update_from_server' and 'populate_from_fo' are turned off...");
    }
  }

  #
  # Custom user transfers
  #

  $s->handle_custom_transfers( $xml );

  my $company = $xml->{weckoffice}->{company};
  my $store   = $xml->{weckoffice}->{store};

  ###################################################################
#  if (uc($company) eq 'EUREST' && $store eq 'CODE:1') {
#    my $str = `net view \\\\192.168.0.100`;
#    #my $str = `ping -n 2 192.168.0.1`;
#    print STDERR "### [192.168.0.100] $str\n";
#
#    $str = `net view \\\\192.168.0.2`;
#    #$str = `ping -n 2 192.168.0.2`;
#    print STDERR "### [192.168.0.2] $str\n";
#  }
  ###################################################################

  $s->sync_fo_information() if $xml->{populate_from_fo};

  if ($s->_check_can_sync($store)) {
    $s->sync_bk_information() if $xml->{update_from_server};
  } else {
    $s->debug("WILL NOT SYNC DATE IS NOT YET OK");
  }

  if (exists $xml->{execute_command_after} && $xml->{execute_command_after}) {
    my $command = $s->_decode_command($xml->{execute_command_after});
    $s->debug("Will execute command: $command");
    my $res = system($command);
    $s->debug("Command '$command' returned an error: $res") unless $res == 0;
  }

  $s->update_last_sync_time();

  ###################################################################
#  my $dsep = $s->directory_separator();
#  my $dir = $xml->{dir} || "";
#
#  if (-d $dir."/files/data") {
#    &find( sub {
#      return unless -f $File::Find::name;
#
#      if (/\.000_0/) {
#        $s->own_remove( $File::Find::name );
#      }
#    }, ( $dir."/files/data" ));
#  }
  ###################################################################

  $s->debug("Done.");

  return $s;
}

sub dir { shift()->{dir} }

# defines if the client was started manually
sub manual { shift()->{manual} }

sub weckoffice_url {
  my $s = shift;

  # service xml
  my $xml = $s->info();

  # weckoffice info
  my $weckoffice = $xml->{weckoffice};

  $ENV{WECKOFFICE_SERVER} = $weckoffice->{host} if ($weckoffice->{host} =~ /\.v2\./);

  my $host = $ENV{WECKOFFICE_SERVER};#$weckoffice->{host};
  my $port = $ENV{WECKOFFICE_SERVER_PORT} || 80;#$weckoffice->{port} || 80;

  my $uri = "Receiver";
   
  if (defined $weckoffice->{version} && ($weckoffice->{version} == 2)) {
    $port = 5001;
    $uri = 'wbo/api/client';

    $ENV{WECKOFFICE_SERVER} = $weckoffice->{host};
    $ENV{WECKOFFICE_SERVER_PORT} = $port;
  }

  return "http://$host:$port/$uri";
}

sub weckoffice_auth {
  my $s = shift;

  my $xml = $s->info();

  my $dsep = $s->directory_separator();

  my $st_name = $xml->{weckoffice}->{store};
  $st_name =~ s/\:/_/g;

  my $auth_key_file = $Bin.$dsep.'.auth_'.$xml->{weckoffice}->{company}.'_'.$st_name.'.key';
  my $auth_key_content = "";

  if (-e $auth_key_file) {
    local $/=undef;

    if (open(F, "< $auth_key_file")) {
      $auth_key_content = <F>;
      close(F);
    }
  }

  # auth info
  my $auth = {
    weckoffice_host     => $xml->{weckoffice}->{host},
    weckoffice_login    => $xml->{weckoffice}->{login},
    weckoffice_passwd   => $xml->{weckoffice}->{passwd},
    weckoffice_company  => $xml->{weckoffice}->{company},
    weckoffice_store    => $xml->{weckoffice}->{store},
    weckoffice_signature => $xml->{weckoffice}->{signature},
    weckoffice_auth_key => $auth_key_content,
  };

  return $auth;
}

sub weckoffice_send_data {
  my $s = shift;
  my $d = shift; # data to send to weckoffice
  my %a = ( dont_unshrink_received_content => 0, @_ );

  my $xml = $s->info();

  my $auth = $s->weckoffice_auth();

  my $url = $s->weckoffice_url();

  my $comm = LWP::UserAgent->new();
  $comm->timeout(7200);
  $comm->default_header('host' => $xml->{weckoffice}->{host});
  
  my $proxy_host = $ENV{WECKOFFICE_PROXY_HOST} || "";
  my $proxy_port = $ENV{WECKOFFICE_PROXY_PORT} || "";
  my $proxy_user = $ENV{WECKOFFICE_PROXY_USER} || "";
  my $proxy_pass = $ENV{WECKOFFICE_PROXY_PASS} || "";
  
  if (defined $xml->{weckoffice}->{version} && ($xml->{weckoffice}->{version} == 2)) { 
    $url .= "/".$d->{action};
  }
  
#  $s->debug(" CALLING $url ");                 

  if ($proxy_host) {
    my $port = $proxy_port ne "" ? ":$proxy_port" : "";
    $s->debug("Will use proxy $proxy_host with port $proxy_port");
    if ($proxy_host =~ /^http/i) {
      $comm->proxy( ['http'] => $proxy_host.$port );
    } else {
      $comm->proxy( ['http'] => "http://".$proxy_host.$port );
    }
  }

  my $r = "";

  # we're sending a file upload
  if (exists $d->{file_upload} && $d->{file_upload}) {
    my $upl = $d->{file_upload};
    delete $d->{file_upload};

    # file is empty
    if (-s $upl == 22) {
      $s->debug("Won't UPLOAD file, ZIP is empty...");
      return "";
    }
    else {
      $s->debug("Will UPLOAD file $upl (".(-s $upl)." bytes)");

      my $req = POST(
        $url,
        Content_Type => 'form-data',
        Content => [ file_upload => [ $upl ], %$d, %$auth ],
      );

      $r = $comm->request($req);
    }
  }
  # we're sending something else
  else {
    $| = 1;
    $r = $comm->post( $url, { %$d, %$auth } );
  }

  # use proxy auth
  if (defined $r && $proxy_user) {
    $r->proxy_authorization_basic($proxy_user, $proxy_pass);
  }
  
  if ($r->code() eq '200') {
    my $cnt = $r->content();
    
    if ($cnt) {
      if ($a{dont_unshrink_received_content}) {
        return $cnt;
      } else {
        my $a = $s->unshrink($cnt);
        return defined $a ? $$a : "";
      }
    } else {
      return 1;
    }
  } else {
    die "weckOFFICE call to ".$d->{action}." returned an error (error code ".$r->code()."\n\n\n".$r->content().") to us...";
  }
}

sub update_last_sync_time {
  my $s = shift;

  $s->debug("update_last_sync_time()");

  $s->weckoffice_send_data( { action => 'UpdateLastSyncTime',
                            }
                          );
}

sub encode_fo_type {
  my $s = shift;

  my $w = ref($s);

  # background compatibility
  $w =~ s/WO::Client:://;
  $w .= "::System";

  $w =~ s/\:/\@/g;

  return $w;
}

sub directory_separator {
  my $self = shift;

  if ($^O eq 'MSWin32') {
    return "\\";
  } else {
    return '/';
  }
}

sub unshrink {
  my $s = shift;
  my $str = shift;
  my $VAR1;

  my $cnt = uncompress(decode_base64($str));

  if ($cnt) {
    eval "$cnt";
    return $VAR1;
  }
}

sub shrink {
  my $s = shift;
  my $ref = shift;
  return encode_base64(compress(Dumper($ref)));
}

sub write_csv_from_iter {
  my $s = shift;
  my( $file, $db_iter, $h ) = @_;

  $h->{control_index} ||= "";
  $h->{control_hash} ||= {};

  open(F, "> ".$file) or die "Can't write to ".$file.": $!";

  my $csv = Text::CSV_XS->new( { binary => 1, eol => "\n", quote_char  => '"', escape_char => '"' } );
  #my $csv = Text::CSV_XS->new( { binary => 1, eol => "\n" } );

  my $c = 0;

  while( my $row = &$db_iter() ) {
    $c++;

    if ($h->{control_index}) {
      next if exists $h->{control_hash}->{ $row->[$h->{control_index}] };
    }

    my $status = $csv->print(\*F, $row);
  }

  $s->debug("Wrote $c records into $file");

  close(F);
}

sub handle_custom_files {
  my $s = shift;
    
  my $xml = $s->info();
  my $dsep = $s->directory_separator();

  if (exists $xml->{files}) {
    my $ref = exists $xml->{files}->{file} ? $xml->{files}->{file} : undef;
    
    if (defined $ref) {
      $s->debug("handle_custom_files()");

      my @files = ref($ref) eq 'HASH' ? ( $ref ) : @$ref;
      
      foreach my $file (@files) {
        my( $path, $content, $copy_when_update ) = ( $file->{path}, $file->{content}, $file->{copy_when_update} );
        
        $path ||= "";
        $content ||= "";
        $copy_when_update ||= "";
        
        if ($path && $content) {
          my $md5_file = $path.".md5";
          my $md5 = undef;
          my $md5_current = undef;
          my $write_file = 0;
          
          if (-e $path && -e $md5_file) {
            $md5_current = &md5_hex($content);
          
            open(F, "< $md5_file") or die "Can't read $md5_file: $!";
            local $/ = undef;
            $md5 = <F>;
            close(F);
            
            if ($md5_current ne $md5) {
              $s->debug("File $path has an MD5 and changed, updating...");
              $write_file = 1;
            }
            else {
              $s->debug("File $path has an MD5 and didn't change, skipping...");
            }
          }
          else {
            $s->debug("File $path doesn't have an MD5, will generate one and save it...");
            $md5_current = &md5_hex($content);
            $write_file = 1;
          }
          
          if ($write_file) {
            open(F, "> $path") or die "Can't write to $path: $!";
            print F $content;
            close(F);
            
            if ($copy_when_update) {
              open(F, "> $copy_when_update") or die "Can't write to $copy_when_update: $!";
              print F $content;
              close(F);
            }
                        
            open(F, "> $md5_file") or die "Can't write to $md5_file: $!";
            print F $md5_current;
            close(F);
          }
        }
      }
    }
  }
}

sub handle_custom_transfers {
  my $s = shift;
  my $xml = shift;

  my $dsep = $s->directory_separator();

  my %transfers = $s->decode_transfer_rules( $xml );

  if (exists $transfers{send}) {
    $s->debug("Will handle files send...");

    foreach my $o (@{$transfers{send}}) {
      $o->{active} ||= 0;

      if (! $o->{active}) {
        $s->debug("Ignoring '".$o->{id}."'...");
        next;
      }

      # <send id="sells" from="c:/fo/export_to_send" recursive="1" remove_after_send="1" cache="1" active="1" />

      $s->debug("RULE: ".$o->{id});

      if (-d $o->{from}) {
        $o->{recursive} ||= 0;
        $o->{use_cache} ||= 0;
        $o->{filter} ||= "";
        $o->{single_remote_file} ||= 0;
        $o->{additional_stores} ||= "";
        my @files;

        my $filter = $o->{filter};

        my $cache_file = $Bin.$dsep.'send#'.$o->{id}.'.cache';
        my %cache = ();

        #$s->debug("will FORCE remove cache...");
        #$s->own_remove($cache_file);

        if ($o->{use_cache} && -f $cache_file) {
          %cache = $s->read_cache_file( $cache_file );
        }

        if ($o->{recursive}) {
          &find({ wanted => sub {
            return unless -f $File::Find::name;

            if ($o->{use_cache}) {
              push @files, $File::Find::name unless exists $cache{ $File::Find::name };
              $cache{ $File::Find::name } = 1;
            } else {
              push @files, $File::Find::name;
            }
          } }, $o->{from});
        } else {
          opendir(DIR, $o->{from}) || die "can't opendir ".$o->{from}.": $!";
          while(my $d = readdir(DIR)) {
            next unless -f $o->{from}.$dsep.$d;

            if ($o->{use_cache}) {
              push @files, $o->{from}.$dsep.$d unless exists $cache{ $o->{from}.$dsep.$d };
              $cache{ $o->{from}.$dsep.$d } = 1;
            } else {
              push @files, $o->{from}.$dsep.$d;
            }
          }
          closedir(DIR);
        }

        if ($filter) {
          my $filters = _decode_filter($filter);
          my $check = "1" x scalar(@$filters);

          my @final = ();

          for my $file (@files) {
            my $control;

            foreach my $filter (@$filters) {
              my $cond = $filter->{condition};

              if ($filter->{not} && $file !~ /$cond/i) {
                $control .= 1;
              }
              elsif (! $filter->{not} && $file =~ /$cond/i) {
                $control .= 1;
              } else {
                $control .= 0;
              }
            }

            push @final, $file if $control eq $check;
          }

          @files = @final;
        }

        $s->debug("Found ".scalar(@files)." files...");

        if (scalar @files) {
          my $files_zip = File::Spec->tmpdir()."/custom_transfer_\@\@".$o->{id}."\@\@.zip";

          my $z = $s->build_zip_file( $files_zip, \@files, $o->{from} );

          if (! defined $z) {
            $s->debug("Error creating zip file from collected files...");
          } else {
            eval {
              $s->weckoffice_send_data( { action => 'FrontofficeReceiveData',
                                          file_upload => $files_zip,
                                          single_remote_file => $o->{single_remote_file},
                                          additional_stores => $o->{additional_stores},
                                        }
                                      );
            };

            if ($@) {
              $s->debug("Error sending zip file to server: $@");
            } else {
              $s->own_remove($files_zip);

              if ($o->{use_cache} && scalar keys %cache) {
                $s->write_cache_file( $cache_file, \%cache );
              }

              if ($o->{remove_after_send}) {
                for (@files) {
                  $s->debug("Will remove $_");
                  $s->own_remove($_)
                }
              }
            }
          }
        }
      }
    }
  }

  if (exists $transfers{receive}) {
    $s->debug("Will handle file receivements...");
    foreach my $o (@{$transfers{receive}}) {
      $o->{active} ||= 0;
      $o->{to}     ||= "";

      if (! $o->{active}) {
        $s->debug("Ignoring '".$o->{id}."'...");
        next;
      }

      if (! $o->{to}) {
        $s->debug("Missing 'to' parameter, ignoring...");
        next;
      }

      #<receive id="sells" to="c:/fo/export_received" active="1" />

      my $cnt;

      eval {
        $cnt = $s->weckoffice_send_data( { action => 'BackofficeReceiveCustomFiles',
                                           id => $o->{id},
                                         },
                                         dont_unshrink_received_content => 1,
                                       );
      };

      if ($@) {
        $s->debug("Error receiving zip file from server: $@");
      } else {
        if (length($cnt) > 1) {
          mkpath($o->{to}) unless -d $o->{to};

          print STDERR "RECEIVED: ".length($cnt)." bytes\n";

          my $files_zip = File::Spec->tmpdir()."/files".$$."_".time.".zip";

          if (open(F, "> $files_zip")) {
            binmode(F);
            print F $cnt;
            close(F);

            my $zip1 = Archive::Zip->new();

            if ($zip1->read( $files_zip ) == AZ_OK) {
              my @members1 = $zip1->members();

#              &mkpath($fo_dir."/".$custom) unless $fo_dir."/".$custom;

              foreach my $member1 (@members1) {
                my $tmp_f = File::Spec->tmpdir()."/".$member1->{fileName};
                $s->own_remove($tmp_f) if -e $tmp_f;

                $zip1->extractMember($member1, $tmp_f);

                my $zip2 = Archive::Zip->new();

                if ($zip2->read( $tmp_f ) == AZ_OK) {
                  my @members2 = $zip2->members();

                  foreach my $member2 (@members2) {
                    my( $file, $folder, $suffix ) = fileparse($member2->{fileName});
                    my $out_f = $o->{to}."/".$file;

                    $s->debug("Will extract received file: $file to: $out_f");

                    my $atime = undef;
                    my $mtime = undef;

                    if (-e $out_f) {
                      ( $atime, $mtime ) = (stat($out_f))[8,9];
                    }

                    $zip2->extractMember($member2, $out_f);

                    if (defined $atime) {
                      my $st = utime $atime, $mtime, $out_f or die "utime: $!\n";
                    }
                  }
                }

                $s->own_remove($tmp_f) if -e $tmp_f;
              }
            }

            $s->own_remove($files_zip) if -e $files_zip;

            # after received, remove from server
            $s->weckoffice_send_data( { action => 'BackofficeCleanupCustomFiles',
                                        id => $o->{id},
                                      },
                                    );
          }
        }
      }
    }
  }
}

#
# Receive files associated with a tree
#

sub receive_backoffice_tree_files {
  my $s = shift;

  $s->debug("receive_backoffice_tree_files()");

  # service xml
  my $xml = $s->info();

  my $dir = $xml->{dir};
  my $dsep = $s->directory_separator();
  my $out_dir = $xml->{use_import_dir} ? 'import' : 'data';

  my $cnt;

  eval {
    $cnt = $s->weckoffice_send_data( { action => 'BackofficeReceiveTreeFiles',
                                     },
                                     dont_unshrink_received_content => 1,
                                   );
  };

  if ($@) {
    $s->debug("Error receiving zip file from server: $@");
  } else {
     if (length($cnt) > 1) {
       my $tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));

       if (open(F, "> $tmp_file")) {
         binmode(F);
         print F $cnt;
         close(F);
       }

       my $zip1 = Archive::Zip->new();

       $s->debug("Will extract files...");

       if ($zip1->read( $tmp_file ) == AZ_OK) {
         my @members1 = $zip1->members();

         foreach my $member1 (@members1) {
           my $out_f = $dir.$dsep."files".$dsep.$out_dir.$dsep.$member1->{fileName};

           $s->debug("Will write '".$member1->{fileName}."' into '$out_f'");
           $zip1->extractMember($member1, $out_f);
         }
       }
     }
     else {
       $s->debug("Received an empty content...");
     }
  }
}

sub decode_transfer_rules {
  my $s = shift;
  my $xml = shift;

  my %info = ();

  if (exists $xml->{files_transfer_rules}) {
    # send rules
    if (exists $xml->{files_transfer_rules}->{send}) {
      if (exists $xml->{files_transfer_rules}->{send}->{id}) {
        push @{ $info{send} }, $xml->{files_transfer_rules}->{send};
      } else {
        for (keys %{$xml->{files_transfer_rules}->{send}}) {
          $xml->{files_transfer_rules}->{send}->{$_}->{id} = $_;
          push @{ $info{send} }, $xml->{files_transfer_rules}->{send}->{$_};
        }
      }
    }

    # receive rules
    if (exists $xml->{files_transfer_rules}->{receive}) {
      if (exists $xml->{files_transfer_rules}->{receive}->{id}) {
        push @{ $info{receive} }, $xml->{files_transfer_rules}->{receive};
      } else {
        for (keys %{$xml->{files_transfer_rules}->{receive}}) {
          $xml->{files_transfer_rules}->{receive}->{$_}->{id} = $_;
          push @{ $info{receive} }, $xml->{files_transfer_rules}->{receive}->{$_};
        }
      }
    }
  }

  return wantarray? %info : \%info;
}

sub build_zip_file {
  my $s = shift;
  my $fn = shift;
  my $files = shift;
  my $fb = shift;

  return unless scalar @$files;
  unlink $fn if -e $fn;

  my $zip = Archive::Zip->new();

  for (@$files) {
    my $n = $_;
    $n =~ s/^$fb// if defined $fb;
    $zip->addFile( $_, $n );
  }

  return undef unless $zip->writeToFileNamed( $fn ) == AZ_OK;

  return 1;
}

sub read_cache_file {
  my $s = shift;
  my $f = shift;

  $s->debug("read_cache_file");

  my %r = ();

  if (open(F, "< $f")) {
    while(<F>) {
      chop;
      $r{$_} = 1;
    }
    close(F);
  }

  return wantarray? %r : \%r;
}

sub write_cache_file {
  my $s = shift;
  my $f = shift;
  my $i = shift;

  $s->debug("write_cache_file");

  if (open(F, "> $f")) {
    for (keys %$i) {
      print F $_."\n";
    }
    close(F);
  }
}

sub tmp_dir {
  my $s = shift;

  my $tmp_dir = File::Spec->tmpdir();

  if ($tmp_dir eq "\\") {
    if ($^O eq 'MSWin32') {
      $tmp_dir = 'c:';
    } 
    else {
      $tmp_dir = '/tmp';
    }
  }

  if (-e "/sbin/init2" && -d "/home/weckOFFICE") {
    mkdir("/home/weckOFFICE/tmp") unless -d "/home/weckOFFICE/tmp";
    $tmp_dir = '/home/weckOFFICE/tmp';
  }

  return $tmp_dir;
}

sub own_remove {
  my $s = shift;
  my $f = shift;
  return unless defined $f && -e $f;

  my $dsep = $s->directory_separator();

#  if (-e $Bin.$dsep."srm") {
#    system($Bin.$dsep."srm -fD \"$f\"");
#    return;
#  }

  my $st = &remove($f);

  if ($st > 0) {
    if ($^O eq 'linux') {
      $st = system("rm -f \"$f\" 2> /dev/null");
      #print STDERR "### $st\n";

      if ($st > 0) {
        $st = system("sudo rm -f \"$f\" 2> /dev/null");
      }
    } else {
      $st = system("del /F /Q \"$f\" 2>> nul");
    }

    if ($st > 0) {
#      print STDERR "Error deleting file '$f': $st\n";
    }
  }
}

sub company_configuration {
  my $s = shift;
  return $s->{__company_configuration} if exists $s->{__company_configuration};

  my $VAR1 = {};

  my $company_config = $s->weckoffice_send_data( { action => 'CompanyConfiguration',
                                                 }
                                               );

  eval "$company_config";
  my %company_config = defined $VAR1 && ref($VAR1) eq 'HASH' ? %$VAR1 : ();

  return $s->{__company_configuration} = \%company_config;
}

sub vfh {
  my $s = shift;

  my( $h, $k ) = @_;

  return undef unless exists $h->{$k};
  return $h->{$k} if defined $h->{$k} && $h->{$k} ne '';
  return undef;
}

sub _misc_validations {
  my $s = shift;
  my %a = ( total_active => 0, visible => 1, @_ );

  my $xml = $s->info();

  $s->debug("_misc_validations()");

  my $dsep = $s->directory_separator();

  my $st_name = $xml->{weckoffice}->{store};
  $st_name =~ s/\:/_/g;

  my $auth_key_file = $Bin.$dsep.'.auth_'.$xml->{weckoffice}->{company}.'_'.$st_name.'.key';

  if (! -e $auth_key_file) {
    my $res = $s->weckoffice_send_data( { action => 'BackofficeGetAuthorizationKey',
                                        }
                                      );

    if ($res ne '' && open(F, "> $auth_key_file")) {
      print F $res;
      close(F);
    }
  }

  # check run mode
  $xml->{update_from_server} ||= "";
  $xml->{populate_from_fo} ||= "";

  $xml->{sync_fo_config} = defined $xml->{sync_fo_config} ? $xml->{sync_fo_config} : 1;

  if ($a{total_active} > 1) {
    $s->debug("You've more than one active <frontoffice>, will disable 'sync_fo_config' for all of them...");
    $xml->{sync_fo_config} = 0;
  }

  $xml->{sync_fo_config} = 0 unless $a{visible};

  if ($xml->{sync_fo_config}) {
    # check client config
    my $res = $s->weckoffice_send_data( { action => 'StoreClientConfig',
                                        }
                                      );
                                      
    # weckoffice has a configuration
    if (length($res) > 1) {
      my $remote_md5 = md5_hex($res);
      my $cur_md5 = md5_hex($s->config_txt());

      if ($remote_md5 ne $cur_md5) {
        $s->debug("Configuration changed in weckOFFICE, will update client and start again...");

        open(F, "> ".$s->config_file()) or die "Can't write to config file '".$s->config_file()."': $!";
        print F $res;
        close(F);

        die "#RELOAD_CLIENT#";
      } else {
        $s->debug("Remote config and client's are equal, skipping update...");
      }
    } else {
      $s->debug("weckOFFICE don't have a configuration, will send ours...");

      $s->weckoffice_send_data( { action => 'StoreClientConfigUpdate',
                                  store_config => $s->config_txt(),
                                }
                              );
    }
  }

  # Handle backup facilities
  if (ref($s) eq 'WO::Client::WinRest') {
    $xml->{files_backup} = {
      backup => [
        {
          'after' => '',
          'recursive' => '0',
          'filter' => '',
          'zip_folder' => '',
          'path' => '/home/winrest/config',
          'before' => '',
          'os' => ''
        },
        {
          'after' => '',
          'recursive' => '0',
          'filter' => '',
          'zip_folder' => '',
          'path' => '/home/winrest/files/data',
          'before' => '',
          'os' => ''
        }
      ],
    };
  }

  if (exists $xml->{files_backup} && exists $xml->{files_backup}->{backup} && ! $s->manual() && $xml->{weckoffice}->{host} !~ /eurest/) {
    my $ref = $xml->{files_backup}->{backup};
    my @rules = ref($ref) eq 'HASH' ? ( $ref ) : @$ref;

    &File::Basename::fileparse_set_fstype('MSWin32');

    my $zip = Archive::Zip->new();

    my @afters;

    my $i = 1;

    foreach my $rule (@rules) {
      $rule->{os}         ||= "";
      $rule->{path}       ||= "";
      $rule->{before}     ||= "";
      $rule->{after}      ||= "";
      $rule->{filter}     ||= "";
      $rule->{recursive}  ||= "";
      $rule->{zip_folder} ||= "";

      my $active = exists $rule->{active} ? $rule->{active} : 1;

      if (! $active) {
        $s->debug("Backup rule is inactive, skipping...");
        next;
      }

      my $folder_name = $rule->{zip_folder} ? $rule->{zip_folder} : $i;

      if (! $rule->{path}) {
        $s->debug("Missing 'path' in backup rule, skipping...");
        next;
      }

      if ($rule->{os} && lc($rule->{os}) eq 'win32') {
        next unless $^O eq 'MSWin32';
      }
      elsif ($rule->{os} && lc($rule->{os}) eq 'linux') {
        next unless $^O eq 'linux';
      }

      if ($rule->{before}) {
        my $res = system($rule->{before});
        $s->debug("Error executing 'before' condition: $res") if $res > 0;
      }

      if (-f $rule->{path}) {
        $s->debug("'filter' will be ignored because it's applied to a".
                  " file and it should be used into folders") if $rule->{filter};

        my( $name, $path, $suffix) = fileparse($rule->{path}, ());

        if ( ( -r $rule->{path} )&& ( ( -f $rule->{path} ) || ( -l $rule->{path} ) ) ){
          $zip->addFile( $rule->{path}, $folder_name."/".$_ );
        }
        else {
          #$s->debug("Won't include file ".$rule->{path});
        }
      }
      elsif (-d $rule->{path}) {
        &find( { wanted => sub {
	        if (! $rule->{recursive}) {
            return if $File::Find::dir ne $rule->{path};
          }
	  
          my $filter = $rule->{filter};

          if ($filter && $_ !~ /$filter/i) {
            return;
          }
	   
          if ( ( -r $File::Find::name )&& ( ( -f $File::Find::name ) || ( -l $File::Find::name ) ) ){
            my $rp = $rule->{path};
            my $df = $File::Find::dir;
            $df =~ s/^$rp//;

            if ($df) {
              $zip->addFile( $File::Find::name, $folder_name."/$df/".$_ );
            }
            else {
              $zip->addFile( $File::Find::name, $folder_name."/".$_ );
            }
          }
          else {
            #$s->debug("Won't include file $File::Find::name");
          }
        }, follow => 1 }, ( $rule->{path} ) );
      }

      if ($rule->{after}) {
        push @afters, $rule->{after};
      }

      $i++;
    }

    # we had some active rule
    if ($i > 1) {
      my $tmp_file = $s->tmp_dir() . $s->directory_separator() . int(rand(1000000));

      if ($zip->writeToFileNamed( $tmp_file ) != AZ_OK) {
        $s->debug("There was an error writing the backup zip file to '$tmp_file...");
      }

      for (@afters) {
        my $res = system($_);
        $s->debug("Error executing 'after' condition '$_': $res") if $res > 0;
      }

      if (-s $tmp_file) {
        eval {
          $s->weckoffice_send_data( { action => 'FrontofficeReceiveBackupData',
                                      file_upload => $tmp_file,
                                    }
                                  );

        };
      
        if ($@) {
          die("Error sending file: $@");
        } else {
          $s->debug("Backup file sent to the server...");
        }
      }
    }
  }
}

sub _decode_command {
  my $s = shift;
  my $str = shift;

  my $xml = $s->info();

  while ($str =~ s/\#(.*?)\#/MYTOK/) {
    my $code = $1;
    my $new_val = eval $code;
    $new_val = 'invalid_token' unless defined $new_val;
    $str =~ s/MYTOK/"$new_val"/;
  }

  return $str;
}

sub _decode_filter {
  my $f = shift;
  my $r = [];

  my $y = strftime("%y", localtime);
  my $m = strftime("%m", localtime);
  my $d = strftime("%d", localtime);
  my $H = strftime("%H", localtime);
  my $M = strftime("%M", localtime);

  my @filters = split /\@/, $f;

  foreach my $filter (@filters) {
    my $not = 0;
    my $cond = $filter;

    if ($filter =~ /^!(.*)$/) {
      $not = 1;
      $cond = $1;
    }

    $cond =~ s/%y/$y/g;
    $cond =~ s/%m/$m/g;
    $cond =~ s/%d/$d/g;
    $cond =~ s/%H/$H/g;
    $cond =~ s/%M/$M/g;

    push @$r, { not => $not, condition => $cond };
  }

  return $r;
}

sub _check_can_sync {
  my $s = shift;
  my $store = shift;

  $s->debug("_check_can_sync is disabled");

  return 1;
}

sub abort {
  my $s = shift;
  my $m = shift;
  die "$m\n";
}

sub get_date_from {
  my $s = shift;

  my $VAR1 = undef;
  my $str = "";

  my $day2sec = 24*60*60;
  my $date_from = strftime("%Y/%m/%d", localtime(time()-90*($day2sec)));

  eval {
    $str = $s->weckoffice_send_data( { action => 'BackofficeGetDateFrom', });
  };

  if ($@) {
    $s->debug("Error: $@");
    return $date_from;
  }

  eval $str;

  if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
    return $VAR1->[0];
  } else {
    return $date_from 
  }
}

1;




