package WO::Client::Integrators::OnlineWinRest;

use strict;
no warnings;

use base qw ( WO::Client::Common );

use File::Copy qw( move copy );
use File::Spec;
use POSIX qw( strftime );
use File::Slurp qw( read_file read_dir );
use Archive::Zip qw( :ERROR_CODES :CONSTANTS );
use File::Basename qw( fileparse );
use File::Find qw( find );
use Data::Dumper;
use FindBin qw( $Bin );
use Encode qw( encode );

use WinRest::File::Articles;
use WinRest::File::Families;
use WinRest::File::Waitress;
use WinRest::File::Clients;
use WinRest::File::Parameters;
use WinRest::File::Promotions;
use WinRest::File::Exclusions;
use WinRest::File::Automatics;
use WinRest::File::FastScreens;
use WinRest::File::HappyHours;

#
# FO to weckOFFICE
#

sub sync_fo_information {
  my $s = shift;

  $s->debug("sync_fo_information()");

  my $xml = $s->info();

  my $fo_type = $s->encode_fo_type();
  $s->debug("Encoded FO type is: $fo_type");

  # frontoffice directory
  my $dir = $xml->{dir} || "";

  # winrest sells dir
  my $sells_dir = $xml->{sells_dir} || "export";

  my $date_from = $s->get_date_from();
  $date_from =~ s/\-//g if defined $date_from;
  $date_from =~ s/\///g if defined $date_from;

  $s->debug("WILL USE $date_from");

  # remove sells after import
  my $remove_sells_after_import = $xml->{remove_sells_after_import} || 0;
  my @possible_remove = ();

  if (! -d $dir) {
    die "Frontoffice dir '$dir' don't exists, where is the frontoffice?";
  }

  # directory separator
  my $dsep = $s->directory_separator();

  my $fo_version = &get_lic_key($dir.$dsep.'config'.$dsep.'wrstsc00.lic', 'VERSION');
  my $hardlock = &get_lic_key($dir.$dsep.'config'.$dsep.'wrstsc00.lic', 'SERIALNUMBER');
  $s->debug("WinREST Version: $fo_version");
  $s->debug("WinREST Hardlock: $hardlock");

  my @pwds = &cserver_passwords();
  $s->debug("CPWDS:");
  $s->dump(\@pwds);

  $xml->{sector_master} ||= "000";

  my $file_warehouses = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmarm.'.$xml->{sector_master};
  my $file_articles = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmart.'.$xml->{sector_master};
  my $file_families = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmfam.'.$xml->{sector_master};
  my $file_waitress = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmemp.'.$xml->{sector_master};
  my $file_params = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrsttprm.'.$xml->{sector_master};
  my $file_clients = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmcli.'.$xml->{sector_master};
  my $file_locals = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrsttloc.'.$xml->{sector_master};
  my $file_promotions = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstprom.'.$xml->{sector_master};
  my $file_exclusions = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstexcl.'.$xml->{sector_master};

  my $dir_movements = $xml->{dir_movements} || undef;
  $dir_movements = $dir.$dsep.'files'.$dsep.$sells_dir unless defined $dir_movements;

  #
  # Prepare ZIP file
  #

  my $zip = Archive::Zip->new();

  my $zipFile = $xml->{dir}.$dsep.$fo_type.'_FO_DATA.zip';
  $s->own_remove( $zipFile ) if -e $zipFile;

  my @files_cleanup = ();
  $s->common_sync_fo_and_sync_bk( $xml, $zip, \@files_cleanup );

  #
  # Get unprocessed files found in the server
  #

  my $unprocessed_files = $s->weckoffice_send_data( { action => 'FrontofficeUnprocessedFiles', }
                                                  );

  my $VAR1 = {};

  my %unprocessed_files = ();
  my @unprocessed_files = split //, $unprocessed_files;
  map { $unprocessed_files{$_} = 1 } @unprocessed_files;

  # default values are 1
  $xml->{sync_locals} = exists $xml->{sync_locals} ? $xml->{sync_locals} : 1;
  $xml->{sync_menus} = exists $xml->{sync_menus} ? $xml->{sync_menus} : 1;
  $xml->{sync_warehouses} = exists $xml->{sync_warehouses} ? $xml->{sync_warehouses} : 1;

  # used to define if we want deleted fo content or not
  $xml->{filter_deleted_content}          = exists $xml->{filter_deleted_content} ? $xml->{filter_deleted_content} : 1;
  $xml->{articles_filter_deleted_content} = exists $xml->{articles_filter_deleted_content} ? $xml->{articles_filter_deleted_content} : 1;
  $xml->{clients_filter_deleted_content}  = exists $xml->{clients_filter_deleted_content} ? $xml->{clients_filter_deleted_content} : 1;
  $xml->{users_filter_deleted_content}    = exists $xml->{users_filter_deleted_content} ? $xml->{users_filter_deleted_content} : 1;
  $xml->{families_filter_deleted_content} = exists $xml->{families_filter_deleted_content} ? $xml->{families_filter_deleted_content} : 1;

  my $company_config = $s->company_configuration();
  $company_config->{winrest_use_new_code_scheme} ||= 0;

  if (exists $company_config->{disable_data_send} && $company_config->{disable_data_send}) {
    $s->debug("Data send is disabled in the company_configurations... disabling masters...");
    $xml->{sync_menus} = 0;
    $xml->{sync_locals} = 0;
    $xml->{sync_clients} = 0;
    $xml->{sync_warehouses} = 0;
    $xml->{sync_families} = 0;
    $xml->{sync_articles} = 0;
    $xml->{sync_users} = 0;
  }

  # default value are 0
  $xml->{check_sell_files_errors} ||= 0;

  # hardcode values
  $xml->{sync_users} = 1;
  $xml->{sync_articles} = 1;
  $xml->{sync_families} = 1;
  $xml->{sync_clients} = 1;
  $xml->{sync_sells} = 1;

  if ($xml->{sync_warehouses}) {
    #
    # Warehouses
    #
    if (-e $file_warehouses) {
      $zip->addFile( $file_warehouses ) unless exists $unprocessed_files{ 'wrstmarm.'.$xml->{sector_master} };
    } else {
      $s->debug("Didn't find the warehouses '$file_warehouses' file...");
    }
  }

  #
  # Parameters
  #

  if (-e $file_params) {
    $zip->addFile( $file_params ) unless exists $unprocessed_files{ 'wrsttprm.'.$xml->{sector_master} };
  } else {
    $s->debug("Didn't find the parameters '$file_params' file...");
  }

  if ($xml->{sync_articles}) {
    #
    # Articles
    #

    if (-e $file_articles) {
    	if ($xml->{filter_deleted_content} || $xml->{articles_filter_deleted_content}) {
        my $fname = &_filter_deleted_from_fo_file( $file_articles, 235 );
        push @files_cleanup, $fname;
			  $zip->addFile( $fname );
    	} else {
    		$zip->addFile( $file_articles ) unless exists $unprocessed_files{ 'wrstmart.'.$xml->{sector_master} };
    	}
    } else {
      $s->debug("Didn't find the articles '$file_articles' file...");
    }
  }

  if ($xml->{sync_families}) {
    #
    # Families
    #

    if (-e $file_families) {
    	if ($xml->{filter_deleted_content} || $xml->{families_filter_deleted_content}) {
    	  my $fname = &_filter_deleted_from_fo_file( $file_families, 198 );
        push @files_cleanup, $fname;
    	  $zip->addFile( $fname );
    	} else {
    		$zip->addFile( $file_families ) unless exists $unprocessed_files{ 'wrstmfam.'.$xml->{sector_master} };
    	}
    } else {
      $s->debug("Didn't find the families '$file_families' file...");
    }
  }

  if ($xml->{sync_users}) {
    #
    # Waitresses
    #

    if (-e $file_waitress) {
    	if ($xml->{filter_deleted_content} || $xml->{users_filter_deleted_content}) {
    	  my $fname = &_filter_deleted_from_fo_file( $file_waitress, 317 );
        push @files_cleanup, $fname;
    	  $zip->addFile( $fname );
    	} else {
    		$zip->addFile( $file_waitress ) unless exists $unprocessed_files{ 'wrstmemp.'.$xml->{sector_master} };
    	}
    } else {
      $s->debug("Didn't find the waitress '$file_waitress' file...");
    }
  }

  if ($xml->{sync_locals}) {
    #
    # Locals
    #

    if (-e $file_locals) {
      $zip->addFile( $file_locals ) unless exists $unprocessed_files{ 'wrsttloc.'.$xml->{sector_master} };
    } else {
      $s->debug("Didn't find the locals '$file_locals' file...");
    }
  }

  if ($xml->{sync_menus}) {
    #
    # Promotions
    #

    if (-e $file_promotions) {
      $zip->addFile( $file_promotions ) unless exists $unprocessed_files{ 'wrstprom.'.$xml->{sector_master} };
    } else {
      $s->debug("Didn't find the promotions '$file_promotions' file...");
    }
  }

    #
    # Exclusions
    #

    if (-e $file_exclusions) {
      $zip->addFile( $file_exclusions );
    }

  if ($xml->{sync_clients}) {
    #
    # Waitresses
    #

    if (-e $file_clients) {
    	if ($xml->{filter_deleted_content} || $xml->{clients_filter_deleted_content}) {
    	  my $fname = &_filter_deleted_from_fo_file( $file_clients, 387 );
        push @files_cleanup, $fname;
    	  $zip->addFile( $fname );
    	} else {
    		$zip->addFile( $file_clients ) unless exists $unprocessed_files{ 'wrstmcli.'.$xml->{sector_master} };
    	}
    } else {
      $s->debug("Didn't find the clients '$file_clients' file...");
    }
  }

  #
  # Movements
  #

  if ($xml->{sync_sells} && -d $dir_movements) {
    my $found_something = $s->sync_movements($zip, \@possible_remove, $date_from, 'sync_fo_information');
  }

  my $file_sizes_watch = $xml->{file_sizes_watch} || "";

  if ($file_sizes_watch) {
    my $file_sizes = $s->tmp_dir() . $dsep . "file_sizes.txt";
    open(F, "> $file_sizes");
    close(F);
    $zip->addFile( $file_sizes );
  }

  die "write error $zipFile" unless $zip->writeToFileNamed( $zipFile ) == AZ_OK;

  #
  # Send ZIP file to the server
  #

  eval {
    $s->weckoffice_send_data( { action => 'FrontofficeReceiveData',
                                file_upload => $zipFile,
                              }
                            );
  };

  $s->own_remove($zipFile) if -e $zipFile;
  $s->own_remove($_) for @files_cleanup;

  if ($@) {
    die "Error sending file: $@";
  } else {
    if ($file_sizes_watch) {
      $s->_save_file_sizes();
    }

    # remove sales after import, if requested
    if ($remove_sells_after_import && scalar @possible_remove) {
      map { $s->own_remove($_) } @possible_remove;
    }
  }

  $s->debug("Done.");
}

#
# Brings weckOFFICE information to FO
#

sub sync_bk_information {
  my $s = shift;

  $s->debug("sync_bk_information - Not implemented...");
}

#
# Sync movements
#

sub sync_movements {
  my $s = shift;
  my $zip = shift;
  my $remove = shift;
  my $date_from = shift;
  my $caller = shift || 'none';

  $s->debug("sync_movements()");

  my $found_something = undef;

  # service xml
  my $xml = $s->info();

  my $check_dates_fast = $xml->{check_dates_fast} || 0;

  my $fo_type = $s->encode_fo_type();

  # frontoffice directory
  my $dir = $xml->{dir} || "";

  my $file_sizes_watch = $xml->{file_sizes_watch} || "";

  # winrest sells dir
  my $sells_dir = $xml->{sells_dir} || "export";

  # directory separator
  my $dsep = $s->directory_separator();

  # movements dir
  my $dir_movements = $xml->{dir_movements} || undef;
  $dir_movements = $dir.$dsep.'files'.$dsep.$sells_dir unless defined $dir_movements;

  # remove sells after import
  my $remove_sells_after_import = $xml->{remove_sells_after_import} || 0;

  my $number_of_days_to_sync = $xml->{number_of_days_to_sync} || 0;

  my @files = &read_dir($dir_movements);

#  use Data::Dumper;
#  print STDERR Dumper \@files;

  #if ($xml->{weckoffice}->{host} eq 'eurest.weckoffice.com' && $xml->{weckoffice}->{store} eq 'CODE:240') {
  #if ($xml->{weckoffice}->{host} eq 'eurestcafe.weckoffice.com') {
  #if ($xml->{weckoffice}->{host} eq 'eurestmakro.weckoffice.com') {
  #if ($xml->{weckoffice}->{host} eq 'trincanela3.weckoffice.com') {
  #if ($xml->{weckoffice}->{host} eq 'giardino.weckoffice.com') {
  #  $s->dump(\@files);
  #}

  if ($xml->{check_sell_files_errors} && scalar @files) {
    $s->debug("Checking sell files for errors...");
    my $res = $s->_check_errors_in_sales($dir_movements, \@files);

    die "Sell files have errors, aborting..." unless $res;
  }

  #
  # Get unprocessed files found in the server
  #

  my $unprocessed_files = $s->weckoffice_send_data( { action => 'FrontofficeUnprocessedFiles',
                                                    }
                                                  );
  my $VAR1 = {};

  my $company_config = $s->company_configuration();

  my %valid_files = (
    b => 1,
    c => 1,
    d => 1,
    g => 1,
    h => 1,
    r => 1,
    s => 1,
    x => 1,
    n => 1,
    o => 1,
  );

  if ($file_sizes_watch) {
    my $sizes = $s->_read_file_sizes();

    foreach my $file (@files) {
      my $file_size = -s $dir_movements . $dsep . $file;

      my( $file_type, $date_str, $sector ) = $file =~ /^(\w)(\d+)\.(\d+)$/;
      next unless defined $date_str;

      my( $year, $mon, $day ) = $date_str =~ /^(\d\d)(\d\d)(\d\d)/;
      $year = "20".$year;

      if (defined $date_from && $year.$mon.$day < $date_from) {
        $s->debug("Ignoring date $year/$mon/$day with the 'date_from' clause...");
        next;
      }

      if (! exists $sizes->{$file} || $sizes->{$file} != $file_size) {
        my( $c, $rest ) = $file =~ /^(.)(.*)$/;
        my $member;

        if (exists $valid_files{ lc($c) }) {
          if (-e $dir_movements.$dsep.lc($c).$rest) {
            $s->debug("Will add file $dir_movements".$dsep.lc($c).$rest);
            $member = $zip->addFile( $dir_movements.$dsep.lc($c).$rest );
          }

          if (-e $dir_movements.$dsep.uc($c).$rest) {
            $s->debug("Will add file $dir_movements".$dsep.uc($c).$rest);
            $member = $zip->addFile( $dir_movements.$dsep.uc($c).$rest );
          }

          $found_something = 1;
          $sizes->{$file} = $file_size;
        }
      }
    }

    $s->{__file_sizes} = $sizes;

    if (! $found_something) {
      $s->debug("!! Using file sizes cache but nothing changed !!");
    }

    return $found_something
  }

  my $act = $check_dates_fast ? 'FrontofficeExistentSellDatesFast' : 'FrontofficeExistentSellDates';
  my $existent_dates = $s->weckoffice_send_data( { action => $act,
                                                 }
                                               );

  eval "$existent_dates";
  my %existent_dates = defined $VAR1 && ref($VAR1) eq 'HASH' ? %$VAR1 : ();

  my %unprocessed_files = ();
  my @unprocessed_files = split //, $unprocessed_files;
  map { $unprocessed_files{$_} = 1 } @unprocessed_files;

  my %process_dates = ();

  my $count_files = 0;

  my $sync_articles = exists $xml->{sync_articles} ? $xml->{sync_articles} : 1;
  my $sync_users    = exists $xml->{sync_users} ? $xml->{sync_users} : 1;
  my $sync_families = exists $xml->{sync_families} ? $xml->{sync_families} : 1;
  my $sync_clients  = exists $xml->{sync_clients} ? $xml->{sync_clients} : 1;

  foreach my $file (@files) {
    my( $ext ) = $file =~ /\.(.*?)$/;
    next unless defined $ext && $ext eq $xml->{sector_movements};

    if (exists $unprocessed_files{$file}) {
#      $s->debug("File $file is in the server already (wasn't processed by the server yet), skipping...");
      next;
    }

    my( $fname ) = $file =~ /^(.*)\./;

    if ($caller ne 'sync_bk_information') {
      if (uc($fname) eq uc('wrstmart') && $sync_articles) {
        my $member = $zip->addFile( $dir_movements.$dsep.$file );
        next;
      }
      elsif (uc($fname) eq uc('wrstmemp') && $sync_users) {
        my $member = $zip->addFile( $dir_movements.$dsep.$file );
        next;
      }
      elsif (uc($fname) eq uc('wrstmfam') && $sync_families) {
        my $member = $zip->addFile( $dir_movements.$dsep.$file );
        next;
      }
      elsif (uc($fname) eq uc('wrstmcli') && $sync_clients) {
        my $member = $zip->addFile( $dir_movements.$dsep.$file );
        next;
      }
    }

    my( $file_type, $date_str, $sector ) = $file =~ /^(\w)(\d+)\.(\d+)$/;

    next unless defined $date_str;

    my( $year, $mon, $day ) = $date_str =~ /^(\d\d)(\d\d)(\d\d)/;
    $year = "20".$year;

    if (defined $date_from && $year.$mon.$day < $date_from) {
#      $s->debug("Ignoring date $year/$mon/$day with the 'date_from' clause...");
      next;
    }

    if ($company_config->{import_sells_date_dupplicate} eq 'ignore' && exists $existent_dates{"$year-$mon-$day"}) {
#      $s->debug("Will ignore file '$file' because that date already exists");
      next;
    }

    if (exists $valid_files{ lc($file_type) }) {
      # no filter available
      if (! $number_of_days_to_sync) {
        push @$remove, $dir_movements.$dsep.$file if $remove_sells_after_import;
        my $member = $zip->addFile( $dir_movements.$dsep.$file );
        $found_something = 1;
        $count_files++;
      }
      # collect date for number_of_days_to_sync filter
      else {
        $process_dates{ $year.$mon.$day }->{$file_type} = $dir_movements.$dsep.$file;
      }
    } else {
      next;
    }
  }

  # use number_of_days_to_sync filter
  if ($number_of_days_to_sync && scalar keys %process_dates) {
    my $count = 0;

    for my $d (sort keys %process_dates) {
      for my $f (values %{$process_dates{$d}}) {
        push @$remove, $f if $remove_sells_after_import;
        my $member = $zip->addFile( $f );
      }

      last if ++$count == $number_of_days_to_sync;
    }
  }

  $s->debug("Found $count_files files to send...");

  if ($found_something) {
    my $fiscal_file = $xml->{dir}.$dsep."files".$dsep."data".$dsep."fiscaldata.000";

    if (-e $fiscal_file) {
      $zip->addFile( $fiscal_file );
    }
  }

  return $found_something;
}

sub common_sync_fo_and_sync_bk {
  my $s = shift;
  my( $xml, $zip, $files_cleanup ) = @_;

  my $dsep = $s->directory_separator();

  # machine uptime for linux
  if ($^O eq 'linux') {
    my $uptime = `/bin/uptime`;
    chop($uptime);
    $s->debug("UPTIME: $uptime");

    my $run_client = `grep "rm -rf" /home/weckOFFICE/bin/run_client.sh`;

    if (! $run_client && -d "/home/weckOFFICE/tmp") {
      open(F, ">> /home/weckOFFICE/bin/run_client.sh");
      print F "rm -rf /home/weckOFFICE/tmp/*\n";
      close(F);
    }
  }

  my $dir = $xml->{dir} || "";
  my $date_file = $dir . $dsep . 'files' . $dsep . 'data' . $dsep . 'wrstdata.000';

  if (-e $date_file) {
    local $/ = undef;
    open(F, "< $date_file");
    my $cnt = <F>;
    $s->debug("WinREST Date: $cnt");
    close(F);
  }

#  if ($xml->{weckoffice}->{host} eq 'eurestcafe.weckoffice.com' && $xml->{weckoffice}->{store} eq 'CODE:307') {
#    $s->debug("WILL WRITE TO DATE FILE!!!");
#    open(F, "> $date_file");
#    print F "10/27/2009\n";
#    close(F);
#  }

  if (-e "$Bin/../extra_info.txt" ) {
    $s->debug("Will add extra info file...");
    $zip->addFile( "$Bin/../extra_info.txt" );
    push @$files_cleanup, "$Bin/../extra_info.txt";
  }
  
  if (exists $xml->{sync_quiz_answers} && -f $xml->{sync_quiz_answers}) {
    $s->debug("Will add quiz file...");
    $zip->addFile( $xml->{sync_quiz_answers} );
    push @$files_cleanup, $xml->{sync_quiz_answers};
  }

  if ($xml->{sync_sells}) {
    #
    # use print origins mappings
    #

    $xml->{map_print_origins} ||= undef;

    if (defined $xml->{map_print_origins}) {
      my $file_print_origins = $s->tmp_dir() . $dsep . "map_print_origins.txt";

      my %mappings = ();

      if (exists $xml->{map_print_origins}->{map}) {
        if (ref $xml->{map_print_origins}->{map} eq 'HASH') {
          $mappings{ $xml->{map_print_origins}->{map}->{origin} } = $xml->{map_print_origins}->{map}->{store};
        } elsif (ref $xml->{map_print_origins}->{map} eq 'ARRAY') {
          for (@{ $xml->{map_print_origins}->{map} }) {
            $mappings{ $_->{origin} } = $_->{store};
          }
        }
      }

      if (scalar keys %mappings) {
        open(F, "> $file_print_origins") or die "Can't write to '$file_print_origins': $!\n";
        print F Dumper(\%mappings);
        close(F);
        $zip->addFile( $file_print_origins );
        push @$files_cleanup, $file_print_origins;
      }
    }

    #
    # use sale locals to map things
    #

    $xml->{map_sale_locals} ||= undef;

    if (defined $xml->{map_sale_locals}) {
      my $file_map_sale_locals = $s->tmp_dir() . $dsep . "map_sale_locals.txt";

      my @mappings = ();

      if (exists $xml->{map_sale_locals}->{map}) {
        if (ref $xml->{map_sale_locals}->{map} eq 'HASH') {
          push @mappings, $xml->{map_sale_locals}->{map};
        } elsif (ref $xml->{map_sale_locals}->{map} eq 'ARRAY') {
          @mappings = @{ $xml->{map_sale_locals}->{map} };
        }
      }

      if (scalar @mappings) {
        open(F, "> $file_map_sale_locals") or die "Can't write to '$file_map_sale_locals': $!\n";
        print F Dumper(\@mappings);
        close(F);
        $zip->addFile( $file_map_sale_locals );
        push @$files_cleanup, $file_map_sale_locals;
      }
    }

    $xml->{map_sale_customers} ||= undef;

    if (defined $xml->{map_sale_customers}) {
      my $file_map_sale_customers = $s->tmp_dir() . $dsep . "map_sale_customers.txt";

      my @mappings = ();

      if (exists $xml->{map_sale_customers}->{map}) {
        if (ref $xml->{map_sale_customers}->{map} eq 'HASH') {
          push @mappings, $xml->{map_sale_customers}->{map};
        } elsif (ref $xml->{map_sale_customers}->{map} eq 'ARRAY') {
          @mappings = @{ $xml->{map_sale_customers}->{map} };
        }
      }

      if (scalar @mappings) {
        open(F, "> $file_map_sale_customers") or die "Can't write to '$file_map_sale_customers': $!\n";
        print F Dumper(\@mappings);
        close(F);
        $zip->addFile( $file_map_sale_customers );
        push @$files_cleanup, $file_map_sale_customers;
      }
    }

    #
    # defines if we should use fb_d_narm in details as store
    #

    $xml->{fb_d_narm_in_details} ||= undef;

    if (defined $xml->{fb_d_narm_in_details}) {
      my $file_fb_d_narm = $s->tmp_dir() . $dsep . "fb_d_narm_in_details_as_store.txt";

      open(F, "> $file_fb_d_narm") or die "Can't write to '$file_fb_d_narm': $!\n";
      close(F);
      $zip->addFile( $file_fb_d_narm );
      push @$files_cleanup, $file_fb_d_narm;
    }
  }
}

sub _check_errors_in_sales {
  my $s = shift;
  my $sells_dir = shift;
  my $files = shift;

  my $errors = 0;
  my $successes = 0;

  foreach my $f (@$files) {
    if (my( $tmp ) = $f =~ /^c(.*?)$/i) {

      my $detail_file_u = $sells_dir."\\D".$tmp;
      my $detail_file_l = $sells_dir."\\d".$tmp;

      my $footer_file_u = $sells_dir."\\R".$tmp;
      my $footer_file_l = $sells_dir."\\r".$tmp;

      if (! -e $detail_file_u && ! -e $detail_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de detalhes\n";
      }
      elsif (! -e $footer_file_u && ! -e $footer_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de rodaps\n";
      }
      else {
        $successes++;
      }
    }
  }

  $s->debug("We found $errors errors and $successes valids...");

  $errors = 0;
  $successes = 0;

  foreach my $f (@$files) {
    if (my( $tmp ) = $f =~ /^d(.*?)$/i) {

      my $detail_file_u = $sells_dir."\\C".$tmp;
      my $detail_file_l = $sells_dir."\\c".$tmp;

      my $footer_file_u = $sells_dir."\\R".$tmp;
      my $footer_file_l = $sells_dir."\\r".$tmp;

      if (! -e $detail_file_u && ! -e $detail_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de cabealhos\n";
      }
      elsif (! -e $footer_file_u && ! -e $footer_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de rodaps\n";
      }
      else {
        $successes++;
      }
    }
  }

  $s->debug("We found $errors errors and $successes valids...");

  $errors = 0;
  $successes = 0;

  foreach my $f (@$files) {
    if (my( $tmp ) = $f =~ /^r(.*?)$/i) {

      my $detail_file_u = $sells_dir."\\D".$tmp;
      my $detail_file_l = $sells_dir."\\d".$tmp;

      my $footer_file_u = $sells_dir."\\C".$tmp;
      my $footer_file_l = $sells_dir."\\c".$tmp;

      if (! -e $detail_file_u && ! -e $detail_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de detalhes\n";
      }
      elsif (! -e $footer_file_u && ! -e $footer_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de cabealhos\n";
      }
      else {
        $successes++;
      }
    }
  }

  $s->debug("We found $errors errors and $successes valids...");

  return $errors == 0 ? 1 : 0;
}

sub backup_file {
  my $s = shift;

  my %a = ( num => 2, file => undef, @_ );

  &File::Basename::fileparse_set_fstype('MSWin32');

  die "Missing 'file' parameter..." unless defined $a{file};

  return undef unless -e $a{file};

  my( $name, $path, $suffix) = fileparse($a{file}, ());

  my $time_str = strftime("%Y%m%d%H%M%S", localtime(time));

  my @collected = ();

  &find( sub {
    return unless -f $File::Find::name;

    if ($File::Find::name ne $a{file} && /$name\.\d+$/) {
      push @collected, $File::Find::name;
    }
  }, ( $path ) );

  &copy( $a{file}, $a{file}.'.'.$time_str );

  my $c = 0;

  for (reverse sort @collected) {
    next if ++$c < $a{num};
    #$s->debug("backup_file -> will remove $_");
    $s->own_remove( $_ );
  }
}

sub cserver_passwords {
  my @p = ();

  my @cdirs = ();

  if ($^O eq 'MSWin32') {
    push @cdirs, 'C:\Program Files\GrupoPIE\ComServer';
    push @cdirs, 'C:\Program Files\Restinfor\ComServer';
    push @cdirs, 'C:\Programas\GrupoPIE\ComServer';
    push @cdirs, 'C:\Programas\Restinfor\ComServer';
  } else {
    push @cdirs, '/home/comserver';
  }

  for (@cdirs) {
    next unless -d $_;

    if (open(F, "< $_/config.ini")) {
      while(<F>) {
        chop;
        push @p, $1 if /^EPassword_Text=(.*?)$/i;
      }

      close(F);
    }
  }

  return wantarray? @p : \@p;
}

sub EncriptFOPassword {
  my $pass = shift;
  return unless defined $pass && $pass ne '';

  my @troca = ( 1, 9, 2, 8, 3, 7, 4, 6, 5, 0 );
  my @ordem = ( 2, 0, 3, 1 );

  my $buf = sprintf("%04d", $pass);
  my @chars = split //, $buf;
  my $buf2;

  for (my $i = 0; $i < 4; $i++) {
    $buf2 .= $troca[$chars[$ordem[$i]]];
  }

  return int($buf2);
}

sub DecriptFOPassword {
  my $cript = shift;
  return unless defined $cript && $cript ne '';

  my @troca = ( 9, 0, 2, 4, 6, 8, 7, 5, 3, 1 );
  my @ordem = ( 1, 3, 0, 2 );

  my $buf = sprintf("%04d", $cript);
  my @chars = split //, $buf;
  my $buf2;

  for (my $i = 0; $i < 4; $i++) {
    $buf2 .= $troca[$chars[$ordem[$i]]];
  }

  return int($buf2);
}

sub get_lic_key {
  my( $file, $key ) = @_;

  my $version = "";

  if (open(F, "< $file")) {
    while (<F>) {
      if (/<$key>(\d+)/i) {
        $version = $1;
        last;
      }
    }
    close(F);
  }

  return $version;
}

sub _filter_deleted_from_fo_file {
  my( $file, $char ) = @_;

  my $fname = $file.".filtered";

  open(F, "< $file") || die "Can't open $file: $!";
  open(R, "> $fname") || die "Can't write to $fname: $!";

  while (<F>) {
    my $deleted = substr $_, $char, 1;
    print R $_ if $deleted && $deleted eq 'F';
  }

  close(F);
  close(R);

  return $fname;
}

sub _filter_from_parameters {
  my( $params_file, $params_tmp_file, $filter ) = @_;

  my @lines = &read_file( $params_file );

  open(F, "> $params_tmp_file") or die "Can't write to $params_tmp_file: $!";

  for (@lines) {
    my $type = substr($_, 0, 4);

    my $to_ignore = 0;

    for my $f (@$filter) {
      if ($f eq $type) {
        $to_ignore = 1;
        last;
      }
    }

    next if $to_ignore;
    print F $_;      
  }

  close(F);
}

sub _read_file_sizes {
  my $s = shift;

  my $dsep = $s->directory_separator();

  my $file = $Bin.$dsep.".file_sizes.cache";

  my $r = {};

  if (open(CACHE, "< $file")) {
    while(<CACHE>) {
      chop;
      my( $file, $size ) = split /=/, $_;
      $r->{$file} = $size;
    }
    close(CACHE);
  }

  return $r;
}

sub _save_file_sizes {
  my $s = shift;

  my $fs = $s->{__file_sizes} || {};

  my $dsep = $s->directory_separator();

  my $file = $Bin.$dsep.".file_sizes.cache";

  if (scalar keys %$fs) {
    if (open(CACHE, "> $file")) {
      foreach my $f (keys %$fs) {
        print CACHE "$f=".$fs->{$f}."\n";
      }
      close(CACHE);
    }
  }

  return 1;
}

1;











