package WO::Client::Integrators::UQ::WinRest;

use strict;
use warnings;

use base qw ( WO::Client::Common );

use File::Remove qw( remove );
use File::Copy qw( move copy );
use File::Slurp qw( read_file read_dir );
use Archive::Zip qw( :ERROR_CODES :CONSTANTS );
use File::Basename;
use File::Find qw( find );
use File::Spec;
use POSIX qw( strftime );

use WinRest::File::Articles;
use WinRest::File::Families;
use WinRest::File::Waitress;
use WinRest::File::Clients;
use WinRest::File::Parameters;

sub tmp_dir { my $s = shift; return $s->tmp_dir() }

#
# FO to weckOFFICE
#

sub sync_fo_information {
  my $s = shift;

  $s->debug("sync_fo_information()");

  my $xml = $s->info();

  my $fo_type = $s->encode_fo_type();
  $s->debug("Encoded FO type is: $fo_type");

  # frontoffice directory
  my $dir = $xml->{dir} || "";

  # winrest sells dir
  my $sells_dir = $xml->{sells_dir} || "export";

  my $date_from = $xml->{date_from} || undef;
  $date_from =~ s/\///g if defined $date_from;

  # remove sells after import
  my $remove_sells_after_import = $xml->{remove_sells_after_import} || 0;
  my @possible_remove = ();

  if (! -d $dir) {
    die "Frontoffice dir '$dir' don't exists, where is the frontoffice?";
  }

  # directory separator
  my $dsep = $s->directory_separator();

  my $file_warehouses = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmarm.'.$xml->{sector_master};
  my $file_articles = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmart.'.$xml->{sector_master};
  my $file_families = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmfam.'.$xml->{sector_master};
  my $file_waitress = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmemp.'.$xml->{sector_master};
  my $file_params = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrsttprm.'.$xml->{sector_master};
  my $file_clients = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmcli.'.$xml->{sector_master};

  my $dir_movements = $xml->{dir_movements} || "";
  $dir_movements = $dir.$dsep.'files'.$dsep.$sells_dir unless $dir_movements;
  $dir_movements =~ s/\//\\/g if $^O eq 'MSWin32';

  #
  # Prepare ZIP file
  #

  my $zip = Archive::Zip->new();

  my $zipFile = $s->tmp_dir().$dsep.$fo_type.'_FO_DATA.zip';
  &remove( $zipFile ) if -e $zipFile;

  #
  # Get unprocessed files found in the server
  #

  my $unprocessed_files = $s->weckoffice_send_data( { action => 'FrontofficeUnprocessedFiles',
                                                    }
                                                  );

  my $VAR1 = {};

  my %unprocessed_files = ();
  my @unprocessed_files = split //, $unprocessed_files;
  map { $unprocessed_files{$_} = 1 } @unprocessed_files;

  # default values are 1
  $xml->{sync_users} = exists $xml->{sync_users} ? $xml->{sync_users} : 1;
  $xml->{sync_articles} = exists $xml->{sync_articles} ? $xml->{sync_articles} : 1;
  $xml->{sync_families} = exists $xml->{sync_families} ? $xml->{sync_families} : 1;
  $xml->{sync_warehouses} = exists $xml->{sync_warehouses} ? $xml->{sync_warehouses} : 1;
  $xml->{sync_clients} = exists $xml->{sync_clients} ? $xml->{sync_clients} : 1;
  $xml->{sync_sells} = exists $xml->{sync_sells} ? $xml->{sync_sells} : 1;

  # default value are 0
  $xml->{check_sell_files_errors} ||= 0;
	
  if ($xml->{sync_warehouses}) {
    #
    # Warehouses
    #
    if (-e $file_warehouses) {
      $zip->addFile( $file_warehouses ) unless exists $unprocessed_files{ 'wrstmarm.'.$xml->{sector_master} };    
    } else {
      $s->debug("Didn't find the warehouses '$file_warehouses' file...");
    }
  }

  #
  # Parameters
  #

  if (-e $file_params) {
    $zip->addFile( $file_params ) unless exists $unprocessed_files{ 'wrsttprm.'.$xml->{sector_master} };
  } else {
    $s->debug("Didn't find the parameters '$file_params' file...");
  }

  if ($xml->{sync_articles}) {
    #
    # Articles
    #

    if (-e $file_articles) {
      $zip->addFile( $file_articles ) unless exists $unprocessed_files{ 'wrstmart.'.$xml->{sector_master} };    
    } else {
      $s->debug("Didn't find the articles '$file_articles' file...");
    }
  }

  if ($xml->{sync_families}) {
    #
    # Families
    #

    if (-e $file_families) {
      $zip->addFile( $file_families ) unless exists $unprocessed_files{ 'wrstmfam.'.$xml->{sector_master} };
    } else {
      $s->debug("Didn't find the families '$file_families' file...");
    }
  }
  
  if ($xml->{sync_users}) {
    #
    # Waitresses
    #

    if (-e $file_waitress) {
      $zip->addFile( $file_waitress ) unless exists $unprocessed_files{ 'wrstmemp.'.$xml->{sector_master} };
    } else {
      $s->debug("Didn't find the waitress '$file_waitress' file...");
    }
  }

  if ($xml->{sync_clients}) {
    #
    # Waitresses
    #

    if (-e $file_clients) {
      $zip->addFile( $file_clients ) unless exists $unprocessed_files{ 'wrstmcli.'.$xml->{sector_master} };
    } else {
      $s->debug("Didn't find the clients '$file_clients' file...");
    }
  }

  #
  # Movements
  #

  if ($xml->{sync_sells}) {
    if (-d $dir_movements) {
      $s->sync_movements($zip, \@possible_remove, $date_from);
    } else {
      die "Can't find movements dir: ".$xml->{sync_sells};
    }
  }

  die "write error" unless $zip->writeToFileNamed( $zipFile ) == AZ_OK;

  #
  # Send ZIP file to the server
  #

  eval {
    $s->weckoffice_send_data( { action => 'FrontofficeReceiveData',
                                file_upload => $zipFile,
                              }
                            );
  };

  if ($@) {
    die "Error sending file: $@";
  } else {
    # remove sales after import, if requested
    if ($remove_sells_after_import && scalar @possible_remove) {
      map { &remove($_) } @possible_remove;
    }
  }

  $s->debug("Done.");
}

#
# Brings weckOFFICE information to FO
#

sub sync_bk_information {
  my $s = shift;

  $s->debug("sync_bk_information");

  # service xml
  my $xml = $s->info();

  my $fo_type = $s->encode_fo_type();

  my $dir = $xml->{dir};

  my $dsep = $s->directory_separator();

  my $parameters_file = $dir.$dsep."files".$dsep."data".$dsep."wrsttprm.".$xml->{sector_master};

  my %tax_codes = ();

  if (-e $parameters_file) {
    my $params = WinRest::File::Parameters->new( file => $parameters_file );
    my @params = $params->get_all();

    foreach my $p (@params) {
      $tax_codes{ $p->param_value() } = $p->code() if $p->param_type() eq 'TIVA';
    }
  } else {
    $s->debug("Can't find WinREST parameters file...");
  }

  my $articles_file   = $dir.$dsep."files".$dsep."data".$dsep."wrstmart".$xml->{sector_master};
  my $families_file   = $dir.$dsep."files".$dsep."data".$dsep."wrstmfam".$xml->{sector_master};
  my $waitresses_file = $dir.$dsep."files".$dsep."data".$dsep."wrstmemp".$xml->{sector_master};
  my $clients_file    = $dir.$dsep."files".$dsep."data".$dsep."wrstmcli".$xml->{sector_master};

  # default values are 1
  $xml->{sync_users} = exists $xml->{sync_users} ? $xml->{sync_users} : 1;
  $xml->{sync_articles} = exists $xml->{sync_articles} ? $xml->{sync_articles} : 1;
  $xml->{sync_families} = exists $xml->{sync_families} ? $xml->{sync_families} : 1;
  $xml->{sync_clients} = exists $xml->{sync_clients} ? $xml->{sync_clients} : 1;
  $xml->{sync_sells} = exists $xml->{sync_sells} ? $xml->{sync_sells} : 1;
  $xml->{transfer_remote_files} = exists $xml->{transfer_remote_files} ? $xml->{transfer_remote_files} : 0;

  my $str;
  my $VAR1 = undef;

  #
  # Sync Families
  #

  if ($xml->{sync_families}) {
    if (! -f $families_file) {
      #die("Can't find families file '$families_file'");
    }

    eval {
      $str = $s->weckoffice_send_data( { action => 'BackofficeGetFamilies',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      # backup original
      &backup_file( file => $families_file, days_to_keep => 10 );

      $s->debug("Will process ".scalar(@$VAR1)." families...");

      my @fam_fields_t = WinRest::File::Families::fields();
      my @fam_defaults = WinRest::File::Families::defaults();

      my @fam_fields = ();
      map { my $key = (keys %$_)[0]; push @fam_fields, $key } @fam_fields_t;

      my %fam_defaults = ();
      map { my $key = (keys %$_)[0]; $fam_defaults{$key} = $_->{$key} } @fam_defaults;

      foreach my $f (@$VAR1) {
        my $code = defined $f->{family_fo_id} ? $f->{family_fo_id} : ($f->{fo_seetings}->{mfam_codg} || "");
        my $name = defined $f->{family_name} ? $f->{family_name} : ($f->{fo_seetings}->{mfam_nome} || "");

        my $nbot = defined $f->{fo_seetings}->{_nbot} && $f->{fo_seetings}->{_nbot} ne 'S/Nome' ? $f->{fo_seetings}->{_nbot} : $f->{family_name};
        my $ncur = defined $f->{fo_seetings}->{_ncur} && $f->{fo_seetings}->{_ncur} ne 'S/Nome' ? $f->{fo_seetings}->{_ncur} : $f->{family_name};

        my %family_main_props = (
          mfam_codg => $code,
          mfam_nome => $name,
          mfam_nbot => $nbot,
          mfam_ncur => $ncur,
          mfam_cbot => $f->{fo_seetings}->{_cbot} ? $f->{fo_seetings}->{_cbot} : "",
        );

        my %all_props = %family_main_props;

        foreach my $field (@fam_fields) {
          next if exists $family_main_props{$field};
          $all_props{$field} = defined $f->{fo_seetings}->{$field} ? $f->{fo_seetings}->{$field} : $fam_defaults{$field};
        }

        WinRest::File::Families->create(
          file => $families_file,
          fields => \%all_props,
        );   
      }
    } else {
      die("Couldn't get families from weckOFFICE server");
    }
  }

  $VAR1 = undef;

  #
  # Sync articles
  #

  if ($xml->{sync_articles}) {
    if (! -f $articles_file) {
      die("Can't find articles file '$articles_file'");
    }

    eval {
      $str = $s->weckoffice_send_data( { action => 'BackofficeGetArticles',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      # backup original
      &backup_file( file => $articles_file, days_to_keep => 10 );

      $s->debug("Will process ".scalar(@$VAR1)." articles...");

      my $cped = 0;

      my @art_fields_t = WinRest::File::Articles::fields();
      my @art_defaults = WinRest::File::Articles::defaults();

      my @art_fields = ();
      map { my $key = (keys %$_)[0]; push @art_fields, $key } @art_fields_t;

      my %art_defaults = ();
      map { my $key = (keys %$_)[0]; $art_defaults{$key} = $_->{$key} } @art_defaults;
      
      foreach my $a (@$VAR1) {


        my $price1 = defined $a->{article_price1} && $a->{article_price1} ne '0.00' ? $a->{article_price1} : ($a->{fo_seetings}->{mart_1_pv} || '0.00');
        my $price2 = defined $a->{article_price2} && $a->{article_price2} ne '0.00' ? $a->{article_price2} : ($a->{fo_seetings}->{mart_2_pv} || '0.00');
        my $price3 = defined $a->{article_price3} && $a->{article_price3} ne '0.00' ? $a->{article_price3} : ($a->{fo_seetings}->{mart_3_pv} || '0.00');
        my $price4 = defined $a->{article_price4} && $a->{article_price4} ne '0.00' ? $a->{article_price4} : ($a->{fo_seetings}->{mart_4_pv} || '0.00');
        my $price5 = defined $a->{article_price5} && $a->{article_price5} ne '0.00' ? $a->{article_price5} : ($a->{fo_seetings}->{mart_5_pv} || '0.00');

        my $tax1 = 0;
        my $tax2 = 0;

        if (defined $a->{article_tax0} && $a->{article_tax0}) {
          $a->{article_tax0} = sprintf("%.2f", $a->{article_tax0});
          $tax1 = exists $tax_codes{ $a->{article_tax0} } ? $tax_codes{ $a->{article_tax0} } : '0';
        } else {
          $tax1 = $a->{fo_seetings}->{mart_iva1} || '0';
        }

        if (defined $a->{article_tax1} && $a->{article_tax1}) {
          $a->{article_tax1} = sprintf("%.2f", $a->{article_tax1});
          $tax2 = exists $tax_codes{ $a->{article_tax1} } ? $tax_codes{ $a->{article_tax1} } : '0';
        } else {
          $tax2 = $a->{fo_seetings}->{mart_iva2} || '0';
        }

        my $name = defined $a->{article_name} ? $a->{article_name} : ($a->{fo_seetings}->{mart_nome} || "");

        my $family = defined $a->{family_fo_id} ? $a->{family_fo_id} : ($a->{fo_seetings}->{mart_fami} || "");

        $cped++;
        my $cped = $a->{fo_seetings}->{mart_cped} ? $a->{fo_seetings}->{mart_cped} : $cped;

        my $code = defined $a->{article_fo_id} ? $a->{article_fo_id} : ($a->{fo_seetings}->{mart_codg} || "");

        my $cbot = $a->{fo_seetings}->{_cbot} ? $a->{fo_seetings}->{_cbot} : "";
        my $ncur = $a->{fo_seetings}->{_ncur} ? $a->{fo_seetings}->{_ncur} : ((length($name) > 13) ? substr($name,0,13) : $name);
        my $nbot = $a->{fo_seetings}->{_nbot} ? $a->{fo_seetings}->{_nbot} : ((length($name) > 32) ? substr($name,0,32) : $name);
        my $vord = $a->{fo_seetings}->{_vord} ? $a->{fo_seetings}->{_vord} : 0;

        my $barcode = defined $a->{article_barcode} ? $a->{article_barcode} : ($a->{fo_seetings}->{mart_cbar} || "");

        my %article_main_props = (
          mart_codg => $code,
          mart_cbot => $cbot,
          mart_cped => $cped,
          mart_fami => $family,
          mart_nome => $name,
          mart_ncur => $ncur,
          mart_nbot => $nbot,
          mart_1_pv => $price1,
          mart_2_pv => $price2,
          mart_3_pv => $price3,
          mart_4_pv => $price4,
          mart_5_pv => $price5,
          mart_iva1 => $tax1,
          mart_iva2 => $tax2,
          mart_cbar => $barcode,
          mart_vord => $vord,
        );

        my %all_props = %article_main_props;

        foreach my $field (@art_fields) {
          next if exists $article_main_props{$field};
          $all_props{$field} = $a->{fo_seetings}->{$field} ? $a->{fo_seetings}->{$field} : $art_defaults{$field};
        }

        WinRest::File::Articles->create(
          file => $articles_file,
          fields => \%all_props,
        );   
      }
    } else {
      die("Couldn't get articles from weckOFFICE server");
    }
  }

  $VAR1 = undef;

  #
  # Sync Waitresses
  #

  if ($xml->{sync_users}) {
    if (! -f $waitresses_file) {
      #die("Can't find waitresses file '$waitresses_file'");
    }

    eval {
      $str = $s->weckoffice_send_data( { action => 'BackofficeGetWaitresses',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      # backup original
      &backup_file( file => $waitresses_file, days_to_keep => 10 );

      $s->debug("Will process ".scalar(@$VAR1)." waitresses...");

      my @wai_fields_t = WinRest::File::Waitress::fields();
      my @wai_defaults = WinRest::File::Waitress::defaults();

      my @wai_fields = ();
      map { my $key = (keys %$_)[0]; push @wai_fields, $key } @wai_fields_t;

      my %wai_defaults = ();
      map { my $key = (keys %$_)[0]; $wai_defaults{$key} = $_->{$key} } @wai_defaults;

      my $cur_codg = 1;

      foreach my $w (@$VAR1) {
        next unless $w->{fo_id} =~ /^\d+$/;

        my %waitress_main_props = (
          memp_ctrb => $w->{nif} || "",
          memp_dadm => $w->{dadmission} || "",
          memp_mora => $w->{address} || "",
          memp_iden => $w->{bi} || "",
          memp_loca => $w->{localidade} || "",
          memp_telf => $w->{telephone} || "",
          memp_nome => $w->{name} || "",
          memp_nint => $w->{fo_login} || "",    # fo_login
          memp_iden => $w->{fo_login} || "",
          memp_pasw => $w->{fo_passwd} || "",  # fo_passwd
          memp_codg => $w->{fo_id},
        );

        # initialize all the unixistant fields
        for (@wai_fields) {
          if (! exists $waitress_main_props{$_}) {
            if (exists $wai_defaults{$_} && defined $wai_defaults{$_}) {
              $waitress_main_props{$_} = $wai_defaults{$_};
            } else {
              $waitress_main_props{$_} = "";
            }
          }
        }

        # if the code is null, create a temp one
        if ($waitress_main_props{memp_codg}) {
          $cur_codg = $waitress_main_props{memp_codg} + 1;
        } else {
          $waitress_main_props{memp_codg} = $cur_codg++;
        }

        WinRest::File::Waitress->create(
          file => $waitresses_file,
          fields => \%waitress_main_props,
        );
      }
    } else {
      die("Couldn't get waitresses from weckOFFICE server");
    }
  }

  $VAR1 = undef;

  #
  # Sync Clients
  #

  if ($xml->{sync_clients}) {
    if (! -f $clients_file) {
      die("Can't find clients file '$clients_file'");
    }

    eval {
      $str = $s->weckoffice_send_data( { action => 'BackofficeGetStoreClients',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      # backup original
      &backup_file( file => $clients_file, days_to_keep => 10 );
      
      $s->debug("Will process ".scalar(@$VAR1)." clients...");

      my @cli_fields_t = WinRest::File::Clients::fields();
      my @cli_defaults = WinRest::File::Clients::defaults();

      my @cli_fields = ();
      map { my $key = (keys %$_)[0]; push @cli_fields, $key } @cli_fields_t;

      my %cli_defaults = ();
      map { my $key = (keys %$_)[0]; $cli_defaults{$key} = $_->{$key} } @cli_defaults;

      my $cur_codg = 1;

      foreach my $w (@$VAR1) {
        my %client_main_props = (
          mcli_codg => $w->{fo_id} || "",
          mcli_nome => $w->{name} || "",
          mcli_telf => $w->{telephone} || "",
          mcli_fax => $w->{fax} || "",
          mcli_loca => $w->{localidade} || "",
          mcli_mora => $w->{address} || "",
          mcli_cpos => $w->{postal_code} || "",
          mcli_ctrb => $w->{ncontrib} || "",
          mcli_dnas => $w->{birth_date} || "",
          mcli_ncrd => $w->{card_number} || "",
          mcli_dcrd => $w->{card_expiration} || "",
          mcli_tcre => $w->{total_credit} || "",
          mcli_tdeb => $w->{total_debt} || "",
          mcli_limi => $w->{credit_limit} || "",
          mcli_sald => $w->{balance} || "",
          mcli_anul => $w->{deleted} || "",
          mcli_obsv => $w->{obs} || "",
        );

        # initialize all the unixistant fields
        for (@cli_fields) {
          if (! exists $client_main_props{$_}) {
            if (exists $cli_defaults{$_} && defined $cli_defaults{$_}) {
              $client_main_props{$_} = $cli_defaults{$_};
            } else {
              $client_main_props{$_} = "";
            }
          }
        }

        # if the code is null, create a temp one
        if ($client_main_props{mcli_codg}) {
          $cur_codg = $client_main_props{mcli_codg} + 1;
        } else {
          $client_main_props{mcli_codg} = $cur_codg++;
        }

        WinRest::File::Clients->create(
          file => $clients_file,
          fields => \%client_main_props,
        );
      }
    } else {
      die("Couldn't get clients from weckOFFICE server");
    }
  }


  #
  # Remote files
  #

  if ($xml->{transfer_remote_files}) {
    $s->debug("Will check 'transfer_remote_files'...");

    eval {
      $str = $s->weckoffice_send_data( { action => 'BackofficeAreFilesAvailable',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    my $upload_state = 0;

    if (defined $VAR1 && ref($VAR1) eq 'HASH') {
      $upload_state = $VAR1->{upload_state};
    }

    #
    # need to receive files
    #

    $s->debug("Store upload_state = $upload_state...");

    if ($upload_state == 2) {
      $s->debug("Will BackofficeGetAvailableFiles...");

      eval {
        $str = $s->weckoffice_send_data( { action => 'BackofficeGetAvailableFiles',
                                         }
                                       );
      };

      if ($@) {
        die("Error: $@");
      }

      eval $str;

      if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
        $s->debug("Will get ".scalar(@$VAR1)." remote files...");

        my @files_to_cleanup = ();

        foreach my $rfile (@$VAR1) {
          $s->debug("Will save '".$rfile->{file_name}."' in '".$rfile->{remote_path}."'");

          eval {
            my $f = $rfile->{remote_path}."/".$rfile->{file_name};

            my $file_content = $s->unshrink($rfile->{file_content});

            # save a backup of original file
            if (-e $f) {
              #my $date_str = strftime("%Y%m%d_%H%M%S", localtime(time));
              my $date_str = 'backup';
              &move( $f, $f.".".$date_str );
            }

            open(F, "> $f") or die "Can't write to '$f': $!\n";
            print F $$file_content;
            close(F);
          };

          if ($@) {
            $s->debug("Error writing file: $@");
          } else {
            $s->debug("Success...");
            push @files_to_cleanup, $rfile->{file_name};
          }
        }

        $s->weckoffice_send_data( { action => 'BackofficeMarkStoreFilesReceived',
                                    cleanup_files => join('', @files_to_cleanup),
                                  }
                                );
      }
    }
  }




  #
  # Movements
  #

  # winrest sells dir
  my $sells_dir = $xml->{sells_dir} || "export";

  # movements dir (overrided by config file)
  my $dir_movements = $xml->{dir_movements} || "";
  $dir_movements = $dir.$dsep.'files'.$dsep.$sells_dir unless $dir_movements;
  $dir_movements =~ s/\//\\/g if $^O eq 'MSWin32';

  if ($xml->{sync_sells}) {
    if (! -d $dir_movements) {
      die "Can't find movements dir: $dir_movements";
    }

    # remove sells after import
    my $remove_sells_after_import = $xml->{remove_sells_after_import} || 0;

    my $zip = Archive::Zip->new();

    my $zipFile = $s->tmp_dir().$dsep.$fo_type.'_FO_DATA.zip';
    &remove( $zipFile ) if -e $zipFile;

    my @possible_remove = ();

    $s->sync_movements($zip, \@possible_remove);

    die('write error') unless $zip->writeToFileNamed( $zipFile ) == AZ_OK;

    #
    # Send ZIP file to the server
    #

    eval {
      $s->weckoffice_send_data( { action => 'FrontofficeReceiveData',
                                  file_upload => $zipFile,
                                }
                              );
    };

    if ($@) {
      die("Error sending file: $@");
    } else {
      # remove sales after import, if requested
      if ($remove_sells_after_import && scalar @possible_remove) {
        map { &remove($_) } @possible_remove;
      }

      &remove($zipFile);
    }

  
  }

  $s->debug("Done.");
}

#
# Sync movements
#

sub sync_movements {
  my $s = shift;
  my $zip = shift;
  my $remove = shift;
  my $date_from = shift;

  # service xml
  my $xml = $s->info();

  my $check_dates_fast = $xml->{check_dates_fast} || 0;

  my $fo_type = $s->encode_fo_type();

  # frontoffice directory
  my $dir = $xml->{dir} || "";

  # winrest sells dir
  my $sells_dir = $xml->{sells_dir} || "export";

  # directory separator
  my $dsep = $s->directory_separator();

  # movements dir
  my $dir_movements = $xml->{dir_movements} || "";
  $dir_movements = $dir.$dsep.'files'.$dsep.$sells_dir unless $dir_movements;
  $dir_movements =~ s/\//\\/g if $^O eq 'MSWin32';

  # remove sells after import
  my $remove_sells_after_import = $xml->{remove_sells_after_import} || 0;

  my $number_of_days_to_sync = $xml->{number_of_days_to_sync} || 0;

  my $VAR1 = {};

  my $company_config = $s->weckoffice_send_data( { action => 'CompanyConfiguration',
                                                 }
                                               );

  eval "$company_config";
  my %company_config = defined $VAR1 && ref($VAR1) eq 'HASH' ? %$VAR1 : ();

  my $act = $check_dates_fast ? 'FrontofficeExistentSellDatesFast' : 'FrontofficeExistentSellDates';
  my $existent_dates = $s->weckoffice_send_data( { action => $act,
                                                 }
                                               );

  eval "$existent_dates";
  my %existent_dates = defined $VAR1 && ref($VAR1) eq 'HASH' ? %$VAR1 : ();


=pod
  my %existent_dates = (
          'DATES_AS400' => {
                             '20060731' => {
                                             'FRONT' => '',
                                             'FECHOFIN' => '0'
                                           },
                             '20060929' => {
                                             'FRONT' => '0',
                                             'FECHOFIN' => '0'
                                           },
                             '20060730' => {
                                             'FRONT' => '0',
                                             'FECHOFIN' => '0'
                                           },
                             '20060801' => {
                                             'FRONT' => '',
                                             'FECHOFIN' => '0'
                                           },
                             '20060928' => {
                                             'FRONT' => '0',
                                             'FECHOFIN' => '0'
                                           },
                             '20060915' => {
                                             'FRONT' => '0',
                                             'FECHOFIN' => '0'
                                           }
                           },
          'FILES_IN_FTP' => {
                              'fechofin' => 1,
                              'vndsct' => 1,
                              'd0610160' => 1,
                              'd0610170' => 1,
                              'c0610150' => 1,
                              'c0610170' => 1,
                              'c0610160' => 1,
                              'd0610150' => 1
                            }
        );
=cut

  #$s->dump(\%existent_dates);

  $s->debug("##### will find files in #$dir_movements#");
  my @files = &read_dir($dir_movements);
  #$s->dump(\@files);
  $s->debug("files to process: ".scalar(@files));

  my %supported_files = (
    vndsct   => 1,
    fechofin => 1,
    f        => 1,
    c        => 1,
    d        => 1,
    o        => 1,
  );

  my @files_to_send = ();

  foreach my $file (@files) {
    my $file_orig = $file;

    $file = lc($file);

    my( $first_char, $name_rest ) = $file =~ /^(.)(.*?)\./;
    next unless defined $first_char && defined $name_rest;

    next unless exists $supported_files{ $first_char.$name_rest } || exists $supported_files{ $first_char };

    my $lc_name_wo_ext = $first_char.$name_rest;

    if ($lc_name_wo_ext eq 'vndsct' && ! exists $existent_dates{FILES_IN_FTP}->{vndsct}) {
      push @files_to_send, $file_orig;
    }
    elsif ($lc_name_wo_ext eq 'fechofin' && ! exists $existent_dates{FILES_IN_FTP}->{fechofin}) {
      push @files_to_send, $file_orig;
    }
    elsif (my( $date) = $name_rest =~ /^(\d+).$/) {
      my $as400_date = "20".$date;

      # file is older than AS400 "check limit". default is 3 months
      next if exists $existent_dates{MIN_DATE} && $as400_date < $existent_dates{MIN_DATE};

      # file is in the FTP already, ignore it
      next if exists $existent_dates{FILES_IN_FTP}->{ $lc_name_wo_ext };

      # date isn't in as400, let's add file
      if (! exists $existent_dates{DATES_AS400}->{ $as400_date }) {
        push @files_to_send, $file_orig;
      } else {
        my $v = exists $existent_dates{DATES_AS400}->{ $as400_date }->{FRONT} ?
                $existent_dates{DATES_AS400}->{ $as400_date }->{FRONT} :
                "";

        # date is in as400 but isn't processed yet
        if ($v ne '0') {
          push @files_to_send, $file_orig;
        }
      }

      # none of the conditions matched, let's move to the next iteration
    }
  }

  $s->dump(\@files_to_send);

  for my $file (@files_to_send) {
    my $member = $zip->addFile( $dir_movements.$dsep.$file );
  }
}

sub _check_errors_in_sales {
  my $s = shift;
  my $sells_dir = shift;
  my $files = shift;

  my $errors = 0;
  my $successes = 0;

  foreach my $f (@$files) {
    if (my( $tmp ) = $f =~ /^c(.*?)$/i) {

      my $detail_file_u = $sells_dir."\\D".$tmp;
      my $detail_file_l = $sells_dir."\\d".$tmp;

      my $footer_file_u = $sells_dir."\\R".$tmp;
      my $footer_file_l = $sells_dir."\\r".$tmp;

      if (! -e $detail_file_u && ! -e $detail_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de detalhes\n";
      }
      elsif (! -e $footer_file_u && ! -e $footer_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de rodaps\n";
      }
      else {
        $successes++;
      }
    }
  }

  $s->debug("We found $errors errors and $successes valids...");

  $errors = 0;
  $successes = 0;

  foreach my $f (@$files) {
    if (my( $tmp ) = $f =~ /^d(.*?)$/i) {

      my $detail_file_u = $sells_dir."\\C".$tmp;
      my $detail_file_l = $sells_dir."\\c".$tmp;

      my $footer_file_u = $sells_dir."\\R".$tmp;
      my $footer_file_l = $sells_dir."\\r".$tmp;

      if (! -e $detail_file_u && ! -e $detail_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de cabealhos\n";
      }
      elsif (! -e $footer_file_u && ! -e $footer_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de rodaps\n";
      }
      else {
        $successes++;
      }
    }
  }

  $s->debug("We found $errors errors and $successes valids...");

  $errors = 0;
  $successes = 0;

  foreach my $f (@$files) {
    if (my( $tmp ) = $f =~ /^r(.*?)$/i) {

      my $detail_file_u = $sells_dir."\\D".$tmp;
      my $detail_file_l = $sells_dir."\\d".$tmp;

      my $footer_file_u = $sells_dir."\\C".$tmp;
      my $footer_file_l = $sells_dir."\\c".$tmp;

      if (! -e $detail_file_u && ! -e $detail_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de detalhes\n";
      }
      elsif (! -e $footer_file_u && ! -e $footer_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de cabealhos\n";
      }
      else {
        $successes++;
      }
    }
  }

  $s->debug("We found $errors errors and $successes valids...");

  return $errors == 0 ? 1 : 0;
}

sub backup_file {
  my %a = ( file => undef, days_to_keep => 5, @_ );

  die "Missing 'file' parameter..." unless defined $a{file};

  my( $file, $folder, $suffix ) = fileparse($a{file});

  my $max_existant = 0;
  &find( sub { &wanted($file, \$max_existant, $a{days_to_keep}) }, $folder );
  &move($a{file}, $a{file}."_".sprintf("%04d", ++$max_existant));
}

sub wanted {
  my $file = shift;
  my $max = shift;
  my $days_to_keep = shift;

  my $found_original = 0;

  if (/^$file/) {
    # backup
    if (/_(\d+)$/) {
      $$max = $1 if $1 > $$max;
    }
  }

  my $file_p = $file."_";

  if (-f $File::Find::name && /^$file_p/) {
    my $days_old = -A $File::Find::name;
    &remove( $File::Find::name ) if $days_old > $days_to_keep;
  }
}

sub EncriptFOPassword {
  my $pass = shift;

  my @troca = ( 1, 9, 2, 8, 3, 7, 4, 6, 5, 0 );
  my @ordem = ( 2, 0, 3, 1 );

  my $buf = sprintf("%04d", $pass);
  my @chars = split //, $buf;
  my $buf2; 

  for (my $i = 0; $i < 4; $i++) {
    $buf2 .= $troca[$chars[$ordem[$i]]];
  }

  return int($buf2);
}

sub DecriptFOPassword {
  my $cript = shift;

  my @troca = ( 9, 0, 2, 4, 6, 8, 7, 5, 3, 1 );
  my @ordem = ( 1, 3, 0, 2 );

  my $buf = sprintf("%04d", $cript);
  my @chars = split //, $buf;
  my $buf2; 

  for (my $i = 0; $i < 4; $i++) {
    $buf2 .= $troca[$chars[$ordem[$i]]];
  }

  return int($buf2);
}

1;

