package WO::Client::Integrators::rTOS::Table;

use strict;
use warnings;
use FileHandle;
use Data::Dumper;
use FindBin qw( $Bin );
use Digest::MD5 qw( md5_hex );

my $LINE_END = $^O eq 'linux' ? "\r\n" : "\n";

sub table { die "You must override the 'table' method..." }

sub fields { () }

sub line_end { return $LINE_END }

sub new {	
  my $c = shift;
  my %a = ( table      => undef,
            cache      => 0,
            cache_prefix => '',
            __cache_changed => 0,
            add_header => 1,
            _idx       => 0,
            fields     => [],
            _records   => [],
            _fields    => [],
            @_,
          );

  my $s = bless { %a }, ref($c) || $c;

  if ($a{cache}) {
    if (! -e $s->cache_file()) {
      # adds the header record for the content
      $s->_add_record( '"'.$s->get_table().'",R,,1'.$s->line_end() ) if $a{add_header};
    }
    else {
      $s->cache_read();
    }
  }
  else {
    # adds the header record for the content
    $s->_add_record( '"'.$s->get_table().'",R,,1'.$s->line_end() ) if $a{add_header};
  }

  # init fields from args
  if (scalar @{ $a{fields} }) {
    for my $f (@{ $a{fields} }) {
      push @{ $s->{_fields} }, { $f => '' };
    }
  }
  # init fields from class
  else {
    my @f = $s->fields();
    $s->{_fields} = \@f;
  }

  return $s;
}

sub get_table {
  my $s = shift;
  return defined $s->{table} ? $s->{table} : $s->table();
}

sub cache_file {
  my $s = shift;

  my $prefix = $s->{cache_prefix} || "";
  $prefix =~ s/\://;
  $prefix = "_".$prefix if $prefix;

  return $Bin."/".$s->get_table().$prefix.".cache";
}

sub add_row {
  my $s = shift;
  my %a = ( values => {}, @_ );

  my $cur_line = "";

  # cycle to all defaults and override received values
  for my $f (@{ $s->{_fields} }) {
    my @a = keys %$f;
    my $k = lc($a[0]);
    my $v = defined $f->{ $k } ? $f->{ $k } : '';

    if ($k eq 'recordid' && (! exists $a{values}->{$k} || ! defined $a{values}->{$k} || $a{values}->{$k} eq '')) {
      $v = ++$s->{_idx};
    }
#    elsif (defined $a{values}->{$k} && $a{values}->{$k} ne '') {
    elsif (exists $a{values}->{$k}) {
      $v = $a{values}->{$k};
    }

    $cur_line .= &_escape($v).",";
  }

  if ($s->{cache}) {
    my $key = "";

    my $changed_record = 0;

    for ($s->cache_fields()) {
      my $f = $s->{_fields}->[$_];
      my @a = keys %$f;
      my $k = lc($a[0]);
      my $v = defined $f->{ $k } ? $f->{ $k } : '';

      if ($k eq 'recordid' && (! exists $a{values}->{$k} || ! defined $a{values}->{$k} || $a{values}->{$k} eq '')) {
        $v = $s->{_idx};
      }
      elsif (exists $a{values}->{$k}) {
        $v = $a{values}->{$k};
      }

      $v = "" unless defined $v;

      $key .= $v."#";
    }

    $key =~ s/#$//;

    my $md5 = md5_hex($cur_line);

    if (! exists $s->{__cache}->{$key}) {
      #print STDERR "[".ref($s)."] Record with key $key doesn't exists in cache, will add it...\n";
      $s->{__cache}->{$key} = $md5;
      $s->{__cache_changed} = 1;

      # adds record
      $s->_add_record( '"'.$s->get_table().'",U,,1,'.$cur_line.$s->line_end() );
      $changed_record = 1;
    }
    elsif (exists $s->{__cache}->{$key} && $s->{__cache}->{$key} ne $md5) {
      print STDERR "[".ref($s)."] Record with key $key exists in cache but changed... will update.\n";
      $s->{__cache}->{$key} = $md5;
      $s->{__cache_changed} = 1;

      # adds record
      $s->_add_record( '"'.$s->get_table().'",U,,1,'.$cur_line.$s->line_end() );
      $changed_record = 1;
    }

    return $changed_record;
  }
  else {
    # adds record
    $s->_add_record( '"'.$s->get_table().'",U,,1,'.$cur_line.$s->line_end() );
  }

  return 1;
}

sub cache_read {
  my $s = shift;

  if (open(F, "< ".$s->cache_file())) {
    while(<F>) {
      my( $k, $v ) = $_ =~ /^(.*?)=(.*?)$/;
      $s->{__cache}->{$k} = $v;
    }
    close(F);
  }
}

sub dump_records {
  my $s = shift;
  my %a = ( to_file => undef, @_ );

  my $created = 0;
  my $fh = FileHandle->new();

  for (@{ $s->{_records} }) {
    my $line = $_;
    if (! defined $a{to_file}) {
      print $line;
    } else {
      if (! $created) {
        if (! $fh->open(">> ".$a{to_file})) {
          die "Can't write to tile ".$a{to_file}.": $!";
        }

        $created = 1;
      }

      print $fh $line;
    }
  }

  $fh->close() if $created;
}

sub DESTROY {
  my $s = shift;

  if ($s->{cache} && $s->{__cache_changed}) {
    #print STDERR "Will flush cache to HDD...\n";

    if (open(F, "> ".$s->cache_file())) {
      foreach my $k (keys %{$s->{__cache}}) {
        print F "$k=".$s->{__cache}->{$k}."\n";
      }
      close(F);
    }
  }
}

# ##############################################################
# Internal methods
# ##############################################################

sub _add_record {
  my $s = shift;
  my $str = shift;

  # TODO: This method was created because all records are in memory.
  # We can change the method to make it save records to hdd instead of memory.

  push @{ $s->{_records} }, $str;
}

sub _escape {
  my $v = shift;
  return "" unless defined $v;

  #return "" if $v eq '0';

  if ($v =~ /^\d+$/ || $v =~ /^\d+\.\d+$/) {
    return $v;
  } else {
    return '"'.$v.'"';
  }
}

1;

