package WO::Client::WinRest;

use strict;
no warnings;

use base qw ( WO::Client::Common );

use File::Copy qw( move copy );
use File::Spec;
use POSIX qw( strftime );
use File::Slurp qw( read_file read_dir );
use Archive::Zip qw( :ERROR_CODES :CONSTANTS );
use File::Basename qw( fileparse );
use File::Find qw( find );
use Data::Dumper;
use FindBin qw( $Bin );
use Encode qw( encode );

use WinRest::File::Articles;
use WinRest::File::Families;
use WinRest::File::Waitress;
use WinRest::File::Clients;
use WinRest::File::Parameters;
use WinRest::File::Promotions;
use WinRest::File::Exclusions;
use WinRest::File::Automatics;
use WinRest::File::FastScreens;
use WinRest::File::HappyHours;

#
# FO to weckOFFICE
#

sub sync_fo_information {
  my $s = shift;

  $s->debug("sync_fo_information()");

  my $xml = $s->info();

  my $fo_type = $s->encode_fo_type();
  $s->debug("Encoded FO type is: $fo_type");

  # frontoffice directory
  my $dir = $xml->{dir} || "";

  # winrest sells dir
  my $sells_dir = $xml->{sells_dir} || "export";

  my $date_from = $s->get_date_from();
  $date_from =~ s/\-//g if defined $date_from;
  $date_from =~ s/\///g if defined $date_from;

  $s->debug("WILL USE $date_from");

  # remove sells after import
  my $remove_sells_after_import = $xml->{remove_sells_after_import} || 0;
  my @possible_remove = ();

  if (! -d $dir) {
    die "Frontoffice dir '$dir' don't exists, where is the frontoffice?";
  }

  # directory separator
  my $dsep = $s->directory_separator();

  my $fo_version = &get_lic_key($dir.$dsep.'config'.$dsep.'wrstsc00.lic', 'VERSION');
  my $hardlock = &get_lic_key($dir.$dsep.'config'.$dsep.'wrstsc00.lic', 'SERIALNUMBER');
  $s->debug("WinREST Version: $fo_version");
  $s->debug("WinREST Hardlock: $hardlock");

  my @pwds = &cserver_passwords();
  $s->debug("CPWDS:");
  $s->dump(\@pwds);

  $xml->{sector_master} ||= "000";

  my $file_warehouses = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmarm.'.$xml->{sector_master};
  my $file_articles = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmart.'.$xml->{sector_master};
  my $file_families = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmfam.'.$xml->{sector_master};
  my $file_waitress = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmemp.'.$xml->{sector_master};
  my $file_params = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrsttprm.'.$xml->{sector_master};
  my $file_clients = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstmcli.'.$xml->{sector_master};
  my $file_locals = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrsttloc.'.$xml->{sector_master};
  my $file_promotions = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstprom.'.$xml->{sector_master};
  my $file_exclusions = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstexcl.'.$xml->{sector_master};

  my $dir_movements = $xml->{dir_movements} || undef;
  $dir_movements = $dir.$dsep.'files'.$dsep.$sells_dir unless defined $dir_movements;

  #
  # Prepare ZIP file
  #

  my $zip = Archive::Zip->new();

  my $zipFile = $xml->{dir}.$dsep.$fo_type.'_FO_DATA.zip';
  $s->own_remove( $zipFile ) if -e $zipFile;

  my @files_cleanup = ();
  $s->common_sync_fo_and_sync_bk( $xml, $zip, \@files_cleanup );

  #
  # Get unprocessed files found in the server
  #

  my $unprocessed_files = $s->weckoffice_send_data( { action => 'FrontofficeUnprocessedFiles', }
                                                  );

  my $VAR1 = {};

  my %unprocessed_files = ();
  my @unprocessed_files = split //, $unprocessed_files;
  map { $unprocessed_files{$_} = 1 } @unprocessed_files;

  # default values are 1
  $xml->{sync_users} = exists $xml->{sync_users} ? $xml->{sync_users} : 1;
  $xml->{sync_articles} = exists $xml->{sync_articles} ? $xml->{sync_articles} : 1;
  $xml->{sync_families} = exists $xml->{sync_families} ? $xml->{sync_families} : 1;
  $xml->{sync_warehouses} = exists $xml->{sync_warehouses} ? $xml->{sync_warehouses} : 1;
  $xml->{sync_clients} = exists $xml->{sync_clients} ? $xml->{sync_clients} : 1;
  $xml->{sync_sells} = exists $xml->{sync_sells} ? $xml->{sync_sells} : 1;
  $xml->{sync_locals} = exists $xml->{sync_locals} ? $xml->{sync_locals} : 1;
  $xml->{sync_menus} = exists $xml->{sync_menus} ? $xml->{sync_menus} : 1;

  # used to define if we want deleted fo content or not
  $xml->{filter_deleted_content}          = exists $xml->{filter_deleted_content} ? $xml->{filter_deleted_content} : 1;
  $xml->{articles_filter_deleted_content} = exists $xml->{articles_filter_deleted_content} ? $xml->{articles_filter_deleted_content} : 1;
  $xml->{clients_filter_deleted_content}  = exists $xml->{clients_filter_deleted_content} ? $xml->{clients_filter_deleted_content} : 1;
  $xml->{users_filter_deleted_content}    = exists $xml->{users_filter_deleted_content} ? $xml->{users_filter_deleted_content} : 1;
  $xml->{families_filter_deleted_content} = exists $xml->{families_filter_deleted_content} ? $xml->{families_filter_deleted_content} : 1;

  my $company_config = $s->company_configuration();
  $company_config->{winrest_use_new_code_scheme} ||= 0;


  if (exists $company_config->{disable_data_send} && $company_config->{disable_data_send}) {
    $s->debug("Data send is disabled in the company_configurations... disabling masters...");
    $xml->{sync_menus} = 0;
    $xml->{sync_locals} = 0;
    $xml->{sync_clients} = 0;
    $xml->{sync_warehouses} = 0;
    $xml->{sync_families} = 0;
    $xml->{sync_articles} = 0;
    $xml->{sync_users} = 0;
  }

  # default value are 0
  $xml->{check_sell_files_errors} ||= 0;

  if ($xml->{sync_warehouses}) {
    #
    # Warehouses
    #
    if (-e $file_warehouses) {
      $zip->addFile( $file_warehouses ) unless exists $unprocessed_files{ 'wrstmarm.'.$xml->{sector_master} };
    } else {
      $s->debug("Didn't find the warehouses '$file_warehouses' file...");
    }
  }

  #
  # Parameters
  #

  if (-e $file_params) {
    $zip->addFile( $file_params ) unless exists $unprocessed_files{ 'wrsttprm.'.$xml->{sector_master} };
  } else {
    $s->debug("Didn't find the parameters '$file_params' file...");
  }

  if ($xml->{sync_articles}) {
    #
    # Articles
    #

    if (-e $file_articles) {
    	if ($xml->{filter_deleted_content} || $xml->{articles_filter_deleted_content}) {
        my $fname = &_filter_deleted_from_fo_file( $file_articles, 235 );
        push @files_cleanup, $fname;
			  $zip->addFile( $fname );
    	} else {
    		$zip->addFile( $file_articles ) unless exists $unprocessed_files{ 'wrstmart.'.$xml->{sector_master} };
    	}
    } else {
      $s->debug("Didn't find the articles '$file_articles' file...");
    }
  }

  if ($xml->{sync_families}) {
    #
    # Families
    #

    if (-e $file_families) {
    	if ($xml->{filter_deleted_content} || $xml->{families_filter_deleted_content}) {
    	  my $fname = &_filter_deleted_from_fo_file( $file_families, 198 );
        push @files_cleanup, $fname;
    	  $zip->addFile( $fname );
    	} else {
    		$zip->addFile( $file_families ) unless exists $unprocessed_files{ 'wrstmfam.'.$xml->{sector_master} };
    	}
    } else {
      $s->debug("Didn't find the families '$file_families' file...");
    }
  }

  if ($xml->{sync_users}) {
    #
    # Waitresses
    #

    if (-e $file_waitress) {
    	if ($xml->{filter_deleted_content} || $xml->{users_filter_deleted_content}) {
    	  my $fname = &_filter_deleted_from_fo_file( $file_waitress, 317 );
        push @files_cleanup, $fname;
    	  $zip->addFile( $fname );
    	} else {
    		$zip->addFile( $file_waitress ) unless exists $unprocessed_files{ 'wrstmemp.'.$xml->{sector_master} };
    	}
    } else {
      $s->debug("Didn't find the waitress '$file_waitress' file...");
    }
  }

  if ($xml->{sync_locals}) {
    #
    # Locals
    #

    if (-e $file_locals) {
      $zip->addFile( $file_locals ) unless exists $unprocessed_files{ 'wrsttloc.'.$xml->{sector_master} };
    } else {
      $s->debug("Didn't find the locals '$file_locals' file...");
    }
  }

  if ($xml->{sync_menus}) {
    #
    # Promotions
    #

    if (-e $file_promotions) {
      $zip->addFile( $file_promotions ) unless exists $unprocessed_files{ 'wrstprom.'.$xml->{sector_master} };
    } else {
      $s->debug("Didn't find the promotions '$file_promotions' file...");
    }
  }

    #
    # Exclusions
    #

    if (-e $file_exclusions) {
      $zip->addFile( $file_exclusions );
    }

  if ($xml->{sync_clients}) {
    #
    # Waitresses
    #

    if (-e $file_clients) {
    	if ($xml->{filter_deleted_content} || $xml->{clients_filter_deleted_content}) {
    	  my $fname = &_filter_deleted_from_fo_file( $file_clients, 387 );
        push @files_cleanup, $fname;
    	  $zip->addFile( $fname );
    	} else {
    		$zip->addFile( $file_clients ) unless exists $unprocessed_files{ 'wrstmcli.'.$xml->{sector_master} };
    	}
    } else {
      $s->debug("Didn't find the clients '$file_clients' file...");
    }
  }

  #
  # Movements
  #

  if ($xml->{sync_sells} && -d $dir_movements) {
    my $found_something = $s->sync_movements($zip, \@possible_remove, $date_from, 'sync_fo_information');
  }

  my $file_sizes_watch = $xml->{file_sizes_watch} || "";

  if ($file_sizes_watch) {
    my $file_sizes = $s->tmp_dir() . $dsep . "file_sizes.txt";
    open(F, "> $file_sizes");
    close(F);
    $zip->addFile( $file_sizes );
  }

  die "write error $zipFile" unless $zip->writeToFileNamed( $zipFile ) == AZ_OK;

  #
  # Send ZIP file to the server
  #

  eval {
    $s->weckoffice_send_data( { action => 'FrontofficeReceiveData',
                                file_upload => $zipFile,
                              }
                            );
  };

  $s->own_remove($zipFile) if -e $zipFile;
  $s->own_remove($_) for @files_cleanup;

  if ($@) {
    die "Error sending file: $@";
  } else {
    if ($file_sizes_watch) {
      $s->_save_file_sizes();
    }

    # remove sales after import, if requested
    if ($remove_sells_after_import && scalar @possible_remove) {
      map { $s->own_remove($_) } @possible_remove;
    }
  }

  $s->debug("Done.");
}

#
# Brings weckOFFICE information to FO
#

sub sync_bk_information {
  my $s = shift;

  $s->debug("sync_bk_information");

  # retrieve files associated with a tree
  #$s->receive_backoffice_tree_files();

  # service xml
  my $xml = $s->info();

  my $fo_type = $s->encode_fo_type();

  my $dir = $xml->{dir};

  if (! -d $dir) {
    die "Frontoffice dir '$dir' don't exists, where is the frontoffice?";
  }
  
  my @tmp_to_move;
  my $count_tmp_to_move = 0;
  my $dsep = $s->directory_separator();

  my $fo_version = &get_lic_key($dir.$dsep.'config'.$dsep.'wrstsc00.lic', 'VERSION');
  my $hardlock = &get_lic_key($dir.$dsep.'config'.$dsep.'wrstsc00.lic', 'SERIALNUMBER');
  $s->debug("WinREST Version: $fo_version");
  $s->debug("WinREST Hardlock: $hardlock");

  # check winrest config for export definitions
  if (-d $dir."/config") {
    &find( sub {
      my $fname = $_;
      
      return unless /wrstsc01\.000(\.\d+)?$/;
      return if /\d\d\d\d\d\d\d\d\d\d\d\d\d\d$/; # ignore backup files
      
      my $change = 0;
      my @lines;
      
      if (open(F, "< $File::Find::name")) {
        while (<F>) {
          if (my( $num ) = $_ =~ /TIMIMP(\d+)/) {
            my $n = length($1)-1;
            my $c = " " x $n;
            my $chars = "0$c";
            s/TIMIMP$1/TIMIMP$chars/;
            $change = 1 if $num > 0;
          }
          
          push @lines, $_;
        }
      }
      
      if ($change) {
        $s->debug("Will disable auto-update from 'import' in $fname...");
        
        my $tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));                   
        open(F, "> $tmp_file") or die "Can't write to $tmp_file: $!";                
        map { print F $_ } @lines;                                                   
        close(F);                                                                    
        $s->backup_file( file => $File::Find::name );                                
        $count_tmp_to_move++;                                                        
        push @tmp_to_move, [ $tmp_file, $File::Find::name ];   
      }
      
=pod
      if (open(F, "< $File::Find::name")) {
        my %export_map = (
          EXPRT1 => 'Familias',
          EXPRT2 => 'Artigos',
          EXPRT3 => 'Empregados',
          EXPRT4 => 'Clientes',
          EXPRT5 => 'Armazens',
          EXPRT6 => 'Parametros',
          EXPRT7 => 'Registos diarios',
          EXPRT8 => 'Para modem',
        );

        my @lines;
        my $valid = 1;
        
        while (<F>) {
          if ($_ !~ /EXPRT/) {
            push @lines, $_;
            next;
          }

          # debug collected data
          s/\r//; 
          s/\n//; 
          
          my( $str, $value ) = $_ =~ /(EXPRT\d)\s+(\d)/;
          $s->debug("\t".($2 ? 'SIM' : 'NAO')." ".$export_map{$1});

          $valid = 0 if $1 eq 'EXPRT1' && $2 == 1 && $valid == 1;
          $valid = 0 if $1 eq 'EXPRT2' && $2 == 1 && $valid == 1;
          $valid = 0 if $1 eq 'EXPRT3' && $2 == 1 && $valid == 1;
          $valid = 0 if $1 eq 'EXPRT4' && $2 == 1 && $valid == 1;
          $valid = 0 if $1 eq 'EXPRT5' && $2 == 1 && $valid == 1;
          $valid = 0 if $1 eq 'EXPRT6' && $2 == 1 && $valid == 1;
          $valid = 0 if $1 eq 'EXPRT7' && $2 == 0 && $valid == 1;
          $valid = 0 if $1 eq 'EXPRT8' && $2 == 1 && $valid == 1;
        }
        close(F);

        if (defined $valid && $valid == 0) {
          $s->debug("Will fix export definitions in ".$File::Find::name." (inactive at the moment)");
#=pod
          push @lines, "     0EXPRT1                     0\r\n";
          push @lines, "     0EXPRT2                     0\r\n";
          push @lines, "     0EXPRT3                     0\r\n";
          push @lines, "     0EXPRT4                     0\r\n";
          push @lines, "     0EXPRT5                     0\r\n";
          push @lines, "     0EXPRT6                     0\r\n";
          push @lines, "     0EXPRT7                     1\r\n";
          push @lines, "     0EXPRT8                     0\r\n";

          my $tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));
          open(F, "> $tmp_file") or die "Can't write to $tmp_file: $!";
          map { print F $_ } @lines;
          close(F);
          $s->backup_file( file => $File::Find::name );
          $count_tmp_to_move++;
          push @tmp_to_move, [ $tmp_file, $File::Find::name ];
#=cut
        } else {
          $s->debug("WinRest export definitions are correct for us...");
        }
      }
=cut
    }, ( $dir."/config" ) );
  }  

  $xml->{use_import_dir} ||= 0;
  $xml->{dont_write_to_data} ||= 0;

  #if ($^O eq 'MSWin32') {
  #  $xml->{dont_write_to_data} = 0;
  #}

  $xml->{sector_master} ||= "000";

  my $out_dir = $xml->{use_import_dir} ? 'import' : 'data';
  my $sector = $xml->{use_import_dir} ? $xml->{sector_movements} : $xml->{sector_master};
  my $export_dir = "$Bin/../to_winrest";

  if ($xml->{dont_write_to_data} && ! -d $export_dir) {
    mkdir($export_dir);
  }

  my $parameters_file   = $dir.$dsep."files".$dsep.$out_dir.$dsep."wrsttprm.".$sector;
  my $articles_file     = $dir.$dsep."files".$dsep.$out_dir.$dsep."wrstmart.".$sector;
  my $families_file     = $dir.$dsep."files".$dsep.$out_dir.$dsep."wrstmfam.".$sector;
  my $waitresses_file   = $dir.$dsep."files".$dsep.$out_dir.$dsep."wrstmemp.".$sector;
  my $clients_file      = $dir.$dsep."files".$dsep.$out_dir.$dsep."wrstmcli.".$sector;
  my $menus_file        = $dir.$dsep."files".$dsep."data".$dsep."wrstprom.".$sector;
  my $excl_file         = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrstexcl.'.$sector;
  my $aut_file          = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrsttaut.'.$sector;
  my $fast_screens_file = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrsttpag.'.$sector;
  my $happy_hours_file  = $dir.$dsep.'files'.$dsep.'data'.$dsep.'wrsthapy.'.$sector;

  my $parameters_file_t   = $parameters_file;
  my $articles_file_t     = $articles_file;
  my $families_file_t     = $families_file;
  my $waitresses_file_t   = $waitresses_file;
  my $clients_file_t      = $clients_file;
  my $menus_file_t        = $menus_file;
  my $excl_file_t         = $excl_file;
  my $aut_file_t          = $aut_file;
  my $fast_screens_file_t = $fast_screens_file;
  my $happy_hours_file_t  = $happy_hours_file;

  # recover a missing parameters file
  if (! -e $parameters_file) {
    $s->debug("Parameters file don't exists, let's try to recover from backup...");

    my $d_dir = $dir.$dsep.'files'.$dsep.'data';
    my @files = &read_dir($d_dir);

    for (@files) {
      next unless /^wrsttprm/i;
      my $size = -s $d_dir.$dsep.$_;

      if ($size) {
        my $orig = $d_dir.$dsep.$_;
        my $dest = $d_dir.$dsep.'wrsttprm.'.$sector;
        &copy( $orig, $dest );
        last;
      }
    }
  }

  if ($xml->{remove_automatic_products_file}) {
    $s->debug("Will remove automatic products file...");
    $s->own_remove($aut_file);
  }

  if ($xml->{dont_write_to_data}) {
    $parameters_file_t   = $export_dir.$dsep."wrsttprm.".$sector;
    $articles_file_t     = $export_dir.$dsep."wrstmart.".$sector;
    $families_file_t     = $export_dir.$dsep."wrstmfam.".$sector;
    $waitresses_file_t   = $export_dir.$dsep."wrstmemp.".$sector;
    $clients_file_t      = $export_dir.$dsep."wrstmcli.".$sector;
    $menus_file_t        = $export_dir.$dsep."wrstprom.".$sector;
    $excl_file_t         = $export_dir.$dsep.'wrstexcl.'.$sector;
    $aut_file_t          = $export_dir.$dsep.'wrsttaut.'.$sector;
    $fast_screens_file_t = $export_dir.$dsep.'wrsttpag.'.$sector;
    $happy_hours_file_t  = $export_dir.$dsep.'wrsthapy.'.$sector;;

    if ($^O eq 'linux' && -e "/sbin/init2") {
      my $install_script =<<END;
#!/bin/sh

HAS_PUBLISH=`grep PUBLISH_DATA /sbin/init2 | grep weckOFFICE`

if [ -z "\$HAS_PUBLISH" ]; then
  HAS_LOAD_CONFIG=`grep "# load configuration" /sbin/init2`

  if [ -z "\$HAS_LOAD_CONFIG" ]; then
    echo "This version of WinREST isn't supported by this script..."
  else
    sed -i 's/# load configuration/\\/home\\/weckOFFICE\\/bin\\/start PUBLISH_DATA\\n    # load configuration/' /sbin/init2
    echo "Done. Please reboot the machine."
  fi
else
  echo "Publish procedure is installed already..."
fi
END

      my $has_publish = `grep PUBLISH_DATA /sbin/init2 | grep weckOFFICE`;

      if ($has_publish eq '') {
        open(F, "> /home/weckOFFICE/bin/install.sh") or die "Can't write to /home/weckOFFICE/bin/install.sh : $!";
        print F $install_script;
        close(F);

        $s->debug("Will execute /home/weckOFFICE/bin/install.sh");
        my $res = system("sh /home/weckOFFICE/bin/install.sh");
        $s->debug("result: $res");

        $s->debug("Will remove /home/weckOFFICE/bin/install.sh");
        $s->own_remove("/home/weckOFFICE/bin/install.sh");
      }
    } # end linux check

    #WINDOWS CHECK
    if ($^O eq 'MSWin32') {
    }
  }

=pod
  if (! -f $parameters_file) {
    if (open(F, "> $parameters_file")) {
      while (<DATA>) {
        print F $_;
      }
      close(F);
    }

    #die("Can't find parameters file '$parameters_file'");

    $s->debug("Can't find parameters file '$parameters_file'");
  }
=cut

  my @pwds = &cserver_passwords();
  $s->debug("CPWDS:");

  my $date_from = $s->get_date_from();
  $date_from =~ s/\-//g if defined $date_from;
  $date_from =~ s/\///g if defined $date_from;
  $s->debug("WILL USE $date_from");

=pod


  if (($xml->{weckoffice}->{host} eq 'eurest.weckoffice.com' || $xml->{weckoffice}->{host} eq 'eurestmakro.weckoffice.com') && -e $parameters_file_t) {
    my $tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));

    my @lines;
    my $last_line = "";
    my $found = 0;

    open(F, "< $parameters_file_t");
    while (<F>) {
      push @lines, $_;

      if (my( $d ) = $_ =~ /TIVA(\d+)/) {
        $found = 1 if $_ =~ /20\.00/;
        $last_line = $_;
      }
    }
    close(F);

    if (! $found) {
      $s->debug("Paramters file don't have the 20% tax, will add it...");
      my( $d ) = $last_line =~ /TIVA(\d+)/;

      if ($d =~ /^\d+$/) {
        if (open(F, "> $tmp_file")) {
          for (@lines) {
            print F $_;
          }
          close(F);
        }

        my $idx = $d+1;

        my %fields = (
          param_type  => 'TIVA',
          code        => $idx,
          description => '20',
          param_value => '20.00',
          text        => '',
        );

        WinRest::File::Parameters->create(
          file => $tmp_file,
          fields => \%fields,
        );

        &move( $tmp_file, $parameters_file_t );
      }
    } else {
      $s->debug("Parameters file already has the 20% tax...");
    }
  }
=cut

  my %tax_codes = ();
  my @params = ();

  eval {
    my $params = WinRest::File::Parameters->new( file => $parameters_file );
    @params = $params->get_all();
  };

  foreach my $p (@params) {
    $tax_codes{ $p->param_value()*1 } = $p->code()*1 if $p->param_type() eq 'TIVA';
  }

  # default values are 1
  $xml->{sync_users}         = exists $xml->{sync_users}         ? $xml->{sync_users}         : 1;
  $xml->{sync_articles}      = exists $xml->{sync_articles}      ? $xml->{sync_articles}      : 1;
  $xml->{sync_families}      = exists $xml->{sync_families}      ? $xml->{sync_families}      : 1;
  $xml->{sync_clients}       = exists $xml->{sync_clients}       ? $xml->{sync_clients}       : 1;
  $xml->{sync_sells}         = exists $xml->{sync_sells}         ? $xml->{sync_sells}         : 1;
  $xml->{sync_menus}         = exists $xml->{sync_menus}         ? $xml->{sync_menus}         : 1;
  $xml->{sync_exclusions}    = exists $xml->{sync_exclusions}    ? $xml->{sync_exclusions}    : 1;
  $xml->{sync_parameters}    = exists $xml->{sync_parameters}    ? $xml->{sync_parameters}    : 0;
  $xml->{sync_auto_articles} = exists $xml->{sync_auto_articles} ? $xml->{sync_auto_articles} : 0;
  $xml->{sync_fast_screens}  = exists $xml->{sync_fast_screens}  ? $xml->{sync_fast_screens} : 0;
  $xml->{sync_fast_screens_use_sort_method}  = exists $xml->{sync_fast_screens_use_sort_method}  ? $xml->{sync_fast_screens_use_sort_method} : 0;
  $xml->{sync_happy_hours}  = exists $xml->{sync_happy_hours}  ? $xml->{sync_happy_hours} : 0;

  my $str;
  my $VAR1 = undef;

  my $families_tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));
  my $articles_tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));
  my $promotions_tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));
  my $params_tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));
  my @params_to_add = ();
  my $params_filter_customer_groups = 0;
  my $params_filter_article_groups = 0;
  my $params_filter_discount_table = 0;
  my $params_filter_fast_screens = 0;

  my @automatic_products = ();
  my @promotions;
  my @articles;
  my @families;

  #
  # Integration control
  #

  my %control_integrations = ();

  eval {
    $str = $s->weckoffice_send_data( { action => 'BackofficeGetControlIntegrations',
                                     }
                                   );
  };

  if ($@) {
    die("Error: $@");
  }

  eval $str;

  if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
    for (@$VAR1) {
      $control_integrations{ $_->{content_type} } = $_;
    }
  }

  my $company_config = $s->company_configuration();
  $company_config->{winrest_use_new_code_scheme} ||= 0;

  if (exists $company_config->{disable_data_send} && $company_config->{disable_data_send}) {
    $s->debug("Data send is disabled in the company_configurations... setting 'manual' to 0...");
    $s->{manual} = 0;
  }

  $s->debug("###### manual: #".$s->manual()."# (sync_parameters: ".$xml->{sync_parameters}.")");

  #
  # Sync Parameters
  #

  if ($s->manual() && $xml->{sync_parameters}) {
    $s->debug("Will process parameters...");

    my @discount_groups = ();
    my @article_groups  = ();
    my @discount_table  = ();

    eval {
      $str = $s->weckoffice_send_data( { action => 'WinRestCompanyDiscountGroups',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      @discount_groups = @$VAR1;
    }

    eval {
      $str = $s->weckoffice_send_data( { action => 'WinRestCompanyArticleGroups',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      @article_groups = @$VAR1;
    }

    eval {
      $str = $s->weckoffice_send_data( { action => 'WinRestCompanyDiscountTable',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      @discount_table = @$VAR1;
    }

    $s->debug("Got ".scalar(@discount_groups)." discount_groups, ".scalar(@article_groups)." article_groups, ".scalar(@discount_table)." discount table");

    for (@discount_groups) {
      my $id   = sprintf("%-10.10s", $_->{id});
      my $name = sprintf("%-20.20s", $_->{name});
      my $value = sprintf("%13.13s", '0.00');
      # winrest doesn't like this, so, we remove it
      $name =~ s/%/ /g;
      $value =~ s/%/ /g;
      next unless $id*1 =~ /^\d+$/;

      $params_filter_customer_groups = 1;

      push @params_to_add, {
        param_type => 'GDSC',
        code        => $id,
        description => $name,
        param_value => $value,
        text        => '',        
      };
    }

    # article groups
    for (@article_groups) {
      my $id    = sprintf("%-10.10s", $_->{id});
      my $name  = sprintf("%-20.20s", $_->{name});
      my $value = sprintf("%13.13s", '0.00');
      # winrest doesn't like this, so, we remove it
      $name =~ s/%/ /g;
      $value =~ s/%/ /g;
      next unless $id*1 =~ /^\d+$/;

      $params_filter_article_groups = 1;

      push @params_to_add, {
        param_type => 'GDSA',
        code        => $id,
        description => $name,
        param_value => $value,
        text        => '',        
      };
    }

    for (@discount_table) {
      my $id    = sprintf("%-10.10s", $_->{discount_group_id} . $_->{article_group_id});
      my $name  = sprintf("%-20.20s", "");
      my $value = sprintf("%13.13s", $_->{value});
      # winrest doesn't like this, so, we remove it
      $name =~ s/%/ /g;
      $value =~ s/%/ /g;
      next unless $id*1 =~ /^\d+$/;

      $params_filter_discount_table = 1;

      push @params_to_add, {
        param_type => 'GDSM',
        code        => $id,
        description => $name,
        param_value => $value,
        text        => '',        
      };
    }
  }

  $VAR1 = undef;

  #
  # Sync articles
  #

  my %used_families = ();
  my %arts_codes;
  my %done_articles;

  if ($s->manual() && $xml->{sync_articles}) {
    if (! -f $articles_file) {
      #die("Can't find articles file '$articles_file'");
    }

    eval {
      $str = $s->weckoffice_send_data( { action => 'BackofficeGetArticles',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      # update export num
      if ($xml->{use_import_dir}) {
        my $export_num = $control_integrations{articles}->{fo_export_seq};
        $export_num++;
        $control_integrations{articles}->{fo_export_seq} = $export_num;
      }

      if (scalar @$VAR1) {

      $s->debug("Will process ".scalar(@$VAR1)." articles...");
      $s->debug("THIS IS PROBABLY A DEMO COMPANY!!!!") if scalar(@$VAR1) == 30;

      my @art_fields_t = WinRest::File::Articles::fields();
      my @art_defaults = WinRest::File::Articles::defaults();

      my @art_fields = ();
      map { my $key = (keys %$_)[0]; push @art_fields, $key } @art_fields_t;

      my %art_defaults = ();
      map { my $key = (keys %$_)[0]; $art_defaults{$key} = $_->{$key} } @art_defaults;

      my $use_vord = exists $xml->{use_vord} ? $xml->{use_vord} : 1;

      #my %done_articles_cp;

      my $deleted = 0;

      foreach my $a (@$VAR1) {
        my $price1 = defined $a->{article_price1} && $a->{article_price1} ne '0.00' ? $a->{article_price1} : ($a->{fo_seetings}->{mart_1_pv} || '0.00');
        my $price2 = defined $a->{article_price2} && $a->{article_price2} ne '0.00' ? $a->{article_price2} : ($a->{fo_seetings}->{mart_2_pv} || '0.00');
        my $price3 = defined $a->{article_price3} && $a->{article_price3} ne '0.00' ? $a->{article_price3} : ($a->{fo_seetings}->{mart_3_pv} || '0.00');
        my $price4 = defined $a->{article_price4} && $a->{article_price4} ne '0.00' ? $a->{article_price4} : ($a->{fo_seetings}->{mart_4_pv} || '0.00');
        my $price5 = defined $a->{article_price5} && $a->{article_price5} ne '0.00' ? $a->{article_price5} : ($a->{fo_seetings}->{mart_5_pv} || '0.00');

        my $tax1 = 0;
        my $tax2 = 0;

        if (defined $a->{article_tax0} && $a->{article_tax0} ne '') {
          #$a->{article_tax0} = sprintf("%.2f", $a->{article_tax0});
          $tax1 = exists $tax_codes{ $a->{article_tax0}*1 } ? $tax_codes{ $a->{article_tax0}*1 } : '0';
        } else {
          $tax1 = $a->{fo_seetings}->{mart_iva1} || '0';
        }

        if (defined $a->{article_tax1} && $a->{article_tax1} ne '') {
          $a->{article_tax1} = sprintf("%.2f", $a->{article_tax1});
          $tax2 = exists $tax_codes{ $a->{article_tax1}*1 } ? $tax_codes{ $a->{article_tax1}*1 } : '0';
        } else {
          $tax2 = $a->{fo_seetings}->{mart_iva2} || '0';
        }

        my $name = defined $a->{article_name} ? $a->{article_name} : ($a->{fo_seetings}->{mart_nome} || "");

        #if ($name =~ /Aguia/i) {
        #  print STDERR "### $name - $tax1 (".$a->{article_tax0}.") - $tax2 (".$a->{article_tax1}.")\n";
        #}

        my $family = defined $a->{family_fo_id} ? $a->{family_fo_id} : ($a->{fo_seetings}->{mart_fami} || "");
        next if $family eq '';

        if (exists $a->{fo_seetings}->{mart_anul} && $a->{fo_seetings}->{mart_anul} eq 'T') {
          $family = 8001;
          $deleted++;
        }

        my $code = defined $a->{article_fo_id} ? $a->{article_fo_id} : ($a->{fo_seetings}->{mart_codg} || "");
        my $cped = exists $a->{fo_seetings}->{mart_cped} && $a->{fo_seetings}->{mart_cped} ne '' ? $a->{fo_seetings}->{mart_cped} : $code;
        $cped = '' if (length($cped) > 5);

        if ($company_config->{winrest_use_new_code_scheme} && exists $a->{fo_code_seq} && $a->{fo_code_seq}) {
          #$cped = $code;
          $cped = $a->{article_store_fo_id};
          $code = $a->{fo_code_seq};
        }

        if ($code !~ /^\d+$/) {
          $s->debug("Articles can only have numbers in their codes (".$code."), skipping...");
          next;
        }

        $arts_codes{ $a->{article_id} } = $code;

        # If the product was processed already, ignore it. WinRest only supports one per family
        if (exists $done_articles{$code}) {
          $s->debug("Product with internal code $code is duplicate, will be ignored (".$a->{article_id}.")...");
          next;
        }

        my $cbot = $a->{fo_seetings}->{_cbot} ? $a->{fo_seetings}->{_cbot}: 255;
        my $ncur = $a->{fo_seetings}->{_ncur} ? $a->{fo_seetings}->{_ncur} : ((length($name) > 13) ? substr($name,0,13) : $name);
        my $nbot = $a->{fo_seetings}->{_nbot} ? $a->{fo_seetings}->{_nbot} : ((length($name) > 32) ? substr($name,0,32) : $name);
        my $vord = $a->{fo_seetings}->{_vord} ? $a->{fo_seetings}->{_vord} : 0;

        my $barcode = defined $a->{article_barcode} ? $a->{article_barcode} : ($a->{fo_seetings}->{mart_cbar} || "");

        $vord = "" unless $use_vord;

        my $first_num;

        if (length($family) < 4) {
          ( $first_num ) = $family =~ /^(\d)/;
        } else {
          ( $first_num ) = $family =~ /^(\d\d)/;
        }

        $used_families{ $first_num }++ if defined $first_num;
		
        my %article_main_props = (
          mart_codg => $code,
          mart_cbot => $cbot,
          mart_cped => $cped,
          mart_fami => $family,
          mart_nome => $name,
          mart_ncur => $ncur,
          mart_nbot => $nbot,
          mart_1_pv => $price1,
          mart_2_pv => $price2,
          mart_3_pv => $price3,
          mart_4_pv => $price4,
          mart_5_pv => $price5,
          mart_iva1 => $tax1,
          mart_iva2 => $tax2,
          mart_cbar => $barcode,
          mart_vord => $vord,
        );

        my %all_props = %article_main_props;

        # initialize all the unixistant fields
        foreach my $field (@art_fields) {
          next if exists $article_main_props{$field};
          $all_props{$field} = exists $a->{fo_seetings}->{$field} ? $a->{fo_seetings}->{$field} : $art_defaults{$field};
        }

        if ($xml->{use_import_dir}) {
          $all_props{mart_nexp} = $control_integrations{articles}->{fo_export_seq};
        }

        # winrest doesn't like this, so, we remove it
        $all_props{mart_nome} =~ s/%/ /g;

        if (exists $all_props{mart_papv} && $all_props{mart_papv} eq '1') {
          $all_props{mart_papv} = "T";
        } elsif (exists $all_props{mart_papv} && $all_props{mart_papv} eq '0') {
          $all_props{mart_papv} = "F";
        }

        $done_articles{ $code } = \%all_props;

        push @articles, \%all_props;
      }

      if ($deleted > 0) {
        $s->debug("$deleted products are deleted (historical)...");
      }

      } else {
        $s->debug("Received 0 articles, ignoring...");
      }
    } else {
      die("Couldn't get articles from weckOFFICE server");
    }
  }

  $VAR1 = undef;

  #
  # Sync articles - Historical
  #

  if ($s->manual() && $xml->{sync_articles_historical}) {
    if (! -f $articles_file) {
      #die("Can't find articles file '$articles_file'");
    }

    eval {
      $str = $s->weckoffice_send_data( { action => 'BackofficeGetArticlesHistorical',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      # update export num
      if ($xml->{use_import_dir}) {
        my $export_num = $control_integrations{articles}->{fo_export_seq};
        $export_num++;
        $control_integrations{articles}->{fo_export_seq} = $export_num;
      }

      if (scalar @$VAR1) {

      $s->debug("Will process ".scalar(@$VAR1)." articles...");
      $s->debug("THIS IS PROBABLY A DEMO COMPANY!!!!") if scalar(@$VAR1) == 30;

      my @art_fields_t = WinRest::File::Articles::fields();
      my @art_defaults = WinRest::File::Articles::defaults();

      my @art_fields = ();
      map { my $key = (keys %$_)[0]; push @art_fields, $key } @art_fields_t;

      my %art_defaults = ();
      map { my $key = (keys %$_)[0]; $art_defaults{$key} = $_->{$key} } @art_defaults;

      my $tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));

      my $use_vord = exists $xml->{use_vord} ? $xml->{use_vord} : 1;

      #my %done_articles; # this is using to top scope variable
      #my %done_articles_cp;

      my $last_fam = "";

      foreach my $a (reverse sort { $a->{article_id} <=> $b->{article_id} } @$VAR1) {
        my $price1 = defined $a->{article_price1} && $a->{article_price1} ne '0.00' ? $a->{article_price1} : ($a->{fo_seetings}->{mart_1_pv} || '0.00');
        my $price2 = defined $a->{article_price2} && $a->{article_price2} ne '0.00' ? $a->{article_price2} : ($a->{fo_seetings}->{mart_2_pv} || '0.00');
        my $price3 = defined $a->{article_price3} && $a->{article_price3} ne '0.00' ? $a->{article_price3} : ($a->{fo_seetings}->{mart_3_pv} || '0.00');
        my $price4 = defined $a->{article_price4} && $a->{article_price4} ne '0.00' ? $a->{article_price4} : ($a->{fo_seetings}->{mart_4_pv} || '0.00');
        my $price5 = defined $a->{article_price5} && $a->{article_price5} ne '0.00' ? $a->{article_price5} : ($a->{fo_seetings}->{mart_5_pv} || '0.00');

        my $tax1 = 0;
        my $tax2 = 0;

        if (defined $a->{article_tax0} && $a->{article_tax0} ne '') {
          #$a->{article_tax0} = sprintf("%.2f", $a->{article_tax0});
          $tax1 = exists $tax_codes{ $a->{article_tax0}*1 } ? $tax_codes{ $a->{article_tax0}*1 } : '0';
        } else {
          $tax1 = $a->{fo_seetings}->{mart_iva1} || '0';
        }

        if (defined $a->{article_tax1} && $a->{article_tax1} ne '') {
          #$a->{article_tax1} = sprintf("%.2f", $a->{article_tax1});
          $tax2 = exists $tax_codes{ $a->{article_tax1}*1 } ? $tax_codes{ $a->{article_tax1}*1 } : '0';
        } else {
          $tax2 = $a->{fo_seetings}->{mart_iva2} || '0';
        }

        my $name = defined $a->{article_name} ? $a->{article_name} : ($a->{fo_seetings}->{mart_nome} || "");

        my $family = defined $a->{family_fo_id} ? $a->{family_fo_id} : ($a->{fo_seetings}->{mart_fami} || "");

        my $code = defined $a->{article_fo_id} ? $a->{article_fo_id} : ($a->{fo_seetings}->{mart_codg} || "");
        my $cped = exists $a->{fo_seetings}->{mart_cped} && $a->{fo_seetings}->{mart_cped} ne '' ? $a->{fo_seetings}->{mart_cped} : $code;

        if ($company_config->{winrest_use_new_code_scheme} && exists $a->{fo_code_seq} && $a->{fo_code_seq}) {
          $cped = $code;
          $code = $a->{fo_code_seq};
        }

        # If the product was processed already, ignore it. WinRest only supports one per family
        if (exists $done_articles{$code}) {
          #$s->debug("Product with internal code $code is duplicate, will be ignored (".$a->{article_id}.")...");
          next;
        }

        #if ($done_articles_cp{$cped}) {
        #  $s->debug("Product with request code $cped is duplicate, will be ignored...");
        #  next if exists $done_articles_cp{ $cped };
        #}

        #$done_articles{ $code } = $name;
        #$done_articles_cp{ $cped } = 1;

        my $cbot = $a->{fo_seetings}->{_cbot} ? $a->{fo_seetings}->{_cbot} : 255;
        my $ncur = $a->{fo_seetings}->{_ncur} ? $a->{fo_seetings}->{_ncur} : ((length($name) > 13) ? substr($name,0,13) : $name);
        my $nbot = $a->{fo_seetings}->{_nbot} ? $a->{fo_seetings}->{_nbot} : ((length($name) > 32) ? substr($name,0,32) : $name);
        my $vord = $a->{fo_seetings}->{_vord} ? $a->{fo_seetings}->{_vord} : 0;

        my $barcode = defined $a->{article_barcode} ? $a->{article_barcode} : ($a->{fo_seetings}->{mart_cbar} || "");

        $vord = "" unless $use_vord;

        my $first_num;
		
		if (! defined $family || $family eq '') {
		  $family = 8001;
		  $first_num = 8001;
		} else {
        $family = 0 if ! defined $family || $family eq '';

        if (length($family) < 4) {
          ( $first_num ) = $family =~ /^(\d)/;
        } else {
          ( $first_num ) = $family =~ /^(\d\d)/;
        }
		}

        $last_fam = $family if $family != 0;
        $family = $last_fam if $family == 0;

        $used_families{ $first_num }++;

        my %article_main_props = (
          mart_codg => $code,
          mart_cbot => $cbot,
          mart_cped => $cped,
          mart_fami => $family,
          mart_nome => $name,
          mart_ncur => $ncur,
          mart_nbot => $nbot,
          mart_1_pv => $price1,
          mart_2_pv => $price2,
          mart_3_pv => $price3,
          mart_4_pv => $price4,
          mart_5_pv => $price5,
          mart_iva1 => $tax1,
          mart_iva2 => $tax2,
          mart_cbar => $barcode,
          mart_vord => $vord,
    		  mart_anul => 'T',
        );

        my %all_props = %article_main_props;

        # initialize all the unixistant fields
        foreach my $field (@art_fields) {
          next if exists $article_main_props{$field};
          $all_props{$field} = exists $a->{fo_seetings}->{$field} ? $a->{fo_seetings}->{$field} : $art_defaults{$field};
        }

        if ($xml->{use_import_dir}) {
          $all_props{mart_nexp} = $control_integrations{articles}->{fo_export_seq};
        }

        # winrest doesn't like this, so, we remove it
        $all_props{mart_nome} =~ s/%/ /g;

        if (exists $all_props{mart_papv} && $all_props{mart_papv} eq '1') {
          $all_props{mart_papv} = "T";
        } elsif (exists $all_props{mart_papv} && $all_props{mart_papv} eq '0') {
          $all_props{mart_papv} = "F";
        }

        $s->debug("HISTORICAL PRODUCT ------------ $code - $name");

        $done_articles{ $code } = \%all_props;

        push @articles, \%all_props;

=pod
        # create article_line
        WinRest::File::Articles->create(
          file => $tmp_file,
          fields => \%all_props,
        );
=cut
      }

=pod
      if (-s $tmp_file) {
        $s->backup_file( file => $articles_file );
        $count_tmp_to_move++;
        push @tmp_to_move, [ $tmp_file, $articles_file_t ];
      }
=cut

      } else {
        $s->debug("Received 0 articles, ignoring...");
      }
    } else {
      die("Couldn't get articles from weckOFFICE server");
    }
  }

  $VAR1 = undef;

  #
  # Sync Families
  #

  my %fams_codes;

  if ($s->manual() && $xml->{sync_families}) {
    if (! -f $families_file) {
      #die("Can't find families file '$families_file'");
    }

    eval {
      $str = $s->weckoffice_send_data( { action => 'BackofficeGetFamilies',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      # update export num
      if ($xml->{use_import_dir}) {
        my $export_num = $control_integrations{families}->{fo_export_seq};
        $export_num++;
        $control_integrations{families}->{fo_export_seq} = $export_num;
      }

      if (scalar @$VAR1) {

      $s->debug("Will process ".scalar(@$VAR1)." families...");

      my @fam_fields_t = WinRest::File::Families::fields();
      my @fam_defaults = WinRest::File::Families::defaults();

      my @fam_fields = ();
      map { my $key = (keys %$_)[0]; push @fam_fields, $key } @fam_fields_t;

      my %fam_defaults = ();
      map { my $key = (keys %$_)[0]; $fam_defaults{$key} = $_->{$key} } @fam_defaults;

      foreach my $f (@$VAR1) {
        $fams_codes{ $f->{family_id} } = $f->{family_fo_id};

        my $code = defined $f->{family_fo_id} ? $f->{family_fo_id} : ($f->{fo_seetings}->{mfam_codg} || "");
        my $name = defined $f->{family_name} ? $f->{family_name} : ($f->{fo_seetings}->{mfam_nome} || "");

        my $nbot = defined $f->{fo_seetings}->{_nbot} && $f->{fo_seetings}->{_nbot} ne 'S/Nome' ? $f->{fo_seetings}->{_nbot} : $f->{family_name};
        my $ncur = defined $f->{fo_seetings}->{_ncur} && $f->{fo_seetings}->{_ncur} ne 'S/Nome' ? $f->{fo_seetings}->{_ncur} : $f->{family_name};

        $name =~ s/&amp;/&/g;
        $nbot =~ s/&amp;/&/g;

        my $cbot = $f->{fo_seetings}->{_cbot} ? $f->{fo_seetings}->{_cbot} : 255;

        my %family_main_props = (
          mfam_codg => $code,
          mfam_nome => $name,
          mfam_nbot => $nbot,
          mfam_ncur => $ncur,
          mfam_cbot => $cbot,
        );

        my %all_props = %family_main_props;

        # initialize all the unixistant fields
        foreach my $field (@fam_fields) {
          next if exists $family_main_props{$field};
          $all_props{$field} = exists $f->{fo_seetings}->{$field} ? $f->{fo_seetings}->{$field} : $fam_defaults{$field};
        }

        if ($xml->{use_import_dir}) {
          $all_props{mfam_nexp} = $control_integrations{families}->{fo_export_seq};
        }

        my $first_num;

        if (length($code) < 4) {
          ( $first_num ) = $code =~ /^(\d)/;
        } else {
          ( $first_num ) = $code =~ /^(\d\d)/;
        }

        if (! exists $used_families{ $first_num }) {
          $s->debug("### Family $code ($name) wasnt used in any article...");
        } elsif ($name eq '') {
          $s->debug("### Family $code has an empty name, ignoring...");
        } else {
          # winrest doesn't like this, so, we remove it
          $all_props{mfam_nome} =~ s/%/ /g;

          if (exists $all_props{mfam_cpat} && $all_props{mfam_cpat} eq '1') {
            $all_props{mfam_cpat} = "T";
          } elsif (exists $all_props{mfam_cpat} && $all_props{mfam_cpat} eq '0') {
            $all_props{mfam_cpat} = "F";
          }

          push @families, \%all_props;
        }
      }
=pod
      my %family_main_props = (
        mfam_codg => '9999',
        mfam_nome => 'Historical',
        mfam_nbot => 'Historical',
        mfam_ncur => 'Historical',
        mfam_anul => 'T',
      );

      push @families, \%family_main_props;
=cut
      push @families, {
          mfam_codg => 8000,
          mfam_nome => 'Historical',
          mfam_ncur => 'Historical',
          mfam_nbot => 'Historical',
          mfam_anul => 'T',
      };

      push @families, {
          mfam_codg => 8001,
          mfam_nome => 'Historical',
          mfam_ncur => 'Historical',
          mfam_nbot => 'HIstorical',
      };

      } else {
        $s->debug("Received 0 families, ignoring...");
      }
    } else {
      die("Couldn't get families from weckOFFICE server");
    }
  }

  $VAR1 = undef;

  #
  # Sync Waitresses
  #

  my %done_users;

  if ($s->manual() && $xml->{sync_users}) {
    if (! -f $waitresses_file) {
      #die("Can't find waitresses file '$waitresses_file'");
    }

    eval {
      $str = $s->weckoffice_send_data( { action => 'BackofficeGetWaitresses',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      # update export num
      if ($xml->{use_import_dir}) {
        my $export_num = $control_integrations{users}->{fo_export_seq};
        $export_num++;
        $control_integrations{users}->{fo_export_seq} = $export_num;
      }

      if (scalar @$VAR1) {

      $s->debug("Will process ".scalar(@$VAR1)." waitresses...");
      $s->debug("THIS IS PROBABLY A DEMO COMPANY!!!!") if scalar(@$VAR1) == 3;

      my @wai_fields_t = WinRest::File::Waitress::fields();
      my @wai_defaults = WinRest::File::Waitress::defaults();

      my @wai_fields = ();
      map { my $key = (keys %$_)[0]; push @wai_fields, $key } @wai_fields_t;

      my %wai_defaults = ();
      map { my $key = (keys %$_)[0]; $wai_defaults{$key} = $_->{$key} } @wai_defaults;

      my $cur_codg = 1;

      my $tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));

      my %map_access_type = (
        10 => 'D',
        11 => 'E',
        12 => 'F',
        13 => 'G',
        14 => 'H',
        15 => 'I',
        16 => 'F',
      );

      foreach my $w (@$VAR1) {
        my $fo_id = exists $w->{real_fo_id} ? $w->{real_fo_id} : $w->{fo_id};

        if ($fo_id !~ /^\d+$/) {
          $s->debug("Waitresses can only have numbers in their codes (".$fo_id."), skipping...");
          next;
        }

        my $login = $w->{fo_login} || "";

        if ($login eq '') {
          $s->debug("Clerk $fo_id | ".$w->{name}." don't have a pos login defined, skipping...");
          next;
        }

        $w->{fo_access_type} ||= 2;

        if ($done_users{$fo_id}) {
          my $code = $fo_id;
          $s->debug("User with internal code $code is duplicate, will be ignored...");
          next if exists $done_users{ $fo_id };
        }

        $done_users{ $fo_id } = 1;

        if (exists $map_access_type{ $w->{fo_access_type} }) {
          $w->{fo_access_type} = $map_access_type{ $w->{fo_access_type} };
        }

        my %waitress_main_props = (
          memp_ctge => $w->{fo_access_type},
          memp_ctrb => $w->{nif} || "",
          
          memp_dadm => $w->{dadmission} || "",

          memp_mora => $w->{address} || "",
          memp_iden => $w->{bi} || "",
          memp_loca => $w->{localidade} || "",
          memp_telf => $w->{telephone} || "",
          memp_nome => $w->{name} || "",
          memp_nint => $login,    # fo_login
          memp_iden => $login,
          memp_pasw => &EncriptFOPassword($w->{fo_passwd}) || &EncriptFOPassword(0),  # fo_passwd
          memp_codg => $fo_id,
        );

        # initialize all the unixistant fields
        foreach my $field (@wai_fields) {
          next if exists $waitress_main_props{$field} && $waitress_main_props{$field} ne '';
          $waitress_main_props{$field} = exists $w->{fo_settings}->{$field} && $w->{fo_settings}->{$field} ne '' ? $w->{fo_settings}->{$field} : $wai_defaults{$field};
        }

        # if the code is null, create a temp one
        if ($waitress_main_props{memp_codg}) {
          $cur_codg = $waitress_main_props{memp_codg} + 1;
        } else {
          $waitress_main_props{memp_codg} = $cur_codg++;
        }

        if ($xml->{use_import_dir}) {
          $waitress_main_props{memp_nexp} = $control_integrations{users}->{fo_export_seq};
        }

        $waitress_main_props{memp_cbot} = $w->{fo_seetings}->{_cbot} ? $w->{fo_seetings}->{_cbot} : 255;

        # winrest doesn't like this, so, we remove it
        $waitress_main_props{memp_nome} =~ s/%/ /g;

        if ($waitress_main_props{memp_dadm}) {
          my( $d2, $m2, $y2 ) = $waitress_main_props{memp_dadm} =~ /^(\d+)[-\/](\d+)[-\/](\d+)$/;
          $waitress_main_props{memp_dadm} = "$m2/$d2/$y2";
        }

        WinRest::File::Waitress->create(
          file => $tmp_file,
          fields => \%waitress_main_props,
        );
      }

      if (-s $tmp_file) {
        $s->backup_file( file => $waitresses_file );
        $count_tmp_to_move++;
        push @tmp_to_move, [ $tmp_file, $waitresses_file_t ];
      }
      } else {
        $s->debug("Received 0 waitresses, ignoring...");
      }
    } else {
      die("Couldn't get waitresses from weckOFFICE server");
    }
  }

  $VAR1 = undef;

  #
  # Sync Menus
  #

  if ($s->manual() && $xml->{sync_menus}) {
    if (! -f $menus_file) {
      #die("Can't find clients file '$menus_file'");
    }

    eval {
      $str = $s->weckoffice_send_data( { action => 'BackofficeGetMenuArticles',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      if (scalar @$VAR1) {
        $s->debug("Will process ".scalar(@$VAR1)." menu records...");

        my @menu_fields_t = WinRest::File::Promotions::fields();
        my @menu_defaults = WinRest::File::Promotions::defaults();

        my @menu_fields = ();
        map { my $key = (keys %$_)[0]; push @menu_fields, $key } @menu_fields_t;

        my %menu_defaults = ();
        map { my $key = (keys %$_)[0]; $menu_defaults{$key} = $_->{$key} } @menu_defaults;

        my $tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));

        foreach my $w (@$VAR1) {
          if ($w->{fo_id} !~ /^\d+$/) {
            next;
          }

          if ($w->{composite_fo_id} !~ /^\d+$/) {
            next;
          }

          my $option = $w->{option} || 1;
          $option = sprintf("%2s", $option);

          my $xtra_cost = $w->{xtra_cost} || 0;
          $xtra_cost = sprintf("%12s", sprintf("%0.2f", $xtra_cost));

          my %menu_main_props = (
            parent_article => $w->{top_internal_code}  || $w->{fo_id} || "",
            step_number    => $option,
            article        => $w->{comp_internal_code} || $w->{composite_fo_id} || "",
            xtra_cost      => $xtra_cost,
          );

          # initialize all the unixistant fields
          for (@menu_fields) {
            if (! exists $menu_main_props{$_}) {
              if (exists $menu_defaults{$_} && defined $menu_defaults{$_}) {
                $menu_main_props{$_} = $menu_defaults{$_};
              } else {
                $menu_main_props{$_} = "";
              }
            }
          }

          push @promotions, \%menu_main_props;
        }
      } else {
        $s->debug("Received 0 menu records, ignoring...");
      }
    } else {
      die("Couldn't get menus from weckOFFICE server");
    }
  }

  $VAR1 = undef;

  #
  # Sync Clients
  #

  if ($s->manual() && $xml->{sync_clients}) {
    eval {
      $str = $s->weckoffice_send_data( { action => 'BackofficeGetStoreClients',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      $xml->{keep_customer_balances_from_till} ||= 0;

      my %customers_credit_cache = ();

      if ($xml->{keep_customer_balances_from_till} && -e $clients_file) {
        my $o = WinRest::File::Clients->new( file => $clients_file );
        my $recs = $o->get_all();

        foreach my $r (@$recs) {
          my $fo_id = $r->mcli_codg();

          $customers_credit_cache{$fo_id} = {
            mcli_tcre => $r->mcli_tcre(),
            mcli_tdeb => $r->mcli_tdeb(),
            mcli_sald => $r->mcli_sald(),
            mcli_pont => $r->mcli_pont(),
          };
        }
      }

      # update export num
      if ($xml->{use_import_dir}) {
        my $export_num = $control_integrations{clients}->{fo_export_seq};
        $export_num++;
        $control_integrations{clients}->{fo_export_seq} = $export_num;
      }

      if (scalar @$VAR1) {

      $s->debug("Will process ".scalar(@$VAR1)." clients...");

      my @cli_fields_t = WinRest::File::Clients::fields();
      my @cli_defaults = WinRest::File::Clients::defaults();

      my @cli_fields = ();
      map { my $key = (keys %$_)[0]; push @cli_fields, $key } @cli_fields_t;

      my %cli_defaults = ();
      map { my $key = (keys %$_)[0]; $cli_defaults{$key} = $_->{$key} } @cli_defaults;

      my $cur_codg = 1;

      my $tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));

      my %done_clients = ();

      my $has_customers_credit_cache = scalar keys %customers_credit_cache;

      foreach my $w (@$VAR1) {
        if ($xml->{weckoffice}->{store} eq 'CODE:299' && $xml->{weckoffice}->{host} eq 'eurest.weckoffice.com') {
          $w->{fo_id} =~ s/^M//i if $w->{fo_id} =~ /^M/i;
        }

        if ($w->{fo_id} !~ /^\d+$/) {
          $s->debug("Customers can only have numbers in their codes (".$w->{fo_id}."), skipping...");
          next;
        }

        if ($done_clients{$w->{fo_id}}) {
          my $code = $w->{fo_id};
          $s->debug("Client with internal code $code is duplicate, will be ignored...");
          next if exists $done_clients{ $w->{fo_id} };
        }

        $done_clients{ $w->{fo_id} } = 1;

        my $card = $w->{card_number} && $w->{card_number} ne '' ? $w->{card_number} : "";

        my %client_main_props = (
          mcli_codg => $w->{fo_id} || "",
          mcli_nome => $w->{name} || "",
          mcli_telf => $w->{telephone} || "",
          mcli_fax => $w->{fax} || "",
          mcli_loca => $w->{localidade} || "",
          mcli_mora => $w->{address} || "",
          mcli_cpos => $w->{postal_code} || "",
          mcli_ctrb => $w->{ncontrib} || "",
          mcli_ncrd => $card,

          mcli_dnas => $w->{birth_date} || "",
          #mcli_dcrd => $w->{card_expiration} || "01/01/2100",
          mcli_dcrd => $w->{card_expiration} || "",

          mcli_tcre => $w->{total_credit} || "",
          mcli_tdeb => $w->{total_debt} || "",
          mcli_limi => $w->{credit_limit} || "9999999",
          mcli_sald => $w->{balance} || "",
          mcli_anul => $w->{deleted} || "",
          mcli_obsv => $w->{obs} || "",
        );

        # initialize all the unixistant fields
        foreach my $field (@cli_fields) {
          next if exists $client_main_props{$field} && $client_main_props{$field} ne '';
          $client_main_props{$field} = exists $w->{fo_settings}->{$field} && defined $w->{fo_settings}->{$field} && $w->{fo_settings}->{$field} ne '' ? $w->{fo_settings}->{$field} : $cli_defaults{$field};
        }

        # if the code is null, create a temp one
        if ($client_main_props{mcli_codg}) {
          $cur_codg = $client_main_props{mcli_codg} + 1;
        } else {
          $client_main_props{mcli_codg} = $cur_codg++;
        }

        if ($xml->{use_import_dir}) {
          $client_main_props{mcli_nexp} = $control_integrations{clients}->{fo_export_seq};
        }

        # winrest doesn't like this, so, we remove it
        $client_main_props{mcli_nome} =~ s/%/ /g;

        if ($has_customers_credit_cache && exists $customers_credit_cache{$client_main_props{mcli_codg}}) {
          my $h = $customers_credit_cache{$client_main_props{mcli_codg}};

          $client_main_props{mcli_tcre} = $h->{mcli_tcre};
          $client_main_props{mcli_tdeb} = $h->{mcli_tdeb};
          $client_main_props{mcli_sald} = $h->{mcli_sald};
          $client_main_props{mcli_pont} = $h->{mcli_pont};
        }

        if ($client_main_props{mcli_dnas}) {
          my( $d1, $m1, $y1 ) = $client_main_props{mcli_dnas} =~ /^(\d+)[-\/](\d+)[-\/](\d+)$/;
          $client_main_props{mcli_dnas} = "$m1/$d1/$y1";
        }

        if ($client_main_props{mcli_dcrd}) {
          my( $d2, $m2, $y2 ) = $client_main_props{mcli_dcrd} =~ /^(\d+)[-\/](\d+)[-\/](\d+)$/;
          eval { $client_main_props{mcli_dcrd} = "$m2/$d2/$y2"; };
        }

        WinRest::File::Clients->create(
          file => $tmp_file,
          fields => \%client_main_props,
        );
      }

      if (-s $tmp_file) {
        $s->backup_file( file => $clients_file );
        $count_tmp_to_move++;
        push @tmp_to_move, [ $tmp_file, $clients_file_t ];
      }
      } else {
        $s->debug("Received 0 clients, ignoring...");
      }
    } else {
      die("Couldn't get clients from weckOFFICE server");
    }
  }

  $VAR1 = undef;

  #
  # Sync Exclusions
  #

  if ($xml->{weckoffice}->{host} eq 'eurest.weckoffice.com') {
    $s->debug("Will backup and remove exclusions file...");
    $s->backup_file( file => $excl_file );
    $s->own_remove($excl_file);
  }

  if ($s->manual() && $xml->{sync_exclusions}) {
    $s->debug("Will process exclusions...");

    eval {
      $str = $s->weckoffice_send_data( { action => 'WinRestGetExclusions',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    my @exclusions = ();

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      @exclusions = @$VAR1;
    }

    $s->debug("Received : ".scalar(@exclusions));

    if (scalar @exclusions) {
      my $tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));

      my %control = ();

      foreach my $e (@exclusions) {
        my $loc_type = defined $e->{local_fo_id} ? 1 : 0;
        my $loc_code = defined $e->{local_fo_id} ? $e->{local_fo_id} : $e->{terminal_fo_id};

        my $content_type = "";
        my $content_code = "";

        if (defined $e->{acess_id}) {
          $content_type = 3;
          $content_code = $e->{acess_id};
        } elsif (defined $e->{article_id}) {
          next unless exists $arts_codes{ $e->{article_id} };
          $content_type = 0;
          $content_code = $arts_codes{ $e->{article_id} };
        } elsif (defined $e->{family_id}) {
          next unless exists $fams_codes{ $e->{family_id} };
          $content_type = 2;
          $content_code = $fams_codes{ $e->{family_id} };
        } elsif (defined $e->{user_fo_id}) {
          $content_type = 1;
          $content_code = $e->{user_fo_id};
        }

        my %all_props = (
          location_code => $loc_code,
          location_type => $loc_type,
          content_code  => $content_code,
          content_type  => $content_type,
        );

        $loc_code = "0" unless defined $loc_code;
        $loc_type = "0" unless defined $loc_type;
        $content_code = "0" unless defined $content_code;
        $content_type = "0" unless defined $content_type;

        next if exists $control{ $loc_code . $loc_type . $content_code . $content_type };
        $control{ $loc_code . $loc_type . $content_code . $content_type } = 1;

        # create article_line
        WinRest::File::Exclusions->create(
          file => $tmp_file,
          fields => \%all_props,
        );
      }

      if (-s $tmp_file) {
        $s->backup_file( file => $excl_file );
        $count_tmp_to_move++;
        push @tmp_to_move, [ $tmp_file, $excl_file_t ];
      }
    } else {
      $s->debug("Didn't receive any exclusion, will backup and remove file...");
      $s->backup_file( file => $excl_file );
      $s->own_remove($excl_file);
    }
  }

  $VAR1 = undef;

  #
  # Automatic Products
  #

  if ($s->manual() && $xml->{sync_auto_articles}) {
    eval {
      $str = $s->weckoffice_send_data( { action => 'BackofficeGetWinRestAutomaticArticles',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      $s->debug("Received ".scalar(@$VAR1)." records of automatic records...");

      foreach my $a (@$VAR1) {
        if (length($a->{article_internal_code}) == 6 && $a->{article_internal_code} =~ /^300/) {
          $a->{article_internal_code} =~ s/^300/301/;
        }

        my %autos_main_props = (
          prod_top        => $a->{article_internal_code},
          prod_auto       => $a->{auto_article_internal_code},
          quantity        => $a->{quantity},
          price_to_use    => $a->{price_to_use},
#          origin_table    => 
#          can_round       =>
#          multiply_factor =>
        );

        push @automatic_products, \%autos_main_props;
      }
    }
  }

  $VAR1 = undef;

  #
  # Fast Screens
  #

  if ($s->manual() && $xml->{sync_fast_screens}) {
    eval {
      $str = $s->weckoffice_send_data( { action => 'BackofficeGetWinRestFastScreens',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      $s->debug("Received ".scalar(@$VAR1)." records of fast screens...");

      my $tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));

      my %pages;
      my $idx = 0;

      if (scalar @$VAR1 && $xml->{sync_fast_screens_use_sort_method}) {
        push @families, {
          mfam_codg => 9000,
          mfam_nome => 'NetBO Fast Screens Top',
          mfam_ncur => 'NetBO Fast Screens Top',
          mfam_nbot => 'NetBO Fast Screens Top',
          mfam_anul => 'T',
        };

        push @families, {
          mfam_codg => 9001,
          mfam_nome => 'NetBO Fast Screens Parent',
          mfam_ncur => 'NetBO Fast Screens Parent',
          mfam_nbot => 'NetBO Fast Screens Parent',
        };
      }

      my @products_to_create = ();

      foreach my $a (@$VAR1) {
        if (! exists $pages{ $a->{name} }) {
          $pages{ $a->{name} }->{auto_complements} = $a->{auto_complements} || 0;
          $pages{ $a->{name} }->{idx} = ++$idx;

          if ($a->{auto_complements}) {
            WinRest::File::FastScreens->create(
              file => $tmp_file,
              fields => { page_code => $idx, product_code => 0 },
            );
          }
        }

        my $new_code = $a->{article_internal_code};

        if ($xml->{sync_fast_screens_use_sort_method}) {
          $new_code = 200000 + $a->{prod_order};
          push @products_to_create, { idx => $a->{prod_order}, new_code => $new_code, internal_code => $a->{article_internal_code} };
        }

        WinRest::File::FastScreens->create(
          file => $tmp_file,
          fields => { page_code => $idx, product_code => $new_code },
        );
      }

      for (@products_to_create) {
        my %info = %{ $done_articles{ $_->{internal_code} } };

        $info{'mart_codg'} = $_->{new_code};
        $info{'mart_fami'} = 9001;
        $info{'mart_cped'} = 0;
        $info{'mart_vord'} = $_->{idx};

        #$info{'mart_ncur'} = '   ';
        #$info{'mart_nome'} = '   ';

        push @articles, \%info;

        push @promotions, {
          parent_article => $_->{new_code},
          step_number    => 1,
          article        => $_->{internal_code},
          xtra_cost      => 0,
        };
      }

      if (-s $tmp_file) {
        $s->backup_file( file => $fast_screens_file );
        $count_tmp_to_move++;
        push @tmp_to_move, [ $tmp_file, $fast_screens_file_t ];
      }

      foreach my $p (sort { $pages{$a}->{idx} <=> $pages{$b}->{idx} } keys %pages) {
        my $h = $pages{$p};

        $params_filter_fast_screens = 1;

        push @params_to_add, {
          param_type => 'PAGR',
          code        => $h->{idx},
          description => $p,
          param_value => '',
          text        => '',        
        };
      }
    }
  }

  $VAR1 = undef;

  #
  # Happy Hours
  #

  if ($s->manual() && $xml->{sync_happy_hours}) {
    eval {
      $str = $s->weckoffice_send_data( { action => 'BackofficeGetWinRestHappyHours',
                                       }
                                     );
    };

    if ($@) {
      die("Error: $@");
    }

    eval $str;

    if (defined $VAR1 && ref($VAR1) eq 'ARRAY') {
      $s->debug("Received ".scalar(@$VAR1)." records of happy hours...");

      my $tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));

      my $cur_idx = 0;
      my $last_idx = -1;

      my $articles = [];
      my %done_qtty = ();

      foreach my $a (@$VAR1) {
        if ($a->{id} != $last_idx) {
          $last_idx = $a->{id};
          $cur_idx++;
        }

        WinRest::File::HappyHours->create(
          file => $tmp_file,
          fields => {
            code    => $cur_idx,
            code_local => $a->{lfo_id},
            week_day => $a->{day},
            start => $a->{start},
            stop => $a->{stop},
            price_to_use => $a->{price},
          },
        );

        if (exists $a->{articles} && scalar @{ $a->{articles} }) {
          for (@{ $a->{articles} }) {
            my $idx = $cur_idx*-1;

            next unless defined $_->{article_fo_id} && $_->{article_fo_id} ne '';

            push @$articles, {
              happy_hour => $idx,
              product => $_->{article_fo_id},
            } unless exists $done_qtty{ $idx }->{ $_->{article_fo_id} };

            $done_qtty{ $idx }->{ $_->{article_fo_id} }++;
          }
        }
      }

      if (-e $tmp_file) {
        $s->backup_file( file => $happy_hours_file );
        $count_tmp_to_move++;
        push @tmp_to_move, [ $tmp_file, $happy_hours_file_t ];
      }

      if (scalar @$articles) {
        foreach my $a (@$articles) {
          my $qtd = $done_qtty{ $a->{happy_hour} }->{ $a->{product} };

          my %autos_main_props = (
            prod_top        => $a->{happy_hour},
            prod_auto       => $a->{product},
            quantity        => $qtd,
            price_to_use    => 0,
  #          origin_table    => 
  #          can_round       =>
  #          multiply_factor =>
          );

          push @automatic_products, \%autos_main_props;
        }
      }
    }
  }

  $VAR1 = undef;

  #
  # Update collected families
  #

  if (scalar @families) {
    $s->backup_file( file => $families_file );

    for my $f (@families) {
      WinRest::File::Families->create(
        file => $families_tmp_file,
        fields => $f,
      );
    }

    $count_tmp_to_move++;
    push @tmp_to_move, [ $families_tmp_file, $families_file_t ];
  }

  #
  # Update collected products
  #

  if (scalar @articles) {
    $s->backup_file( file => $articles_file );

    for my $a (@articles) {
      WinRest::File::Articles->create(
        file => $articles_tmp_file,
        fields => $a,
      );
    }

    $count_tmp_to_move++;
    push @tmp_to_move, [ $articles_tmp_file, $articles_file_t ];
  }

  #
  # Update collected promotions
  #

  if (scalar @promotions) {
    $s->backup_file( file => $menus_file );

    for my $prom (@promotions) {
      next if $prom->{parent_article} eq '0' || $prom->{article} eq '0';

      WinRest::File::Promotions->create(
        file => $promotions_tmp_file,
        fields => $prom,
        line_end => $^O eq 'MSWin32' ? "\n" : "\r\n",
      );
    }

    $count_tmp_to_move++;
    push @tmp_to_move, [ $promotions_tmp_file, $menus_file_t ];
  }

  #
  # Update collected automatic products
  #

  if (scalar @automatic_products) {
    my $tmp_file = $s->tmp_dir() . $dsep . int(rand(1000000));

    $s->backup_file( file => $aut_file );

    for my $aut (@automatic_products) {
      WinRest::File::Automatics->create(
        file => $tmp_file,
        fields => $aut,
      );
    }

    $count_tmp_to_move++;
    push @tmp_to_move, [ $tmp_file, $aut_file_t ];
  }

  #
  # Update collected parameters
  #

  if (scalar @params_to_add) {
    my @filters = ();
    push @filters, 'GDSC' if $params_filter_customer_groups;
    push @filters, 'GDSA' if $params_filter_article_groups;
    push @filters, 'GDSM' if $params_filter_discount_table;
    push @filters, 'PAGR' if $params_filter_fast_screens;

    &_filter_from_parameters( $parameters_file, $params_tmp_file, \@filters ) if scalar @filters;

    for (@params_to_add) {
      WinRest::File::Parameters->create(
        file => $params_tmp_file,
        fields => {
          param_type  => $_->{param_type},
          code        => $_->{code},
          description => $_->{description},
          param_value => $_->{param_value},
          text        => $_->{text},
        },
      );
    }

    $s->backup_file( file => $parameters_file );
    $count_tmp_to_move++;
    push @tmp_to_move, [ $params_tmp_file, $parameters_file_t ];
  }

  #
  # Update integrations control
  #

  my @ci = values %control_integrations;
  $s->weckoffice_send_data( { action => 'BackofficeSetControlIntegrations',
                              control_integrations => $s->shrink(\@ci),
                            }
                          );

  #
  # Movements
  #

  # winrest sells dir
  my $sells_dir = $xml->{sells_dir} || "export";

  # movements dir
  my $dir_movements = $xml->{dir_movements} || undef;
  $dir_movements = $dir.$dsep.'files'.$dsep.$sells_dir unless defined $dir_movements;

  my $zip = Archive::Zip->new();

  my $zipFile = $xml->{dir}.$dsep.$fo_type.'_FO_DATA.zip';
  $s->own_remove( $zipFile ) if -e $zipFile;

  my @files_cleanup = ();
  $s->common_sync_fo_and_sync_bk( $xml, $zip, \@files_cleanup );

  #
  # move only in the end...
  #

  if (scalar @tmp_to_move && scalar @tmp_to_move == $count_tmp_to_move) {
    for (@tmp_to_move) {
      print STDERR "Will move ".$_->[0]." to ".$_->[1]."\n";

      if (-s $_->[0]) {
        &move( $_->[0], $_->[1] );

        if ($^O eq 'MSWin32') {
          # English
          if ($ENV{PROGRAMFILES} =~ /Program Files/) {
            my $st = system("cacls \"".$_->[1]."\" /E /G Everyone:F");
          }
          # Portuguese
          elsif ($ENV{PROGRAMFILES} =~ /Programas/) {
            my $st = system("cacls \"".$_->[1]."\" /E /G Todos:F");
          }
          else {
            $s->debug("Uknown language... can't call cacls");
          }
        }
      } else {
        $s->debug("Didn't move ".$_->[0]." to ".$_->[1]." because it had size 0!!!!!!!!");
      }
    }

    if (-e "/bin/sync") {
      my $res = system("/bin/sync");
      $s->debug("Result from sync: $res");
    }
    
    $xml->{use_update_warning} ||= 0;
    
    if ($^O eq 'linux' && $xml->{use_update_warning} && -f "/home/weckOFFICE/bin/done_updating_data.sh") {
      my @all = `ps xa | grep done_updating_data | grep -v grep`;

      if (! -e "/home/weckOFFICE/bin/kill.sh") {
        open(F, "> /home/weckOFFICE/bin/kill.sh");
        print F "#!/bin/bash\n";
        print F "MY_PID=\$1\n";
        print F "kill -9 \$MY_PID\n";
        close(F);
        system("chmod 755 /home/weckOFFICE/bin/kill.sh");
      }

      foreach my $res (@all) {
        chop($res);

        if ($res) {
          $res =~ s/^\s+//;
          my( $pid ) = $res =~ /^(\d+)/;

          if ($pid) {
            system("/home/weckOFFICE/bin/kill.sh $pid");
            #system("kill -9 $pid");
          }
        }
      }

      system("/bin/wconnect /home/weckOFFICE/bin/done_updating_data.sh &");
    }

    #WINDOWS CHECK
    if ($^O eq 'MSWin32') {
      opendir(DIR, $Bin."/../to_winrest");
      my @files = readdir(DIR);
      closedir(DIR);

      if ($xml->{dont_write_to_data} && scalar(@files) > 2) {
        my $ds = $s->directory_separator();
        my $f = $Bin.$ds."update_data.vbs";

        my $winrest_dir = $xml->{dir};
        my $to_winrest_dir = $Bin."/../to_winrest";
        $winrest_dir =~ s/\//\\/g;
        $to_winrest_dir =~ s/\//\\/g;

        my $str =<<END;
Option Explicit

Dim text, result, WShell

result = MsgBox("Uma Actualizacao de Dados esta disponivel deseja actualizar ?" & vbNewline & "Caso responda nao tera que sincronizar novamente", vbYesNo, "NetBO Client")

Set WShell = WScript.CreateObject("WScript.Shell")
WShell.CurrentDirectory = "$winrest_dir"

On Error Resume Next

If result = vbYes Then
  WShell.Run "taskkill.exe /F /IM winrestw.exe"
  WScript.Sleep 2000
  WShell.Run "cmd /C move /Y $to_winrest_dir\\* $winrest_dir\\files\\data"
  WScript.Sleep 1000
  WShell.Run "$winrest_dir\\winrestw.exe"
Else
End If
END
        open(F, "> $f");
        print F $str;
        close(F);

        my $script = $Bin.$dsep."update_data.vbs";

        system($script) if -e $script;
      }
    }   
  }

  $VAR1 = undef;

  if ($xml->{sync_sells} && -d $dir_movements) {
    # remove sells after import
    my $remove_sells_after_import = $xml->{remove_sells_after_import} || 0;

    my @possible_remove = ();

    $s->sync_movements($zip, \@possible_remove, $date_from, 'sync_bk_information');

    die('write error') unless $zip->writeToFileNamed( $zipFile ) == AZ_OK;

    #
    # Send ZIP file to the server
    #

    eval {
      $s->weckoffice_send_data( { action => 'FrontofficeReceiveData',
                                  file_upload => $zipFile,
                                }
                              );
    };

    if ($@) {
      die("Error sending file: $@");
    } else {
      my $file_sizes_watch = $xml->{file_sizes_watch} || "";

      if ($file_sizes_watch) {
        $s->_save_file_sizes();
      }

      # remove sales after import, if requested
      if ($remove_sells_after_import && scalar @possible_remove) {
        map { $s->own_remove($_) } @possible_remove;
      }
    }
  }

  $s->own_remove($_) for @files_cleanup;
  $s->own_remove($zipFile) if -e $zipFile;

  # retrieve files associated with a tree
  $s->receive_backoffice_tree_files();

  $s->debug("Done.");
}

#
# Sync movements
#

sub sync_movements {
  my $s = shift;
  my $zip = shift;
  my $remove = shift;
  my $date_from = shift;
  my $caller = shift || 'none';

  $s->debug("sync_movements()");

  my $found_something = undef;

  # service xml
  my $xml = $s->info();

  my $check_dates_fast = $xml->{check_dates_fast} || 0;

  my $fo_type = $s->encode_fo_type();

  # frontoffice directory
  my $dir = $xml->{dir} || "";

  my $file_sizes_watch = $xml->{file_sizes_watch} || "";

  # winrest sells dir
  my $sells_dir = $xml->{sells_dir} || "export";

  # directory separator
  my $dsep = $s->directory_separator();

  # movements dir
  my $dir_movements = $xml->{dir_movements} || undef;
  $dir_movements = $dir.$dsep.'files'.$dsep.$sells_dir unless defined $dir_movements;

  # remove sells after import
  my $remove_sells_after_import = $xml->{remove_sells_after_import} || 0;

  my $number_of_days_to_sync = $xml->{number_of_days_to_sync} || 0;

  my @files = &read_dir($dir_movements);

#  use Data::Dumper;
#  print STDERR Dumper \@files;

  #if ($xml->{weckoffice}->{host} eq 'eurest.weckoffice.com' && $xml->{weckoffice}->{store} eq 'CODE:240') {
  #if ($xml->{weckoffice}->{host} eq 'eurestcafe.weckoffice.com') {
  #if ($xml->{weckoffice}->{host} eq 'eurestmakro.weckoffice.com') {
  #if ($xml->{weckoffice}->{host} eq 'trincanela3.weckoffice.com') {
  #if ($xml->{weckoffice}->{host} eq 'giardino.weckoffice.com') {
  #  $s->dump(\@files);
  #}

  if ($xml->{check_sell_files_errors} && scalar @files) {
    $s->debug("Checking sell files for errors...");
    my $res = $s->_check_errors_in_sales($dir_movements, \@files);

    die "Sell files have errors, aborting..." unless $res;
  }

  #
  # Get unprocessed files found in the server
  #

  my $unprocessed_files = $s->weckoffice_send_data( { action => 'FrontofficeUnprocessedFiles',
                                                    }
                                                  );
  my $VAR1 = {};

  my $company_config = $s->company_configuration();

  my %valid_files = (
    b => 1,
    c => 1,
    d => 1,
    g => 1,
    h => 1,
    r => 1,
    s => 1,
    x => 1,
    n => 1,
    o => 1,
  );

  if ($file_sizes_watch) {
    my $sizes = $s->_read_file_sizes();

    foreach my $file (@files) {
      my $file_size = -s $dir_movements . $dsep . $file;

      my( $file_type, $date_str, $sector ) = $file =~ /^(\w)(\d+)\.(\d+)$/;
      next unless defined $date_str;

      my( $year, $mon, $day ) = $date_str =~ /^(\d\d)(\d\d)(\d\d)/;
      $year = "20".$year;

      if (defined $date_from && $year.$mon.$day < $date_from) {
        $s->debug("Ignoring date $year/$mon/$day with the 'date_from' clause...");
        next;
      }

      if (! exists $sizes->{$file} || $sizes->{$file} != $file_size) {
        my( $c, $rest ) = $file =~ /^(.)(.*)$/;
        my $member;

        if (exists $valid_files{ lc($c) }) {
          if (-e $dir_movements.$dsep.lc($c).$rest) {
            $s->debug("Will add file $dir_movements".$dsep.lc($c).$rest);
            $member = $zip->addFile( $dir_movements.$dsep.lc($c).$rest );
          }

          if (-e $dir_movements.$dsep.uc($c).$rest) {
            $s->debug("Will add file $dir_movements".$dsep.uc($c).$rest);
            $member = $zip->addFile( $dir_movements.$dsep.uc($c).$rest );
          }

          $found_something = 1;
          $sizes->{$file} = $file_size;
        }
      }
    }

    $s->{__file_sizes} = $sizes;

    if (! $found_something) {
      $s->debug("!! Using file sizes cache but nothing changed !!");
    }

    return $found_something
  }

  my $act = $check_dates_fast ? 'FrontofficeExistentSellDatesFast' : 'FrontofficeExistentSellDates';
  my $existent_dates = $s->weckoffice_send_data( { action => $act,
                                                 }
                                               );

  eval "$existent_dates";
  my %existent_dates = defined $VAR1 && ref($VAR1) eq 'HASH' ? %$VAR1 : ();

  my %unprocessed_files = ();
  my @unprocessed_files = split //, $unprocessed_files;
  map { $unprocessed_files{$_} = 1 } @unprocessed_files;

  my %process_dates = ();

  my $count_files = 0;

  my $sync_articles = exists $xml->{sync_articles} ? $xml->{sync_articles} : 1;
  my $sync_users    = exists $xml->{sync_users} ? $xml->{sync_users} : 1;
  my $sync_families = exists $xml->{sync_families} ? $xml->{sync_families} : 1;
  my $sync_clients  = exists $xml->{sync_clients} ? $xml->{sync_clients} : 1;

  foreach my $file (@files) {
    my( $ext ) = $file =~ /\.(.*?)$/;
    next unless defined $ext && $ext eq $xml->{sector_movements};

    if (exists $unprocessed_files{$file}) {
#      $s->debug("File $file is in the server already (wasn't processed by the server yet), skipping...");
      next;
    }

    my( $fname ) = $file =~ /^(.*)\./;

    if ($caller ne 'sync_bk_information') {
      if (uc($fname) eq uc('wrstmart') && $sync_articles) {
        my $member = $zip->addFile( $dir_movements.$dsep.$file );
        next;
      }
      elsif (uc($fname) eq uc('wrstmemp') && $sync_users) {
        my $member = $zip->addFile( $dir_movements.$dsep.$file );
        next;
      }
      elsif (uc($fname) eq uc('wrstmfam') && $sync_families) {
        my $member = $zip->addFile( $dir_movements.$dsep.$file );
        next;
      }
      elsif (uc($fname) eq uc('wrstmcli') && $sync_clients) {
        my $member = $zip->addFile( $dir_movements.$dsep.$file );
        next;
      }
    }

    my( $file_type, $date_str, $sector ) = $file =~ /^(\w)(\d+)\.(\d+)$/;

    next unless defined $date_str;

    my( $year, $mon, $day ) = $date_str =~ /^(\d\d)(\d\d)(\d\d)/;
    $year = "20".$year;

    if (defined $date_from && $year.$mon.$day < $date_from) {
#      $s->debug("Ignoring date $year/$mon/$day with the 'date_from' clause...");
      next;
    }

    if ($company_config->{import_sells_date_dupplicate} eq 'ignore' && exists $existent_dates{"$year-$mon-$day"}) {
#      $s->debug("Will ignore file '$file' because that date already exists");
      next;
    }

    if (exists $valid_files{ lc($file_type) }) {
      # no filter available
      if (! $number_of_days_to_sync) {
        push @$remove, $dir_movements.$dsep.$file if $remove_sells_after_import;
        my $member = $zip->addFile( $dir_movements.$dsep.$file );
        $found_something = 1;
        $count_files++;
      }
      # collect date for number_of_days_to_sync filter
      else {
        $process_dates{ $year.$mon.$day }->{$file_type} = $dir_movements.$dsep.$file;
      }
    } else {
      next;
    }
  }

  # use number_of_days_to_sync filter
  if ($number_of_days_to_sync && scalar keys %process_dates) {
    my $count = 0;

    for my $d (sort keys %process_dates) {
      for my $f (values %{$process_dates{$d}}) {
        push @$remove, $f if $remove_sells_after_import;
        my $member = $zip->addFile( $f );
      }

      last if ++$count == $number_of_days_to_sync;
    }
  }

  $s->debug("Found $count_files files to send...");

  if ($found_something) {
    my $fiscal_file = $xml->{dir}.$dsep."files".$dsep."data".$dsep."fiscaldata.000";

    if (-e $fiscal_file) {
      $zip->addFile( $fiscal_file );
    }
  }

  return $found_something;
}

sub common_sync_fo_and_sync_bk {
  my $s = shift;
  my( $xml, $zip, $files_cleanup ) = @_;

  my $dsep = $s->directory_separator();

  # machine uptime for linux
  if ($^O eq 'linux') {
    my $uptime = `/bin/uptime`;
    chop($uptime);
    $s->debug("UPTIME: $uptime");

    my $run_client = `grep "rm -rf" /home/weckOFFICE/bin/run_client.sh`;

    if (! $run_client && -d "/home/weckOFFICE/tmp") {
      open(F, ">> /home/weckOFFICE/bin/run_client.sh");
      print F "rm -rf /home/weckOFFICE/tmp/*\n";
      close(F);
    }
  }

  my $dir = $xml->{dir} || "";
  my $date_file = $dir . $dsep . 'files' . $dsep . 'data' . $dsep . 'wrstdata.000';

  if (-e $date_file) {
    local $/ = undef;
    open(F, "< $date_file");
    my $cnt = <F>;
    $s->debug("WinREST Date: $cnt");
    close(F);
  }

#  if ($xml->{weckoffice}->{host} eq 'eurestcafe.weckoffice.com' && $xml->{weckoffice}->{store} eq 'CODE:307') {
#    $s->debug("WILL WRITE TO DATE FILE!!!");
#    open(F, "> $date_file");
#    print F "10/27/2009\n";
#    close(F);
#  }

  if (-e "$Bin/../extra_info.txt" ) {
    $s->debug("Will add extra info file...");
    $zip->addFile( "$Bin/../extra_info.txt" );
    push @$files_cleanup, "$Bin/../extra_info.txt";
  }
  
  if (exists $xml->{sync_quiz_answers} && -f $xml->{sync_quiz_answers}) {
    $s->debug("Will add quiz file...");
    $zip->addFile( $xml->{sync_quiz_answers} );
    push @$files_cleanup, $xml->{sync_quiz_answers};
  }

  if ($xml->{sync_sells}) {
    #
    # use print origins mappings
    #

    $xml->{map_print_origins} ||= undef;

    if (defined $xml->{map_print_origins}) {
      my $file_print_origins = $s->tmp_dir() . $dsep . "map_print_origins.txt";

      my %mappings = ();

      if (exists $xml->{map_print_origins}->{map}) {
        if (ref $xml->{map_print_origins}->{map} eq 'HASH') {
          $mappings{ $xml->{map_print_origins}->{map}->{origin} } = $xml->{map_print_origins}->{map}->{store};
        } elsif (ref $xml->{map_print_origins}->{map} eq 'ARRAY') {
          for (@{ $xml->{map_print_origins}->{map} }) {
            $mappings{ $_->{origin} } = $_->{store};
          }
        }
      }

      if (scalar keys %mappings) {
        open(F, "> $file_print_origins") or die "Can't write to '$file_print_origins': $!\n";
        print F Dumper(\%mappings);
        close(F);
        $zip->addFile( $file_print_origins );
        push @$files_cleanup, $file_print_origins;
      }
    }

    #
    # use sale locals to map things
    #

    $xml->{map_sale_locals} ||= undef;

    if (defined $xml->{map_sale_locals}) {
      my $file_map_sale_locals = $s->tmp_dir() . $dsep . "map_sale_locals.txt";

      my @mappings = ();

      if (exists $xml->{map_sale_locals}->{map}) {
        if (ref $xml->{map_sale_locals}->{map} eq 'HASH') {
          push @mappings, $xml->{map_sale_locals}->{map};
        } elsif (ref $xml->{map_sale_locals}->{map} eq 'ARRAY') {
          @mappings = @{ $xml->{map_sale_locals}->{map} };
        }
      }

      if (scalar @mappings) {
        open(F, "> $file_map_sale_locals") or die "Can't write to '$file_map_sale_locals': $!\n";
        print F Dumper(\@mappings);
        close(F);
        $zip->addFile( $file_map_sale_locals );
        push @$files_cleanup, $file_map_sale_locals;
      }
    }

    $xml->{map_sale_customers} ||= undef;

    if (defined $xml->{map_sale_customers}) {
      my $file_map_sale_customers = $s->tmp_dir() . $dsep . "map_sale_customers.txt";

      my @mappings = ();

      if (exists $xml->{map_sale_customers}->{map}) {
        if (ref $xml->{map_sale_customers}->{map} eq 'HASH') {
          push @mappings, $xml->{map_sale_customers}->{map};
        } elsif (ref $xml->{map_sale_customers}->{map} eq 'ARRAY') {
          @mappings = @{ $xml->{map_sale_customers}->{map} };
        }
      }

      if (scalar @mappings) {
        open(F, "> $file_map_sale_customers") or die "Can't write to '$file_map_sale_customers': $!\n";
        print F Dumper(\@mappings);
        close(F);
        $zip->addFile( $file_map_sale_customers );
        push @$files_cleanup, $file_map_sale_customers;
      }
    }

    #
    # defines if we should use fb_d_narm in details as store
    #

    $xml->{fb_d_narm_in_details} ||= undef;

    if (defined $xml->{fb_d_narm_in_details}) {
      my $file_fb_d_narm = $s->tmp_dir() . $dsep . "fb_d_narm_in_details_as_store.txt";

      open(F, "> $file_fb_d_narm") or die "Can't write to '$file_fb_d_narm': $!\n";
      close(F);
      $zip->addFile( $file_fb_d_narm );
      push @$files_cleanup, $file_fb_d_narm;
    }
  }
}

sub _check_errors_in_sales {
  my $s = shift;
  my $sells_dir = shift;
  my $files = shift;

  my $errors = 0;
  my $successes = 0;

  foreach my $f (@$files) {
    if (my( $tmp ) = $f =~ /^c(.*?)$/i) {

      my $detail_file_u = $sells_dir."\\D".$tmp;
      my $detail_file_l = $sells_dir."\\d".$tmp;

      my $footer_file_u = $sells_dir."\\R".$tmp;
      my $footer_file_l = $sells_dir."\\r".$tmp;

      if (! -e $detail_file_u && ! -e $detail_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de detalhes\n";
      }
      elsif (! -e $footer_file_u && ! -e $footer_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de rodaps\n";
      }
      else {
        $successes++;
      }
    }
  }

  $s->debug("We found $errors errors and $successes valids...");

  $errors = 0;
  $successes = 0;

  foreach my $f (@$files) {
    if (my( $tmp ) = $f =~ /^d(.*?)$/i) {

      my $detail_file_u = $sells_dir."\\C".$tmp;
      my $detail_file_l = $sells_dir."\\c".$tmp;

      my $footer_file_u = $sells_dir."\\R".$tmp;
      my $footer_file_l = $sells_dir."\\r".$tmp;

      if (! -e $detail_file_u && ! -e $detail_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de cabealhos\n";
      }
      elsif (! -e $footer_file_u && ! -e $footer_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de rodaps\n";
      }
      else {
        $successes++;
      }
    }
  }

  $s->debug("We found $errors errors and $successes valids...");

  $errors = 0;
  $successes = 0;

  foreach my $f (@$files) {
    if (my( $tmp ) = $f =~ /^r(.*?)$/i) {

      my $detail_file_u = $sells_dir."\\D".$tmp;
      my $detail_file_l = $sells_dir."\\d".$tmp;

      my $footer_file_u = $sells_dir."\\C".$tmp;
      my $footer_file_l = $sells_dir."\\c".$tmp;

      if (! -e $detail_file_u && ! -e $detail_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de detalhes\n";
      }
      elsif (! -e $footer_file_u && ! -e $footer_file_l) {
        $errors++;
        print STDERR "($f) no existe ficheiro de cabealhos\n";
      }
      else {
        $successes++;
      }
    }
  }

  $s->debug("We found $errors errors and $successes valids...");

  return $errors == 0 ? 1 : 0;
}

sub backup_file {
  my $s = shift;

  my %a = ( num => 2, file => undef, @_ );

  &File::Basename::fileparse_set_fstype('MSWin32');

  die "Missing 'file' parameter..." unless defined $a{file};

  return undef unless -e $a{file};

  my( $name, $path, $suffix) = fileparse($a{file}, ());

  my $time_str = strftime("%Y%m%d%H%M%S", localtime(time));

  my @collected = ();

  &find( sub {
    return unless -f $File::Find::name;

    if ($File::Find::name ne $a{file} && /$name\.\d+$/) {
      push @collected, $File::Find::name;
    }
  }, ( $path ) );

  &copy( $a{file}, $a{file}.'.'.$time_str );

  my $c = 0;

  for (reverse sort @collected) {
    next if ++$c < $a{num};
    #$s->debug("backup_file -> will remove $_");
    $s->own_remove( $_ );
  }
}

sub cserver_passwords {
  my @p = ();

  my @cdirs = ();

  if ($^O eq 'MSWin32') {
    push @cdirs, 'C:\Program Files\GrupoPIE\ComServer';
    push @cdirs, 'C:\Program Files\Restinfor\ComServer';
    push @cdirs, 'C:\Programas\GrupoPIE\ComServer';
    push @cdirs, 'C:\Programas\Restinfor\ComServer';
  } else {
    push @cdirs, '/home/comserver';
  }

  for (@cdirs) {
    next unless -d $_;

    if (open(F, "< $_/config.ini")) {
      while(<F>) {
        chop;
        push @p, $1 if /^EPassword_Text=(.*?)$/i;
      }

      close(F);
    }
  }

  return wantarray? @p : \@p;
}

sub EncriptFOPassword {
  my $pass = shift;
  return unless defined $pass && $pass ne '';

  my @troca = ( 1, 9, 2, 8, 3, 7, 4, 6, 5, 0 );
  my @ordem = ( 2, 0, 3, 1 );

  my $buf = sprintf("%04d", $pass);
  my @chars = split //, $buf;
  my $buf2;

  for (my $i = 0; $i < 4; $i++) {
    $buf2 .= $troca[$chars[$ordem[$i]]];
  }

  return int($buf2);
}

sub DecriptFOPassword {
  my $cript = shift;
  return unless defined $cript && $cript ne '';

  my @troca = ( 9, 0, 2, 4, 6, 8, 7, 5, 3, 1 );
  my @ordem = ( 1, 3, 0, 2 );

  my $buf = sprintf("%04d", $cript);
  my @chars = split //, $buf;
  my $buf2;

  for (my $i = 0; $i < 4; $i++) {
    $buf2 .= $troca[$chars[$ordem[$i]]];
  }

  return int($buf2);
}

sub get_lic_key {
  my( $file, $key ) = @_;

  my $version = "";

  if (open(F, "< $file")) {
    while (<F>) {
      if (/<$key>(\d+)/i) {
        $version = $1;
        last;
      }
    }
    close(F);
  }

  return $version;
}

sub _filter_deleted_from_fo_file {
  my( $file, $char ) = @_;

  my $fname = $file.".filtered";

  open(F, "< $file") || die "Can't open $file: $!";
  open(R, "> $fname") || die "Can't write to $fname: $!";

  while (<F>) {
    my $deleted = substr $_, $char, 1;
    print R $_ if $deleted && $deleted eq 'F';
  }

  close(F);
  close(R);

  return $fname;
}

sub _filter_from_parameters {
  my( $params_file, $params_tmp_file, $filter ) = @_;

  my @lines = &read_file( $params_file );

  open(F, "> $params_tmp_file") or die "Can't write to $params_tmp_file: $!";

  for (@lines) {
    my $type = substr($_, 0, 4);

    my $to_ignore = 0;

    for my $f (@$filter) {
      if ($f eq $type) {
        $to_ignore = 1;
        last;
      }
    }

    next if $to_ignore;
    print F $_;      
  }

  close(F);
}

sub _read_file_sizes {
  my $s = shift;

  my $dsep = $s->directory_separator();

  my $file = $Bin.$dsep.".file_sizes.cache";

  my $r = {};

  if (open(CACHE, "< $file")) {
    while(<CACHE>) {
      chop;
      my( $file, $size ) = split /=/, $_;
      $r->{$file} = $size;
    }
    close(CACHE);
  }

  return $r;
}

sub _save_file_sizes {
  my $s = shift;

  my $fs = $s->{__file_sizes} || {};

  my $dsep = $s->directory_separator();

  my $file = $Bin.$dsep.".file_sizes.cache";

  if (scalar keys %$fs) {
    if (open(CACHE, "> $file")) {
      foreach my $f (keys %$fs) {
        print CACHE "$f=".$fs->{$f}."\n";
      }
      close(CACHE);
    }
  }

  return 1;
}

1;











