package WinRest::File::Fields;

use strict;
use warnings;
use Tie::File;
use Data::Dumper;
use Encode qw( decode encode );
use Digest::MD5 qw( md5_hex );
use File::Path qw( mkpath rmtree );
use Compress::Zlib;
use Carp qw( confess );

$Carp::CarpLevel = 1;

sub cache_db_dir    { 'weckoffice_caches' }
sub cache_key_field { confess ref(shift())." -> You must override the 'cache_key_field' method...\n" }

sub fields { confess ref(shift())." -> You must override the 'fields' method...\n" }

sub defaults { () }

#
# trim_words parameter, when disabled, will not remove spaces from left or write of the word
#

sub new {
  my $c = shift;
  my %a = ( die_on_error => 0, file => undef, trim_words => 1, cache_usage => 0, @_ );

  my $s = bless {%a}, ref($c) || $c;

  confess "Missing the 'file' parameter, " unless defined $a{file};

  if (! -e $a{file}) {
    confess "File $a{file} don't exists\n";
  }

  if ($a{cache_usage}) {
    confess "'cache_usage' can't be used on Win32 systems due to the lack of BerkeleyDB::Lite module" if $^O eq 'MSWin32';
    confess "Missing the 'cache_root_dir' parameter, " unless defined $a{cache_root_dir};
    my $cache_dir = $a{cache_root_dir}."/".$s->cache_db_dir();
    mkpath($cache_dir) unless -d $cache_dir;
  }

  my @lines = ();

  # this code handles zip files or normal ones
  my $gz = gzopen($a{file}, "rb") or confess "Cannot open $a{file}: $gzerrno\n" ;
  my $buffer;

  while ($gz->gzreadline($buffer) > 0) {
    $buffer =~ s/\n//;
    $buffer =~ s/\r//;
    next unless $buffer ne "";
    push @lines, $buffer;
  }

  $s->{__lines} = \@lines;

  return $s;
}

sub create {
  my $c = shift;
  my %a = ( file => undef, fields => {}, line_end => "\n", @_ );

  confess "You didn't specify nothing in the 'fields' parameter\n"
    unless scalar keys %{$a{fields}};

  my $s = bless {%a}, ref($c) || $c;

  # all fields with sizes of this class
  my %fields = ();
  map { my @k = keys %$_; $fields{$k[0]} = $_->{$k[0]} } $s->fields();

  # validate received fields
  foreach my $k (keys %{$a{fields}}) {
    if (! exists $fields{$k}) {
      confess "You're trying to set field '$k' but it's not defined in ".ref($s)."\n";
    }
  }

  # set field defaults
  my %defaults = ();
  map { my @k = keys %$_; $defaults{$k[0]} = $_->{$k[0]} } $s->defaults();

  my $buff = "";

  # generate field line
  foreach my $field ($s->fields()) {
    my( $name, $len ) = each %$field;

    my $field_value = "";

    if (exists $a{fields}->{$name} && defined $a{fields}->{$name}) {
      $field_value = $a{fields}->{$name};
    } else {
      $field_value = $defaults{$name};
    }

    $field_value = defined $field_value ? $field_value : "";
    $field_value =~ s/[\000-\011]|[\013-\037]//g;
    $field_value = &_encode_str($field_value);

    if (defined $field_value) {
#      if ($field_value =~ /^\d+$/ || $field_value =~ /^\d+\.\d+$/) {
#        $field_value = sprintf("%".$len."s", $field_value);
#      } else {
        $field_value = sprintf("%-".$len."s", $field_value);
#      }
      $field_value = substr($field_value, 0, $len);
      $buff .= $field_value;

#      print STDERR "## $name, $len, ".length($field_value)."\n";
    }
    # this is a HARD FIX!! see WinRest::File::Waitresses -> memp_tmp
    elsif ($name eq 'memp_tmp') {
      print STDERR "CRAP\n";
      $buff .= " ";
    }
  }

  open(F, ">> $a{file}") || confess "Error appending to $a{file}: $!\n";
  #print F "$buff\r\n";
  print F "$buff".$a{line_end};
  close(F);
}

sub set_lines {
  my $s = shift;
  my $l = shift || [];
  $s->{__lines} = $l;
}

sub total_char_number {
  my $s = shift;

  my $total_len = 0;

  foreach my $field ($s->fields()) {
    my( $name, $length ) = each %$field;
    $total_len += $length;
  }

  return $total_len;
}

sub get_all {
  my $s = shift;
  return $s->search( all => 1 );
}

sub search {
  my $s = shift;
  my %a = ( field => undef, value => undef, all => 0, @_ );

  my $db = undef;

  if ($s->{cache_usage}) {
    require BerkeleyDB::Lite;

    $db = BerkeleyDB::Lite::Hash->new(
      home     => $s->cache_db_dir(),
      rootdir  => $s->{cache_root_dir},
      filename => ref($s),
    );
  }

  my @results = ();
  my $total_char_number = $s->total_char_number();
  my $line_count = 0;

  foreach my $line (@{$s->{__lines}}) {
    $line =~ s/
//g;

    $line_count++;

    my $proc = 0;
    my $line_len = length $line;
    my %tmp = ();
    my $found = 0;

    if ($line_len != $total_char_number) {
      my $diff = abs($line_len - $total_char_number);

      if ($s->{die_on_errors}) {
        # Winrest sometimes dupplicate FOUR LAST FIELDS, this ignores them
        if ($diff > 9) {
          confess ref($s).": Line $line_count in file ".$s->{file}." should have $total_char_number chars and have $line_len...\n";
        }
      } else {
        # Winrest sometimes dupplicate FOUR LAST FIELDS, this ignores them
        if ($diff > 9) {
          print STDERR ref($s).": Line $line_count in file ".$s->{file}." should have $total_char_number chars and have $line_len...\n";
        }
      }
    }

    eval {
      foreach my $field ($s->fields()) {
        my( $name, $length ) = each %$field;
        
        my $value;

        if ($s->{trim_words}) {
          $value = &_trim_word(substr($line, $proc, $length));
        } else {
          $value = substr($line, $proc, $length);
        }
   
        $proc += $length;

        $tmp{ $name } = $value;

        if ($a{all} == 1) {
          $found = 1;
        } elsif ($name eq $a{field} && $value eq $a{value}) {
          $found = 1;
        }
      }
    };

    if ($@) {
      print STDERR "Error parsing line $line_count ($@)... moving on...\n";
      next;
    }

    if ($found > 0) {      
      my $rec = WinRest::File::Fields::Record->new();

      if ($s->{cache_usage} && defined $db) {
        if (! exists $tmp{ $s->cache_key_field() }) {
          confess "You're trying to use '".$s->cache_key_field()."' as a cache_key_field but I didn't find that field...";
        } else {
          if (exists $db->{$tmp{ $s->cache_key_field() }}) {
            my $md5 = $s->_gen_md5(\%tmp);
            my $cached_value = $db->{$tmp{ $s->cache_key_field() }}; # scalar ref

            if ($$cached_value eq $md5) {
              $rec->is_new(0);
              $rec->is_update(0);
            } else {
              $rec->is_new(0);
              $rec->is_update(1);
              $db->{$tmp{ $s->cache_key_field() }} = $md5;
            }
          } else {
            $db->{$tmp{ $s->cache_key_field() }} = $s->_gen_md5(\%tmp);
            $rec->is_new(1);
            $rec->is_update(0);
          }
        }
      } else {
        $rec->is_new(0);
        $rec->is_update(0);
      }

      map { $rec->$_($tmp{$_}) } keys %tmp;
      
      # saves the original string line of the record
      $rec->line_str($line);
      
      push @results, $rec;
    }
  }

  return wantarray? @results : \@results;
}

sub _trim_word {
  my $w = shift;

  if (defined $w) {
    $w =~ s/^\s+//;
    $w =~ s/\s+$//;

    return $w;
  } else {
    return "";
  }
}

sub _decode_str {
  my $w = shift;
  return encode("iso-8859-1", decode("cp850", $w));
}

sub _encode_str {
  my $w = shift;
  return encode("cp850", $w);
}

sub _gen_md5 {
  my $s = shift;
  my $r = shift;

  my $str = "";
  my @keys = sort { $a cmp $b } keys %$r;
  map { $str .= $r->{$_} if defined $r->{$_} } @keys;

  return md5_hex( $str );
}

package WinRest::File::Fields::Record;

use strict;
use warnings;
use vars qw( $AUTOLOAD );

sub new {
  my $c = shift;
  my %a = @_;

  my $s = bless {%a}, ref($c) || $c;

  return $s;
}

sub AUTOLOAD {
  my $s = shift;
  my $type = ref($s) || croak("$s is not an object");
  my $name = $AUTOLOAD;
  $name =~ s/.*://;

  return unless $name ne 'DESTROY';

  if (@_) {
    my $v = shift;
    if (defined $v && !ref($v)) {
      return $s->{$name} = $v;
    }
    return $s->{$name} = $v;
  }
  # this is called from the parent class because if the method was in this
  # one, we would have a deep recursion in the AUTOLOAD
  return &WinRest::File::Fields::_decode_str($s->{$name});
}

1;

