package WinTouch::Export;

use strict;
use warnings;
use Data::Dumper;
use DB_File;
use Digest::MD5 qw( md5_hex );
use FindBin qw( $Bin );

#
# Overridable methods
#

sub fields { return () }

sub defaults { return () }

sub key_fields { return () }

#
# Public Methods
#

sub new {
  my $c = shift;
  my %a = @_;

  my $s = bless {%a}, ref($c) || $c;

  $s->_init_fields();

  $s->{_records} = [];

  return $s;
}

sub records {
  my $s = shift;
  my %a = ( validate_fields => 0, data => [], @_ );

  if (scalar @{ $a{data} }) {
    my @fields = $s->fields();

    foreach my $rec (@{ $a{data} }) {
      my $record = {};

      if ($a{validate_fields}) {
        for (keys %$rec) {
          if (! exists $s->{__fields}->{$_}) {
            #print STDERR "Invalid field used in ".ref($s).": $_, ignoring...\n";
            #next;
            die "Invalid field used in ".ref($s).": $_\n" unless exists $s->{__fields}->{$_};
          }
        }
      }

      foreach my $field (@fields) {
        $record->{ $field } = exists $rec->{ $field } ? $rec->{ $field } : $s->{__fields}->{$field};
      }

      push @{ $s->{_records} }, $record;
    }
  } else {
    return wantarray? @{$s->{_records}} : $s->{_records};
  }
}

sub export_name {
  my $s = shift;
  my $cname = ref($s);
  $cname =~ s/^.*\:\:(.*?)$/$1/;
  return $cname;
}

sub to_array {
  my $s = shift;

  my @lines = ();

  my @fields = $s->fields();
  my @records = $s->records();

  push @lines, "I" . $s->export_name() . "," . join(",", @fields);

  foreach my $record (@records) {
    my @rec = ();

    foreach my $field (@fields) {
      push @rec, $record->{$field};
    }

    push @lines, "D" . join("", @rec) . "";
  }

  return wantarray? @lines : \@lines;
}

sub to_fh {
  my $s = shift;
  my %a = ( fh => undef, cache_control => 0, @_ );

  die "Missing 'fh' parameter..." unless defined $a{fh};

  my $fh = $a{fh};

  my %cache_info;
  my $cache_just_init = 0;

  if ($a{cache_control}) {
    if (! -e $s->cache_file()) {
      $s->_init_cache();
      $cache_just_init = 1;
    } else {
      tie %cache_info, "DB_File", $s->cache_file();
    } 
  }

  my @fields = $s->fields();
  my @records = $s->records();

  print $fh "I" . $s->export_name() . "," . join(",", @fields) . "\r\n";

  my @keys = $s->key_fields();

  foreach my $record (@records) {
    # does cache control only if available
    if ($a{cache_control} && scalar @keys && ! $cache_just_init) {
      # build key str
      my $key_str;
      map { $key_str .= $record->{$_} if defined $record->{$_} } @keys;

      my $value_str;
      map { $value_str .= $record->{$_} if defined $record->{$_} } sort keys %$record;

      my $gen_md5 = &md5_hex($value_str);

      if (! exists $cache_info{ $key_str }) {
        print STDERR "Don't exists in cache...\n";
        $cache_info{ $key_str } = $gen_md5;
      } elsif ($gen_md5 eq $cache_info{ $key_str }) {
        next;
      } else {
        print STDERR "Strange... $gen_md5 - $key_str - $value_str\n";
        $cache_info{ $key_str } = $gen_md5;
      }
    }

    #print STDERR ref($s)." -> Will process one...\n";

    my @rec = ();

    foreach my $field (@fields) {
      push @rec, $record->{$field} || "";
    }

    print $fh "D" . join("", @rec) . "\r\n";
  }

  return 1;
}

sub cache_file {
  my $s = shift;
  return $Bin."/".$s->class_to_name().".dat";
}

sub class_to_name {
  my $s = shift;

  my $c = ref($s);
  $c =~ s/\://g;

  return $c;
}

#
# Internal Methods
#

sub _init_cache {
  my $s = shift;

  my @keys = $s->key_fields();
  return unless scalar @keys;

  my @records = $s->records();
  return unless scalar @records;

  my %info = ();
  tie %info, "DB_File", $s->cache_file();

  foreach my $record (@records) {
    # build key str
    my $keys_str;
    map { $keys_str .= $record->{$_} if defined $record->{$_} } @keys;

    my $values_str;
    map { $values_str .= $record->{$_} if defined $record->{$_} } sort keys %$record;

    $info{ $keys_str } = &md5_hex( $values_str );
  }
}

sub _init_fields {
  my $s = shift;

  my @f = $s->fields();
  my %d = $s->defaults();

  map { $s->{__fields}->{$_} = exists $d{$_} ? $d{$_} : "" } @f;
}

1;

